@@ -96,7 +96,7 @@ two_factor_bound(GEN x)
   z = shiftr(sqrtr(z), n);
   z = divrr(z, dbltor(pow((double)n, 0.75)));
   z = roundr_safe(sqrtr(z));
-  z = mulii(z, absi(gel(x,n)));
+  z = mulii(z, absi_shallow(gel(x,n)));
   return gerepileuptoint(av, shifti(z, 1));
 }
 #endif
@@ -106,7 +106,7 @@ static GEN
 Mignotte_bound(GEN S)
 {
   long i, d = degpol(S);
-  GEN C, N2, t, binlS, lS = leading_coeff(S), bin = vecbinome(d-1);
+  GEN C, N2, t, binlS, lS = leading_coeff(S), bin = vecbinomial(d-1);
 
   N2 = sqrtr(RgX_fpnorml2(S,DEFAULTPREC));
   binlS = is_pm1(lS)? bin: ZC_Z_mul(bin, lS);
@@ -130,7 +130,7 @@ Beauzamy_bound(GEN S)
   const long prec = DEFAULTPREC;
   long i, d = degpol(S);
   GEN bin, lS, s, C;
-  bin = vecbinome(d);
+  bin = vecbinomial(d);
 
   s = real_0(prec);
   for (i=0; i<=d; i++)
@@ -143,7 +143,7 @@ Beauzamy_bound(GEN S)
   /* s = [S]_2^2 */
   C = powruhalf(stor(3,prec), 3 + 2*d); /* 3^{3/2 + d} */
   C = divrr(mulrr(C, s), mulur(4*d, mppi(prec)));
-  lS = absi(leading_coeff(S));
+  lS = absi_shallow(leading_coeff(S));
   return mulir(lS, sqrtr(C));
 }
 
@@ -158,7 +158,7 @@ factor_bound(GEN S)
     err_printf("Mignotte bound: %Ps\n",a);
     err_printf("Beauzamy bound: %Ps\n",b);
   }
-  return gerepileupto(av, ceil_safe(gmin(a, b)));
+  return gerepileupto(av, ceil_safe(gmin_shallow(a, b)));
 }
 
 /* Naive recombination of modular factors: combine up to maxK modular
@@ -184,8 +184,8 @@ cmbf(GEN pol, GEN famod, GEN bound, GEN p, long a, long b,
   GEN fa       = cgetg(lfamod+1, t_VEC);
 
   *pmaxK = cmbf_maxK(lfamod);
-  lc = absi(leading_coeff(pol));
-  if (is_pm1(lc)) lc = NULL;
+  lc = absi_shallow(leading_coeff(pol));
+  if (equali1(lc)) lc = NULL;
   lcpol = lc? ZX_Z_mul(pol, lc): pol;
 
   {
@@ -267,7 +267,7 @@ nextK:
         if (y) q = mulii(y, q);
         y = centermodii(q, pa, pas2);
       }
-      if (!signe(y) || remii(constant_coeff(lcpol), y) != gen_0)
+      if (!signe(y) || !dvdii(constant_coeff(lcpol), y))
       {
         if (DEBUGLEVEL>3) err_printf("T");
         avma = av; goto NEXT;
@@ -312,7 +312,7 @@ nextK:
       if (lfamod < 2*K) goto END;
       i = 1; curdeg = deg[ind[1]];
       bound = factor_bound(pol);
-      if (lc) lc = absi(leading_coeff(pol));
+      if (lc) lc = absi_shallow(leading_coeff(pol));
       lcpol = lc? ZX_Z_mul(pol, lc): pol;
       if (DEBUGLEVEL>3)
         err_printf("\nfound factor %Ps\nremaining modular factor(s): %ld\n",
@@ -366,12 +366,12 @@ shifteval(GEN Q, long n)
 static GEN
 root_bound(GEN P0)
 {
-  GEN Q = leafcopy(P0), lP = absi(leading_coeff(Q)), x,y,z;
+  GEN Q = leafcopy(P0), lP = absi_shallow(leading_coeff(Q)), x,y,z;
   long k, d = degpol(Q);
 
   /* P0 = lP x^d + Q, deg Q < d */
   Q = normalizepol_lg(Q, d+2);
-  for (k=lg(Q)-1; k>1; k--) gel(Q,k) = absi(gel(Q,k));
+  for (k=lg(Q)-1; k>1; k--) gel(Q,k) = absi_shallow(gel(Q,k));
   k = (long)(fujiwara_bound(P0));
   for (  ; k >= 0; k--)
   {
@@ -419,15 +419,15 @@ chk_factors(GEN P, GEN M_L, GEN bound, GEN famod, GEN pa)
 
   r  = lg(piv)-1;
   list = cgetg(r+1, t_VEC);
-  lt = absi(leading_coeff(pol));
-  if (is_pm1(lt)) lt = NULL;
+  lt = absi_shallow(leading_coeff(pol));
+  if (equali1(lt)) lt = NULL;
   ltpol = lt? ZX_Z_mul(pol, lt): pol;
   paov2 = shifti(pa,-1);
   for (i = 1;;)
   {
     if (DEBUGLEVEL) err_printf("LLL_cmbf: checking factor %ld\n",i);
     y = chk_factors_get(lt, famod, gel(piv,i), NULL, pa);
-    y = FpX_center(y, pa, paov2);
+    y = FpX_center_i(y, pa, paov2);
     if (! (pol = ZX_divides_i(ltpol,y,bound)) ) return NULL;
     if (lt) y = Q_primpart(y);
     gel(list,i) = y;
@@ -436,7 +436,7 @@ chk_factors(GEN P, GEN M_L, GEN bound, GEN famod, GEN pa)
     if (lt)
     {
       pol = ZX_Z_divexact(pol, leading_coeff(y));
-      lt = absi(leading_coeff(pol));
+      lt = absi_shallow(leading_coeff(pol));
       ltpol = ZX_Z_mul(pol, lt);
     }
     else
@@ -488,14 +488,14 @@ LLL_cmbf(GEN P, GEN famod, GEN p, GEN pa, GEN bound, long a, long rec)
   const long N0 = 1; /* # of traces added at each step */
   double BitPerFactor = 0.4; /* nb bits in p^(a-b) / modular factor */
   long i,j,tmax,n0,C, dP = degpol(P);
-  double logp = log((double)itos(p)), LOGp2 = LOG2/logp;
+  double logp = log((double)itos(p)), LOGp2 = M_LN2/logp;
   double b0 = log((double)dP*2) / logp, logBr;
   GEN lP, Br, Bnorm, Tra, T2, TT, CM_L, m, list, ZERO;
   pari_sp av, av2;
   long ti_LLL = 0, ti_CF  = 0;
 
-  lP = absi(leading_coeff(P));
-  if (is_pm1(lP)) lP = NULL;
+  lP = absi_shallow(leading_coeff(P));
+  if (equali1(lP)) lP = NULL;
   Br = root_bound(P);
   if (lP) Br = mulii(lP, Br);
   logBr = gtodouble(glog(Br, DEFAULTPREC)) / logp;
@@ -642,7 +642,7 @@ AGAIN:
 static int
 cmbf_precs(GEN q, GEN A, GEN B, long *pta, long *ptb, GEN *qa, GEN *qb)
 {
-  long a,b,amin,d = (long)(31 * LOG2/gtodouble(glog(q,DEFAULTPREC)) - 1e-5);
+  long a,b,amin,d = (long)(31 * M_LN2/gtodouble(glog(q,DEFAULTPREC)) - 1e-5);
   int fl = 0;
 
   b = logintall(B, q, qb) + 1;
@@ -678,7 +678,7 @@ combine_factors(GEN target, GEN famod, GEN p, long klim)
 
   A = factor_bound(target);
 
-  la = absi(leading_coeff(target));
+  la = absi_shallow(leading_coeff(target));
   B = mului(n, sqri(mulii(la, root_bound(target)))); /* = bound for S_2 */
 
   (void)cmbf_precs(p, A, B, &a, &b, &pa, &pb);
@@ -985,8 +985,7 @@ nfrootsQ(GEN x)
   x = Q_primpart(x);
   RgX_check_ZX(x,"nfrootsQ");
   val = ZX_valrem(x, &x);
-  (void)ZX_gcd_all(x, ZX_deriv(x), &x);
-  z = DDF_roots(x);
+  z = DDF_roots( ZX_radical(x) );
   if (val) z = shallowconcat(z, gen_0);
   return gerepileupto(av, sort(z));
 }
@@ -1027,7 +1026,7 @@ maxnorm(GEN p)
     if (abscmpii(x,m) > 0) m = x;
   }
   m = divii(m, gel(p,n));
-  return gerepileuptoint(av, addis(absi(m),1));
+  return gerepileuptoint(av, addiu(absi_shallow(m),1));
 }
 #endif
 
@@ -1116,6 +1115,8 @@ ZX_gcd_all(GEN A, GEN B, GEN *Anew)
 }
 GEN
 ZX_gcd(GEN A, GEN B) { return ZX_gcd_all(A,B,NULL); }
+GEN
+ZX_radical(GEN A) { GEN B; (void)ZX_gcd_all(A,ZX_deriv(A),&B); return B; }
 
 static GEN
 _gcd(GEN a, GEN b)
@@ -1133,7 +1134,7 @@ QX_gcd(GEN A0, GEN B0)
 
   D = ZX_gcd(Q_primitive_part(A0, &a), Q_primitive_part(B0, &b));
   av2 = avma; a = _gcd(a,b);
-  if (isint1(a)) avma = av2; else D = RgX_Rg_mul(D, a);
+  if (isint1(a)) avma = av2; else D = ZX_Q_mul(D, a);
   return gerepileupto(av, D);
 }
 
@@ -1228,10 +1229,10 @@ BD(GEN f)
   { /* contains all Phi_n originally dividing f, n = 2 mod 4, n > 2;
      * and only those
      * In that case, Graeffe(Phi_n) = Phi_{n/2}, and Phi_n = Phi_{n/2}(-x) */
-    Gi = BD_odd(ZX_unscale(f, gen_m1));
+    Gi = BD_odd(ZX_z_unscale(f, -1));
     if (Gi)
     { /* N.B. Phi_2 does not divide f */
-      Gi = ZX_unscale(Gi, gen_m1);
+      Gi = ZX_z_unscale(Gi, -1);
       /* Gi is the product of all Phi_n | f, n = 2 mod 4 */
       G = myconcat(G, Gi);
     }
@@ -1251,8 +1252,7 @@ polcyclofactors(GEN f)
   RgX_check_ZX(f,"polcyclofactors");
   if (degpol(f))
   {
-    (void)ZX_gcd_all(f, ZX_deriv(f), &f);
-    f = BD(f);
+    f = BD(ZX_radical(f));
     if (f) return gerepilecopy(av, f);
   }
   avma = av; return cgetg(1,t_VEC);
@@ -1327,7 +1327,7 @@ BD_iscyclo(GEN f)
   /* f = product of Phi_n, n odd */
   if (ZX_equal(f, f1)) { avma = av; return BD_odd_iscyclo(f); }
 
-  fn = ZX_unscale(f, gen_m1); /* f(-x) */
+  fn = ZX_z_unscale(f, -1); /* f(-x) */
   /* f = product of Phi_n, n = 2 mod 4 */
   if (ZX_equal(f1, fn)) return 2*BD_odd_iscyclo(fn);
 
@@ -1360,7 +1360,7 @@ poliscycloprod(GEN f)
   long i, d = degpol(f);
   if (typ(f) != t_POL) pari_err_TYPE("poliscycloprod",f);
   if (!RgX_is_ZX(f)) return 0;
-  if (!equali1(leading_coeff(f)) || !is_pm1(constant_coeff(f))) return 0;
+  if (!ZX_is_monic(f) || !is_pm1(constant_coeff(f))) return 0;
   if (d < 2) return (d == 1);
   if ( degpol(ZX_gcd_all(f, ZX_deriv(f), &f)) )
   {
