@@ -368,7 +368,7 @@ RgX_rescale(GEN P, GEN h)
 {
   long i, l = lg(P);
   GEN Q = cgetg(l,t_POL), hi = h;
-  Q[l-1] = P[l-1];
+  gel(Q,l-1) = gel(P,l-1);
   for (i=l-2; i>=2; i--)
   {
     gel(Q,i) = gmul(gel(P,i), hi);
