@@ -371,6 +371,7 @@ ZX_rescale_lt(GEN P)
 static GEN
 ZX_eval2BILspec(GEN x, long k, long nx)
 {
+  pari_sp av = avma;
   long i,j, lz = k*nx, ki;
   GEN pz = cgetipos(2+lz);
   GEN nz = cgetipos(2+lz);
@@ -390,7 +391,7 @@ ZX_eval2BILspec(GEN x, long k, long nx)
       for (j=0; j<lc; j++) *int_W(nz,j+ki) = *int_W(c,j);
   }
   pz = int_normalize(pz,0);
-  nz = int_normalize(nz,0); return subii(pz,nz);
+  nz = int_normalize(nz,0); return gerepileuptoint(av, subii(pz,nz));
 }
 
 static long
@@ -417,7 +418,7 @@ Z_mod2BIL_ZX(GEN x, long bs, long d, long vx)
   {
     pari_sp av = avma;
     long lz = minss(bs, lm-offset);
-    GEN z = adduispec_offset(carry, x, offset, lz);
+    GEN z = lz > 0 ?adduispec_offset(carry, x, offset, lz): utoi(carry);
     if (lgefint(z) == 3+bs) { carry = 1; z = gen_0;}
     else
     {
@@ -675,22 +676,13 @@ ZXV_equal(GEN V, GEN W)
 }
 
 GEN
-ZXV_Z_mul(GEN y, GEN x)
-{
-  long i, l;
-  GEN z = cgetg_copy(y, &l);
-  for(i=1; i<l; i++) gel(z,i) = ZX_Z_mul(gel(y,i), x);
-  return z;
-}
+ZXV_Z_mul(GEN x, GEN y)
+{ pari_APPLY_same(ZX_Z_mul(gel(x,i), y)) }
 
 GEN
-ZXV_remi2n(GEN y, long N)
-{
-  long i, l;
-  GEN z = cgetg_copy(y, &l);
-  for(i=1; i<l; i++) gel(z,i) = ZX_remi2n(gel(y,i), N);
-  return z;
-}
+ZXV_remi2n(GEN x, long N)
+{ pari_APPLY_same(ZX_remi2n(gel(x,i), N)) }
+
 GEN
 ZXV_dotproduct(GEN x, GEN y)
 {
@@ -707,6 +699,213 @@ ZXV_dotproduct(GEN x, GEN y)
   return gerepileupto(av, c);
 }
 
+/*******************************************************************/
+/*                                                                 */
+/*                                ZXQM                             */
+/*                                                                 */
+/*******************************************************************/
+
+GEN
+ZXn_mul(GEN x, GEN y, long n)
+{ return RgXn_red_shallow(ZX_mul(x, y), n); }
+
+GEN
+ZXn_sqr(GEN x, long n)
+{ return RgXn_red_shallow(ZX_sqr(x), n); }
+
+/*******************************************************************/
+/*                                                                 */
+/*                                ZXQM                             */
+/*                                                                 */
+/*******************************************************************/
+
+static long
+ZX_expi(GEN x)
+{
+  if (signe(x)==0) return 0;
+  if (typ(x)==t_INT) return expi(x);
+  return ZX_expispec(x+2, lgpol(x));
+}
+
+static long
+ZXC_expi(GEN x)
+{
+  long i, l = lg(x), m=0;
+  for(i = 1; i < l; i++)
+  {
+    long e = ZX_expi(gel(x,i));
+    if (e > m) m = e;
+  }
+  return m;
+}
+
+static long
+ZXM_expi(GEN x)
+{
+  long i, l = lg(x), m=0;
+  for(i = 1; i < l; i++)
+  {
+    long e = ZXC_expi(gel(x,i));
+    if (e > m) m = e;
+  }
+  return m;
+}
+
+static GEN
+ZX_eval2BIL(GEN x, long k)
+{
+  if (signe(x)==0) return gen_0;
+  if (typ(x)==t_INT) return x;
+  return ZX_eval2BILspec(x+2, k, lgpol(x));
+}
+
+/*Eval x in 2^(k*BIL) in linear time*/
+static GEN
+ZXC_eval2BIL(GEN x, long k)
+{
+  long i, lx = lg(x);
+  GEN A = cgetg(lx, t_COL);
+  for (i=1; i<lx; i++) gel(A,i) = ZX_eval2BIL(gel(x,i), k);
+  return A;
+}
+
+static GEN
+ZXM_eval2BIL(GEN x, long k)
+{
+  long i, lx = lg(x);
+  GEN A = cgetg(lx, t_MAT);
+  for (i=1; i<lx; i++) gel(A,i) = ZXC_eval2BIL(gel(x,i), k);
+  return A;
+}
+
+static GEN
+Z_mod2BIL_ZXQ(GEN x, long bs, GEN T)
+{
+  pari_sp av = avma;
+  long v = varn(T), d = 2*(degpol(T)-1);
+  GEN z = Z_mod2BIL_ZX(x, bs, d, 0);
+  setvarn(z, v);
+  return gerepileupto(av, ZX_rem(z, T));
+}
+
+static GEN
+ZC_mod2BIL_ZXQC(GEN x, long bs, GEN T)
+{
+  long i, lx = lg(x);
+  GEN A = cgetg(lx, t_COL);
+  for (i=1; i<lx; i++) gel(A,i) = Z_mod2BIL_ZXQ(gel(x,i), bs, T);
+  return A;
+}
+
+static GEN
+ZM_mod2BIL_ZXQM(GEN x, long bs, GEN T)
+{
+  long i, lx = lg(x);
+  GEN A = cgetg(lx, t_MAT);
+  for (i=1; i<lx; i++) gel(A,i) = ZC_mod2BIL_ZXQC(gel(x,i), bs, T);
+  return A;
+}
+
+GEN
+ZXQM_mul(GEN x, GEN y, GEN T)
+{
+  long d = degpol(T);
+  GEN z;
+  pari_sp av = avma;
+  if (d == 0)
+    z = ZM_mul(simplify_shallow(x),simplify_shallow(y));
+  else
+  {
+    long ex = ZXM_expi(x), ey = ZXM_expi(y), d= degpol(T), n = lg(x)-1;
+    long e = ex + ey + expu(d) + expu(n) + 4;
+    long N = divsBIL(e)+1;
+    z = ZM_mul(ZXM_eval2BIL(x,N), ZXM_eval2BIL(y,N));
+    z = ZM_mod2BIL_ZXQM(z, N, T);
+  }
+  return gerepileupto(av, z);
+}
+
+GEN
+ZXQM_sqr(GEN x, GEN T)
+{
+  long d = degpol(T);
+  GEN z;
+  pari_sp av = avma;
+  if (d == 0)
+    z = ZM_sqr(simplify_shallow(x));
+  else
+  {
+    long ex = ZXM_expi(x), d = degpol(T), n = lg(x)-1;
+    long e = 2*ex + expu(d) + expu(n) + 4;
+    long N = divsBIL(e)+1;
+    z = ZM_sqr(ZXM_eval2BIL(x,N));
+    z = ZM_mod2BIL_ZXQM(z, N, T);
+  }
+  return gerepileupto(av, z);
+}
+
+GEN
+QXQM_mul(GEN x, GEN y, GEN T)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN dy, ny = Q_primitive_part(y, &dy);
+  GEN z = ZXQM_mul(nx, ny, T);
+  if (dx || dy)
+  {
+    GEN d = dx ? dy ? gmul(dx, dy): dx : dy;
+    if (!gequal1(d)) z = RgM_Rg_mul(z, d);
+  }
+  return z;
+}
+
+GEN
+QXQM_sqr(GEN x, GEN T)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN z = ZXQM_sqr(nx, T);
+  if (dx) z = RgM_Rg_mul(z, gsqr(dx));
+  return z;
+}
+
+static GEN
+Z_mod2BIL_Fq(GEN x, long bs, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  long v = varn(T), d = 2*(degpol(T)-1);
+  GEN z = Z_mod2BIL_ZX(x, bs, d, 0);
+  setvarn(z, v);
+  return gerepileupto(av, FpX_rem(z, T, p));
+}
+
+static GEN
+ZC_mod2BIL_FqC(GEN x, long bs, GEN T, GEN p)
+{
+  long i, lx = lg(x);
+  GEN A = cgetg(lx, t_COL);
+  for (i=1; i<lx; i++) gel(A,i) = Z_mod2BIL_Fq(gel(x,i), bs, T, p);
+  return A;
+}
+
+static GEN
+ZM_mod2BIL_FqM(GEN x, long bs, GEN T, GEN p)
+{
+  long i, lx = lg(x);
+  GEN A = cgetg(lx, t_MAT);
+  for (i=1; i<lx; i++) gel(A,i) = ZC_mod2BIL_FqC(gel(x,i), bs, T, p);
+  return A;
+}
+
+GEN
+FqM_mul_Kronecker(GEN x, GEN y, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  long ex = ZXM_expi(x), ey = ZXM_expi(y), d= degpol(T), n = lg(x)-1;
+  long e = ex + ey + expu(d) + expu(n) + 4;
+  long N = divsBIL(e)+1;
+  GEN  z = ZM_mul(ZXM_eval2BIL(x,N), ZXM_eval2BIL(y,N));
+  return gerepileupto(av, ZM_mod2BIL_FqM(z, N, T, p));
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                                ZXX                              */
@@ -806,7 +1005,7 @@ ZXX_to_Kronecker_spec(GEN P, long lP, long n)
   {
     long j;
     GEN c = gel(P,i);
-    if (typ(c)==t_INT)
+    if (typ(c)!=t_POL)
     {
       gel(y,k++) = c;
       j = 3;
@@ -830,3 +1029,58 @@ ZXX_to_Kronecker(GEN P, long n)
   GEN z = ZXX_to_Kronecker_spec(P+2, lgpol(P), n);
   setvarn(z,varn(P)); return z;
 }
+
+GEN
+ZXQX_sqr(GEN x, GEN T)
+{
+  pari_sp av = avma;
+  long n = degpol(T);
+  GEN z = ZXX_sqr_Kronecker(x, n);
+  z = Kronecker_to_ZXX(z, n, varn(T));
+  return gerepileupto(av, z);
+}
+
+GEN
+ZXQX_mul(GEN x, GEN y, GEN T)
+{
+  pari_sp av = avma;
+  long n = degpol(T);
+  GEN z = ZXX_mul_Kronecker(x, y, n);
+  z = Kronecker_to_ZXX(z, n, varn(T));
+  return gerepileupto(av, z);
+}
+
+GEN
+QX_mul(GEN x, GEN y)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN dy, ny = Q_primitive_part(y, &dy);
+  GEN z = ZX_mul(nx, ny);
+  if (dx || dy)
+  {
+    GEN d = dx ? dy ? gmul(dx, dy): dx : dy;
+    return ZX_Q_mul(z, d);
+  } else
+    return z;
+}
+
+GEN
+QX_sqr(GEN x)
+{
+  GEN dx, nx = Q_primitive_part(x, &dx);
+  GEN z = ZX_sqr(nx);
+  if (dx)
+    return ZX_Q_mul(z, gsqr(dx));
+  else
+    return z;
+}
+
+GEN
+QX_ZX_rem(GEN x, GEN y)
+{
+  pari_sp av = avma;
+  GEN d, nx = Q_primitive_part(x, &d);
+  GEN r = ZX_rem(nx, y);
+  if (d) r = ZX_Q_mul(r, d);
+  return gerepileupto(av, r);
+}
