@@ -109,7 +109,8 @@ charpoly0(GEN x, long v, long flag)
     case 5:
       return charpoly(x, v);
   }
-  pari_err_FLAG("charpoly"); return NULL; /* not reached */
+  pari_err_FLAG("charpoly");
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* characteristic pol. Easy cases. Return NULL in case it's not so easy. */
@@ -157,7 +158,7 @@ easychar(GEN x, long v)
       return NULL;
   }
   pari_err_TYPE("easychar",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 /* compute charpoly by mapping to Fp first, return lift to Z */
 static GEN
@@ -178,22 +179,29 @@ GEN
 charpoly(GEN x, long v)
 {
   GEN T, p = NULL;
+  long prec;
   if ((T = easychar(x,v))) return T;
-  if (RgM_is_ZM(x))
+  switch(RgM_type(x, &p,&T,&prec))
   {
-    T = ZM_charpoly(x);
-    setvarn(T, v);
+    case t_INT:
+      T = ZM_charpoly(x); setvarn(T, v); break;
+    case t_INTMOD:
+      if (!BPSW_psp(p)) T = carberkowitz(x, v);
+      else
+      {
+        pari_sp av = avma;
+        T = RgM_Fp_charpoly(x,p,v);
+        T = gerepileupto(av, FpX_to_mod(T,p));
+      }
+      break;
+    case t_REAL:
+    case t_COMPLEX:
+    case t_PADIC:
+      T = carhess(x, v);
+      break;
+    default:
+      T = carberkowitz(x, v);
   }
-  else if (RgM_is_FpM(x, &p) && BPSW_psp(p))
-  {
-    pari_sp av = avma;
-    T = RgM_Fp_charpoly(x,p,v);
-    T = gerepileupto(av, FpX_to_mod(T,p));
-  }
-  else if (isinexact(x))
-    T = carhess(x, v);
-  else
-    T = carberkowitz(x, v);
   return T;
 }
 
@@ -375,7 +383,8 @@ matadjoint0(GEN x, long flag)
     case 0: return adj(x);
     case 1: return adjsafe(x);
   }
-  pari_err_FLAG("matadjoint"); return NULL; /* not reached */
+  pari_err_FLAG("matadjoint");
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /*******************************************************************/
@@ -672,7 +681,7 @@ matfrobenius(GEN M, long flag, long v)
   default:
     pari_err_FLAG("matfrobenius");
   }
-  return NULL; /*NOT REACHED*/
+  return NULL; /*LCOV_EXCL_LINE*/
 }
 
 /*******************************************************************/
@@ -684,40 +693,42 @@ matfrobenius(GEN M, long flag, long v)
 static GEN
 easymin(GEN x, long v)
 {
-  pari_sp ltop=avma;
   GEN G, R, dR;
-  if (typ(x)==t_POLMOD && !issquarefree(gel(x,1)))
-    return NULL;
+  if (typ(x)==t_POLMOD && !issquarefree(gel(x,1))) return NULL;
   R = easychar(x, v);
-  if (!R) return R;
-  dR=RgX_deriv(R);
-  if (!lgpol(dR)) {avma=ltop; return NULL;}
-  G=RgX_gcd(R,dR);
-  G=RgX_Rg_div(G,leading_coeff(G));
-  return gerepileupto(ltop, RgX_div(R,G));
+  if (!R) return NULL;
+  dR = RgX_deriv(R);
+  if (!lgpol(dR)) return NULL;
+  G = RgX_normalize(RgX_gcd(R,dR));
+  return RgX_div(R,G);
+}
+static GEN
+RgXQ_minpoly_naive(GEN y, GEN P)
+{
+  long n = lgpol(P);
+  GEN M = ker(RgXQ_matrix_pow(y,n,n,P));
+  return content(RgM_to_RgXV(M,varn(P)));
 }
-
 GEN
 minpoly(GEN x, long v)
 {
-  pari_sp ltop=avma;
+  pari_sp av = avma;
   GEN P;
-  if (v<0) v = 0;
-  if (typ(x)==t_FFELT)
+  if (v < 0) v = 0;
+  if (typ(x) == t_FFELT)
   {
-      GEN p1 = FpX_to_mod(FF_minpoly(x), FF_p_i(x));
-      setvarn(p1,v); return gerepileupto(ltop,p1);
+    P = FpX_to_mod(FF_minpoly(x), FF_p_i(x));
+    setvarn(P,v); return gerepileupto(av,P);
   }
-
   P = easymin(x,v);
-  if (P) return P;
-  if (typ(x)==t_POLMOD)
+  if (P) return gerepileupto(av,P);
+  avma = av;
+  if (typ(x) == t_POLMOD)
   {
-    P = gcopy(RgXQ_minpoly_naive(gel(x,2), gel(x,1)));
-    setvarn(P,v);
-    return gerepileupto(ltop,P);
+    P = RgXQ_minpoly_naive(gel(x,2), gel(x,1));
+    setvarn(P,v); return gerepileupto(av,P);
   }
-  if (typ(x)!=t_MAT) pari_err_TYPE("minpoly",x);
+  if (typ(x) != t_MAT) pari_err_TYPE("minpoly",x);
   if (lg(x) == 1) return pol_1(v);
   return RgM_minpoly(x,v);
 }
@@ -727,24 +738,36 @@ minpoly(GEN x, long v)
 /*                       HESSENBERG FORM                           */
 /*                                                                 */
 /*******************************************************************/
-GEN
-hess(GEN x)
+static int
+relative0(GEN a, GEN a0, long bit)
+{ return gequal0(a) || (bit && gexpo(a)-gexpo(a0) < bit); }
+/* assume x a non-empty square t_MAT */
+static GEN
+RgM_hess(GEN x0, long prec)
 {
-  pari_sp av = avma;
-  long lx = lg(x), m, i, j;
+  pari_sp av;
+  long lx = lg(x0), bit = prec? 8 - bit_accuracy(prec): 0, m, i, j;
+  GEN x;
 
-  if (typ(x) != t_MAT) pari_err_TYPE("hess",x);
-  if (lx == 1) return cgetg(1,t_MAT);
-  if (lgcols(x) != lx) pari_err_DIM("hess");
-
-  x = RgM_shallowcopy(x);
+  if (bit) x0 = RgM_shallowcopy(x0);
+  av = avma; x = RgM_shallowcopy(x0);
   for (m=2; m<lx-1; m++)
   {
     GEN t = NULL;
-    for (i=m+1; i<lx; i++) { t = gcoeff(x,i,m-1); if (!gequal0(t)) break; }
+    for (i=m+1; i<lx; i++)
+    {
+      t = gcoeff(x,i,m-1);
+      if (!relative0(t, gcoeff(x0,i,m-1), bit)) break;
+    }
     if (i == lx) continue;
     for (j=m-1; j<lx; j++) swap(gcoeff(x,i,j), gcoeff(x,m,j));
-    swap(gel(x,i), gel(x,m)); t = ginv(t);
+    swap(gel(x,i), gel(x,m));
+    if (bit)
+    {
+      for (j=m-1; j<lx; j++) swap(gcoeff(x0,i,j), gcoeff(x0,m,j));
+      swap(gel(x0,i), gel(x0,m));
+    }
+    t = ginv(t);
 
     for (i=m+1; i<lx; i++)
     {
@@ -763,7 +786,25 @@ hess(GEN x)
       }
     }
   }
-  return gerepilecopy(av,x);
+  return x;
+}
+
+GEN
+hess(GEN x)
+{
+  pari_sp av = avma;
+  GEN p = NULL, T = NULL;
+  long prec, lx = lg(x);
+  if (typ(x) != t_MAT) pari_err_TYPE("hess",x);
+  if (lx == 1) return cgetg(1,t_MAT);
+  if (lgcols(x) != lx) pari_err_DIM("hess");
+  switch(RgM_type(x, &p, &T, &prec))
+  {
+    case t_REAL:
+    case t_COMPLEX: break;
+    default: prec = 0;
+  }
+  return gerepilecopy(av, RgM_hess(x,prec));
 }
 
 GEN
@@ -837,18 +878,14 @@ FpM_hess(GEN x, GEN p)
   }
   return gerepilecopy(av,x);
 }
-GEN
-carhess(GEN x, long v)
+/* H in Hessenberg form */
+static GEN
+RgM_hess_charpoly(GEN H, long v)
 {
-  pari_sp av;
-  long lx, r, i;
-  GEN y, H;
-
-  if ((H = easychar(x,v))) return H;
-
-  lx = lg(x); av = avma; y = cgetg(lx+1, t_VEC);
-  gel(y,1) = pol_1(v); H = hess(x);
-  for (r = 1; r < lx; r++)
+  long n = lg(H), r, i;
+  GEN y = cgetg(n+1, t_VEC);
+  gel(y,1) = pol_1(v);
+  for (r = 1; r < n; r++)
   {
     pari_sp av2 = avma;
     GEN z, a = gen_1, b = pol_0(v);
@@ -862,7 +899,16 @@ carhess(GEN x, long v)
                 RgX_Rg_mul(gel(y,r), gcoeff(H,r,r)));
     gel(y,r+1) = gerepileupto(av2, RgX_sub(z, b)); /* (X - H[r,r])y[r] - b */
   }
-  return fix_pol(av, gel(y,lx));
+  return gel(y,n);
+}
+GEN
+carhess(GEN x, long v)
+{
+  pari_sp av;
+  GEN y;
+  if ((y = easychar(x,v))) return y;
+  av = avma; y = RgM_hess_charpoly(hess(x), v);
+  return fix_pol(av, y);
 }
 
 /* Bound for sup norm of charpoly(M/dM), M integral: let B = |M|oo / |dM|,
@@ -1050,7 +1096,7 @@ gnorm(GEN x)
     case t_QUAD:    av = avma; return gerepileupto(av, quadnorm(x));
 
     case t_POL: case t_SER: case t_RFRAC: av = avma;
-      return gerepileupto(av, greal(gmul(gconj(x),x)));
+      return gerepileupto(av, greal(gmul(conj_i(x),x)));
 
     case t_FFELT:
       y = cgetg(3, t_INTMOD);
@@ -1069,7 +1115,7 @@ gnorm(GEN x)
       return y;
   }
   pari_err_TYPE("gnorm",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* return |q|^2, complex modulus */
@@ -1104,7 +1150,7 @@ gnorml2_i(GEN x, long prec)
     case t_MAT: lx = lg(x); break;
 
     default: pari_err_TYPE("gnorml2",x);
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   if (lx == 1) return gen_0;
   av = avma;
@@ -1209,7 +1255,7 @@ gnorml1(GEN x,long prec)
       break;
 
     default: pari_err_TYPE("gnorml1",x);
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   return gerepileupto(av, s);
 }
@@ -1245,7 +1291,7 @@ gnorml1_fake(GEN x)
       break;
 
     default: pari_err_TYPE("gnorml1_fake",x);
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   return gerepileupto(av, s);
 }
@@ -1273,7 +1319,7 @@ gsupnorm_aux(GEN x, GEN *m, GEN *msq, long prec)
     case t_MAT: lx = lg(x); break;
 
     default: pari_err_TYPE("gsupnorm",x);
-      return; /* not reached */
+      return; /* LCOV_EXCL_LINE */
   }
   for (i=1; i<lx; i++) gsupnorm_aux(gel(x,i), m, msq, prec);
 }
@@ -1371,7 +1417,7 @@ gtrace(GEN x)
 
 
     case t_RFRAC:
-      return gadd(x, gconj(x));
+      av = avma; return gerepileupto(av, gadd(x, conj_i(x)));
 
     case t_VEC: case t_COL:
       y = cgetg_copy(x, &lx);
@@ -1385,7 +1431,7 @@ gtrace(GEN x)
       av = avma; return gerepileupto(av, mattrace(x));
   }
   pari_err_TYPE("gtrace",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* Cholesky decomposition for positive definite matrix a
@@ -1679,20 +1725,22 @@ jacobi(GEN a, long prec)
 
 /*************************************************************************/
 /**                                                                     **/
-/**              MATRICE RATIONNELLE --> ENTIERE                        **/
+/**                   Q-vector space -> Z-modules                       **/
 /**                                                                     **/
 /*************************************************************************/
 
 GEN
 matrixqz0(GEN x,GEN p)
 {
-  if (typ(x) != t_MAT) pari_err_TYPE("QM_minors_coprime",x);
+  if (typ(x) != t_MAT) pari_err_TYPE("matrixqz",x);
   if (!p) return QM_minors_coprime(x,NULL);
-  if (typ(p) != t_INT) pari_err_TYPE("QM_minors_coprime",p);
+  if (typ(p) != t_INT) pari_err_TYPE("matrixqz",p);
   if (signe(p)>=0) return QM_minors_coprime(x,p);
+  if (!RgM_is_QM(x)) pari_err_TYPE("matrixqz", x);
   if (absequaliu(p,1)) return QM_ImZ_hnf(x); /* p = -1 */
   if (absequaliu(p,2)) return QM_ImQ_hnf(x); /* p = -2 */
-  pari_err_FLAG("QM_minors_coprime"); return NULL; /* not reached */
+  pari_err_FLAG("QM_minors_coprime");
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
@@ -1754,147 +1802,81 @@ QM_minors_coprime(GEN x, GEN D)
 }
 
 static GEN
-RgC_Z_mul(GEN A, GEN u)
+QM_ImZ_hnfall_i(GEN A, GEN *U, long remove)
 {
-  long s = signe(u);
-  if (is_pm1(u)) return s > 0? A: RgC_neg(A);
-  return s? gmul(u,A): zerocol(lg(A)-1);
-}
-
-/* u,v integral, A,B RgC */
-static GEN
-RgC_lincomb(GEN u, GEN v, GEN A, GEN B)
-{
-  if (!signe(u)) return RgC_Z_mul(B,v);
-  if (!signe(v)) return RgC_Z_mul(A,u);
-  return RgC_add(RgC_Z_mul(A,u), RgC_Z_mul(B,v));
-}
-
-/* cf ZC_elem */
-/* zero aj = Aij (!= 0)  using  ak = Aik (maybe 0), via linear combination of
- * A[j] and A[k] of determinant 1. */
-static void
-QC_elem(GEN aj, GEN ak, GEN A, long j, long k)
-{
-  GEN p1, u, v, d;
-
-  if (gequal0(ak)) { swap(gel(A,j), gel(A,k)); return; }
-  if (typ(aj) == t_INT) {
-    if (typ(ak) != t_INT) { aj = mulii(aj, gel(ak,2)); ak = gel(ak,1); }
-  } else {
-    if (typ(ak) == t_INT) { ak = mulii(ak, gel(aj,2)); aj = gel(aj,1); }
-    else {
-      GEN daj = gel(aj,2), dak = gel(ak,2), D = gcdii(daj, dak);
-      aj = gel(aj,1); ak = gel(ak,1);
-      if (!is_pm1(D)) { daj = diviiexact(daj, D); dak = diviiexact(dak, D); }
-      if (!is_pm1(dak)) aj = mulii(aj, dak);
-      if (!is_pm1(daj)) ak = mulii(ak, daj);
-    }
-  }
-  /* aj,ak were multiplied by their least common denominator */
-
-  d = bezout(aj,ak,&u,&v);
-  /* frequent special case (u,v) = (1,0) or (0,1) */
-  if (!signe(u))
-  { /* ak | aj */
-    GEN c = negi(diviiexact(aj,ak));
-    gel(A,j) = RgC_lincomb(gen_1, c, gel(A,j), gel(A,k));
-    return;
-  }
-  if (!signe(v))
-  { /* aj | ak */
-    GEN c = negi(diviiexact(ak,aj));
-    gel(A,k) = RgC_lincomb(gen_1, c, gel(A,k), gel(A,j));
-    swap(gel(A,j), gel(A,k));
-    return;
-  }
-
-  if (!is_pm1(d)) { aj = diviiexact(aj,d); ak = diviiexact(ak,d); }
-  p1 = gel(A,k);
-  gel(A,k) = RgC_lincomb(u,v, gel(A,j),p1);
-  gel(A,j) = RgC_lincomb(negi(aj),ak, p1,gel(A,j));
-}
-
-static GEN
-QM_imZ_hnf_aux(GEN A)
-{
-  pari_sp av = avma;
-  long i,j,k,n,m;
-
-  n = lg(A);
-  if (n == 1) return cgetg(1,t_MAT);
-  if (n == 2) {
-    GEN c;
-    A = Q_primitive_part(A, &c);
-    if (!c) A = ZM_copy(A); else if ( isintzero(c) ) A = cgetg(1,t_MAT);
-    return A;
-  }
-  m = lgcols(A);
-  for (i=1; i<m; i++)
+  GEN V = NULL, D;
+  A = Q_remove_denom(A,&D);
+  if (D)
   {
-    GEN b;
-    for (j = k = 1; j<n; j++)
+    long l, lA;
+    V = matkermod(A,D,NULL);
+    l = lg(V); lA = lg(A);
+    if (l == 1) V = scalarmat_shallow(D, lA-1);
+    else
     {
-      GEN a = gcoeff(A,i,j);
-      if (gequal0(a)) continue;
-
-      k = j+1; if (k == n) k = 1;
-      /* zero a = Aij  using  b = Aik */
-      QC_elem(a, gcoeff(A,i,k), A, j,k);
-    }
-    b = gcoeff(A,i,k);
-    if (!gequal0(b))
-    {
-      b = Q_denom(b);
-      if (!is_pm1(b)) gel(A,k) = RgC_Rg_mul(gel(A,k), b);
-    }
-    if (gc_needed(av,1))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"QM_imZ_hnf_aux");
-      A = gerepilecopy(av,A);
+      if (l < lA) V = hnfmodid(V,D);
+      A = ZM_Z_divexact(ZM_mul(A, V), D);
     }
   }
-  return ZM_hnf(A);
+  A = ZM_hnflll(A, U, remove);
+  if (U && V) *U = ZM_mul(V,*U);
+  return A;
 }
-
 GEN
-QM_ImZ_hnf(GEN x)
+QM_ImZ_hnfall(GEN x, GEN *U, long remove)
 {
   pari_sp av = avma;
-  return gerepileupto(av, QM_imZ_hnf_aux( RgM_shallowcopy(x) ));
+  x = QM_ImZ_hnfall_i(x, U, remove);
+  if (U) gerepileall(av, 2, &x, &U); else x = gerepileupto(av, x);
+  return x;
 }
+GEN
+QM_ImZ_hnf(GEN x) { return QM_ImZ_hnfall(x, NULL, 1); }
 
 GEN
-QM_ImQ_hnf(GEN x)
+QM_ImQ_hnfall(GEN x, GEN *U, long remove)
 {
   pari_sp av = avma, av1;
-  long j,j1,k,m,n;
-  GEN c;
+  long k, r, m, n = lg(x);
+  GEN c, V;
 
-  n = lg(x); if (n==1) return gcopy(x);
+  if (U) *U = matid(n-1);
+  if (n==1) return gcopy(x);
   m = lgcols(x); x = RgM_shallowcopy(x);
-  c = zero_zv(n-1);
+  c = zero_zv(n-1); r = 1;
   av1 = avma;
-  for (k=1; k<m; k++)
+  for (k = 1; k < m; k++)
   {
-    j=1; while (j<n && (c[j] || gequal0(gcoeff(x,k,j)))) j++;
+    long j = 1, a;
+    GEN p;
+    while (j<n && (c[j] || isintzero(gcoeff(x,k,j)))) j++;
     if (j==n) continue;
 
-    c[j]=k; gel(x,j) = RgC_Rg_div(gel(x,j),gcoeff(x,k,j));
-    for (j1=1; j1<n; j1++)
-      if (j1!=j)
+    c[j]=k; p = gcoeff(x,k,j);
+    gel(x,j) = RgC_Rg_div(gel(x,j), p);
+    if (U) gel(*U,j) = RgC_Rg_div(gel(*U,j), p);
+    for (a = 1; a < n; a++)
+      if (a != j)
       {
-        GEN t = gcoeff(x,k,j1);
-        if (!gequal0(t)) gel(x,j1) = RgC_sub(gel(x,j1), RgC_Rg_mul(gel(x,j),t));
+        GEN t = gcoeff(x,k,a);
+        if (gequal0(t)) continue;
+        if (U) gel(*U,a) = RgC_sub(gel(*U,a), RgC_Rg_mul(gel(*U,j),t));
+        gel(x,a) = RgC_sub(gel(x,a), RgC_Rg_mul(gel(x,j),t));
       }
+    if (++r == n) break;
     if (gc_needed(av1,1))
     {
       if(DEBUGMEM>1) pari_warn(warnmem,"QM_ImQ_hnf");
-      x = gerepilecopy(av1,x);
+      gerepileall(av1, U? 2: 1, &x, U);
     }
   }
-  return gerepileupto(av, QM_imZ_hnf_aux(x));
+  x = QM_ImZ_hnfall_i(x, U? &V: NULL, remove);
+  if (U) { *U = RgM_mul(*U,V); gerepileall(av,2,&x,U); }
+  else x = gerepileupto(av,x);
+  return x;
 }
+GEN
+QM_ImQ_hnf(GEN x) { return QM_ImQ_hnfall(x, NULL, 1); }
 
 GEN
 intersect(GEN x, GEN y)
