@@ -35,7 +35,7 @@ boundfact(GEN n, ulong lim)
     }
   }
   pari_err_TYPE("boundfact",n);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* NOT memory clean */
@@ -141,7 +141,7 @@ factor_eulerphi(GEN n)
   A = Z_factor(n); P = gel(A,1); E = gel(A,2); l = lg(P);
   for(i = 1; i < l; i++)
   {
-    GEN p = gel(P,i), q = subis(p,1), fa;
+    GEN p = gel(P,i), q = subiu(p,1), fa;
     long e = itos(gel(E,i)), w;
 
     w = vali(q); v += w; q = shifti(q,-w);
@@ -181,7 +181,7 @@ factor_eulerphi(GEN n)
 /**         CHECK FACTORIZATION FOR ARITHMETIC FUNCTIONS              **/
 /**                                                                   **/
 /***********************************************************************/
-static int
+int
 RgV_is_ZVpos(GEN v)
 {
   long i, l = lg(v);
@@ -193,7 +193,7 @@ RgV_is_ZVpos(GEN v)
   return 1;
 }
 /* check whether v is a ZV with non-0 entries */
-static int
+int
 RgV_is_ZVnon0(GEN v)
 {
   long i, l = lg(v);
@@ -401,44 +401,86 @@ divisors_init(GEN n, GEN *pP, GEN *pE)
   *pP = P; *pE = e; return isint;
 }
 
+static long
+ndiv(GEN E)
+{
+  long i, l;
+  GEN e = cgetg_copy(E, &l); /* left on stack */
+  ulong n;
+  for (i=1; i<l; i++) e[i] = E[i]+1;
+  n = itou_or_0( zv_prod_Z(e) );
+  if (!n || n & ~LGBITS) pari_err_OVERFLOW("divisors");
+  return n;
+}
+static int
+cmpi1(void *E, GEN a, GEN b) { (void)E; return cmpii(gel(a,1), gel(b,1)); }
+/* P a t_COL of objects, E a t_VECSMALL of exponents, return cleaned-up
+ * factorization (removing 0 exponents) as a t_MAT with 2 cols. */
+static GEN
+fa_clean(GEN P, GEN E)
+{
+  long i, j, l = lg(E);
+  GEN Q = cgetg(l, t_COL);
+  for (i = j = 1; i < l; i++)
+    if (E[i]) { gel(Q,j) = gel(P,i); E[j] = E[i]; j++; }
+  setlg(Q,j);
+  setlg(E,j); return mkmat2(Q,Flc_to_ZC(E));
+}
 GEN
-divisors(GEN n)
+divisors_factored(GEN N)
 {
   pari_sp av = avma;
-  long i, j, l;
-  ulong ndiv;
-  GEN *d, *t, *t1, *t2, *t3, P, E, e;
-  int isint = divisors_init(n, &P, &E);
+  GEN *d, *t1, *t2, *t3, D, P, E;
+  int isint = divisors_init(N, &P, &E);
+  GEN (*mul)(GEN,GEN) = isint? mulii: gmul;
+  long i, j, l, n = ndiv(E);
 
-  l = lg(E); e = cgetg(l, t_VECSMALL);
-  for (i=1; i<l; i++) e[i] = E[i]+1;
-  ndiv = itou_or_0( zv_prod_Z(e) );
-  if (!ndiv || ndiv & ~LGBITS) pari_err_OVERFLOW("divisors");
-  d = t = (GEN*) cgetg(ndiv+1,t_VEC);
+  D = cgetg(n+1,t_VEC); d = (GEN*)D;
+  l = lg(E);
+  *++d = mkvec2(gen_1, const_vecsmall(l-1,0));
+  for (i=1; i<l; i++)
+    for (t1=(GEN*)D,j=E[i]; j; j--,t1=t2)
+      for (t2=d, t3=t1; t3<t2; )
+      {
+        GEN a, b;
+        a = mul(gel(*++t3,1), gel(P,i));
+        b = leafcopy(gel(*t3,2)); b[i]++;
+        *++d = mkvec2(a,b);
+      }
+  if (isint) gen_sort_inplace(D,NULL,&cmpi1,NULL);
+  for (i = 1; i <= n; i++) gmael(D,i,2) = fa_clean(P, gmael(D,i,2));
+  return gerepilecopy(av, D);
+}
+GEN
+divisors(GEN N)
+{
+  long i, j, l;
+  GEN *d, *t1, *t2, *t3, D, P, E;
+  int isint = divisors_init(N, &P, &E);
+  GEN (*mul)(GEN,GEN) = isint? mulii: gmul;
+
+  D = cgetg(ndiv(E)+1,t_VEC); d = (GEN*)D;
+  l = lg(E);
   *++d = gen_1;
-  if (isint)
-  {
-    for (i=1; i<l; i++)
-      for (t1=t,j=E[i]; j; j--,t1=t2)
-        for (t2=d, t3=t1; t3<t2; ) *++d = mulii(*++t3, gel(P,i));
-    e = ZV_sort((GEN)t);
-  } else {
-    for (i=1; i<l; i++)
-      for (t1=t,j=E[i]; j; j--,t1=t2)
-        for (t2=d, t3=t1; t3<t2; ) *++d = gmul(*++t3, gel(P,i));
-    e = (GEN)t;
-  }
-  return gerepileupto(av, e);
+  for (i=1; i<l; i++)
+    for (t1=(GEN*)D,j=E[i]; j; j--,t1=t2)
+      for (t2=d, t3=t1; t3<t2; ) *++d = mul(*++t3, gel(P,i));
+  if (isint) ZV_sort_inplace(D);
+  return D;
+}
+GEN
+divisors0(GEN N, long flag)
+{
+  if (flag && flag != 1) pari_err_FLAG("divisors");
+  return flag? divisors_factored(N): divisors(N);
 }
 
 GEN
-divisorsu_fact(GEN P, GEN E)
+divisorsu_fact(GEN fa)
 {
+  GEN d, t, t1, t2, t3, P = gel(fa,1), E = gel(fa,2);
   long i, j, l = lg(P);
-  ulong nbdiv = 1;
-  GEN d, t, t1, t2, t3;
-  for (i=1; i<l; i++) nbdiv *= 1+E[i];
-  d = t = cgetg(nbdiv+1,t_VECSMALL);
+  d = t = cgetg(numdivu_fact(fa) + 1,t_VECSMALL);
   *++d = 1;
   for (i=1; i<l; i++)
     for (t1=t,j=E[i]; j; j--,t1=t2)
@@ -446,11 +488,10 @@ divisorsu_fact(GEN P, GEN E)
   vecsmall_sort(t); return t;
 }
 GEN
-divisorsu(ulong n)
+divisorsu(ulong N)
 {
   pari_sp av = avma;
-  GEN fa = factoru(n);
-  return gerepileupto(av, divisorsu_fact(gel(fa,1), gel(fa,2)));
+  return gerepileupto(av, divisorsu_fact(factoru(N)));
 }
 
 static GEN
@@ -678,6 +719,23 @@ eulerphi(GEN n)
   return gerepileuptoint(av, ZV_prod(Q));
 }
 
+long
+numdivu_fact(GEN fa)
+{
+  GEN E = gel(fa,2);
+  long n = 1, i, l = lg(E);
+  for (i = 1; i < l; i++) n *= E[i]+1;
+  return n;
+}
+long
+numdivu(long N)
+{
+  pari_sp av;
+  GEN fa;
+  if (N == 1) return 1;
+  av = avma; fa = factoru(N);
+  avma = av; return numdivu_fact(fa);
+}
 static GEN
 numdiv_aux(GEN F)
 {
@@ -692,19 +750,13 @@ numdiv(GEN n)
 {
   pari_sp av = avma;
   GEN F, E;
-  long i, l;
   if ((F = check_arith_non0(n,"numdiv")))
   {
     F = clean_Z_factor(F);
     E = numdiv_aux(F);
   }
   else if (lgefint(n) == 3)
-  {
-    if (n[2] == 1) return gen_1;
-    F = factoru(n[2]);
-    E = gel(F,2); l = lg(E);
-    for (i=1; i<l; i++) E[i]++;
-  }
+    return utoipos(numdivu(n[2]));
   else
     E = numdiv_aux(absZ_factor(n));
   return gerepileuptoint(av, zv_prod_Z(E));
@@ -715,7 +767,7 @@ static GEN
 u_euler_sumdiv(ulong p, long v)
 {
   GEN u = utoipos(1 + p); /* can't overflow */
-  for (; v > 1; v--) u = addsi(1, mului(p, u));
+  for (; v > 1; v--) u = addui(1, mului(p, u));
   return u;
 }
 /* 1 + q + ... + q^v */
@@ -844,6 +896,19 @@ Z_issquarefree(GEN n)
   }
   return moebius(n)? 1: 0;
 }
+
+static int
+fa_issquarefree(GEN F)
+{
+  GEN P = gel(F,1), E = gel(F,2);
+  long i, s, l = lg(P);
+  if (l == 1) return 1;
+  s = signe(gel(P,1)); /* = signe(x) */
+  if (!s) return 0;
+  for(i = 1; i < l; i++)
+    if (!equali1(gel(E,i))) return 0;
+  return 1;
+}
 long
 issquarefree(GEN x)
 {
@@ -856,8 +921,10 @@ issquarefree(GEN x)
       if (!signe(x)) return 0;
       av = avma; d = RgX_gcd(x, RgX_deriv(x));
       avma = av; return (lg(d) == 3);
+    case t_VEC:
+    case t_MAT: return fa_issquarefree(check_arith_all(x,"issquarefree"));
     default: pari_err_TYPE("issquarefree",x);
-      return 0; /* not reached */
+      return 0; /* LCOV_EXCL_LINE */
   }
 }
 
@@ -1018,13 +1085,12 @@ digits(GEN x, GEN B)
       avma = av; return binary_2k(x, k);
     }
   }
-  if (signe(x) < 0) x = absi(x);
+  x = absi_shallow(x);
   lz = logint(x,B) + 1;
   if (lgefint(B)>3)
   {
     z = gerepileupto(av, gen_digits_i(x, B, lz, NULL, &Z_ring, _dvmdii));
-    vecreverse_inplace(z);
-    return z;
+    vecreverse_inplace(z); return z;
   }
   else
   {
