@@ -32,11 +32,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
  * p is a rational prime, a belongs to Z_K, e=e(P/p), f=f(P/p), and b
  * is Lenstra's constant, such that p.P^(-1)= p Z_K + b Z_K.
  *
- * An extended ideal is a couple [I,F] where I is a valid ideal and F is
- * either an algebraic number, or a factorization matrix attached to an
- * algebraic number. All routines work with either extended ideals or ideals
- * (an omitted F is assumed to be [;] <-> 1).
- * All ideals are output in HNF form. */
+ * An extended ideal is a couple [I,F] where I is an ideal and F is either an
+ * algebraic number, or a factorization matrix attached to an algebraic number.
+ * All routines work with either extended ideals or ideals (an omitted F is
+ * assumed to be factor(1)). All ideals are output in HNF form. */
 
 /* types and conversions */
 
@@ -46,10 +45,18 @@ idealtyp(GEN *ideal, GEN *arch)
   GEN x = *ideal;
   long t,lx,tx = typ(x);
 
-  if (tx==t_VEC && lg(x)==3)
-  { *arch = gel(x,2); x = gel(x,1); tx = typ(x); }
+  if (tx!=t_VEC || lg(x)!=3) *arch = NULL;
   else
-    *arch = NULL;
+  {
+    GEN a = gel(x,2);
+    if (typ(a) == t_MAT && lg(a) != 3)
+    { /* allow [;] */
+      if (lg(a) != 1) pari_err_TYPE("idealtyp [extended ideal]",x);
+      a = trivial_fact();
+    }
+    *arch = a;
+    x = gel(x,1); tx = typ(x);
+  }
   switch(tx)
   {
     case t_MAT: lx = lg(x);
@@ -66,12 +73,12 @@ idealtyp(GEN *ideal, GEN *arch)
       t = id_PRINCIPAL; break;
     default:
       pari_err_TYPE("idealtyp",x);
-      return 0; /*not reached*/
+      return 0; /*LCOV_EXCL_LINE*/
   }
   *ideal = x; return t;
 }
 
-/* nf a true nf; v = [a,x,...], a in Z. Return (a,x) */
+/* true nf; v = [a,x,...], a in Z. Return (a,x) */
 GEN
 idealhnf_two(GEN nf, GEN v)
 {
@@ -79,11 +86,15 @@ idealhnf_two(GEN nf, GEN v)
   if (typ(m) == t_INT) return scalarmat(gcdii(m,p), nf_get_degree(nf));
   return ZM_hnfmodid(m, p);
 }
-
-static GEN
-ZM_Q_mul(GEN x, GEN y)
-{ return typ(y) == t_INT? ZM_Z_mul(x,y): RgM_Rg_mul(x,y); }
-
+/* true nf */
+GEN
+pr_hnf(GEN nf, GEN pr)
+{
+  GEN p = pr_get_p(pr), m;
+  if (pr_is_inert(pr)) return scalarmat(p, nf_get_degree(nf));
+  m = zk_scalar_or_multable(nf, pr_get_gen(pr));
+  return ZM_hnfmodprime(m, p);
+}
 
 GEN
 idealhnf_principal(GEN nf, GEN x)
@@ -94,7 +105,7 @@ idealhnf_principal(GEN nf, GEN x)
   {
     case t_COL: break;
     case t_INT:  if (!signe(x)) return cgetg(1,t_MAT);
-      return scalarmat(absi(x), nf_get_degree(nf));
+      return scalarmat(absi_shallow(x), nf_get_degree(nf));
     case t_FRAC:
       return scalarmat(Q_abs_shallow(x), nf_get_degree(nf));
     default: pari_err_TYPE("idealhnf",x);
@@ -116,6 +127,7 @@ vec_mulid(GEN nf, GEN x, long nx, long N)
     for (j=1; j<=N; j++) gel(m, k++) = zk_ei_mul(nf, gel(x,i),j);
   return m;
 }
+/* true nf */
 GEN
 idealhnf_shallow(GEN nf, GEN x)
 {
@@ -123,7 +135,7 @@ idealhnf_shallow(GEN nf, GEN x)
 
   /* cannot use idealtyp because here we allow non-square matrices */
   if (tx == t_VEC && lx == 3) { x = gel(x,1); tx = typ(x); lx = lg(x); }
-  if (tx == t_VEC && lx == 6) return idealhnf_two(nf,x); /* PRIME */
+  if (tx == t_VEC && lx == 6) return pr_hnf(nf,x); /* PRIME */
   switch(tx)
   {
     case t_MAT:
@@ -205,7 +217,7 @@ idealdiv0(GEN nf, GEN x, GEN y, long flag)
     case 1: return idealdivexact(nf,x,y);
     default: pari_err_FLAG("idealdiv");
   }
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
@@ -478,26 +490,30 @@ idealHNF_norm_pval(GEN x, GEN p, long Zval)
   return v;
 }
 
-/* return P, primes dividing Nx and xZ = x\cap Z, set v_p(Nx), v_p(xZ);
- * x integral in HNF */
+/* x integral in HNF, f0 = partial factorization of a multiple of
+ * x[1,1] = x\cap Z */
 GEN
-idealHNF_Z_factor(GEN x, GEN *pvN, GEN *pvZ)
+idealHNF_Z_factor_i(GEN x, GEN f0, GEN *pvN, GEN *pvZ)
 {
-  GEN xZ = gcoeff(x,1,1), f, P, E, vN, vZ;
+  GEN P, E, vN, vZ, xZ = gcoeff(x,1,1), f = f0? f0: Z_factor(xZ);
   long i, l;
-  if (typ(xZ) != t_INT) pari_err_TYPE("idealfactor",x);
-  f = Z_factor(xZ);
   P = gel(f,1); l = lg(P);
   E = gel(f,2);
   *pvN = vN = cgetg(l, t_VECSMALL);
   *pvZ = vZ = cgetg(l, t_VECSMALL);
   for (i = 1; i < l; i++)
   {
-    vZ[i] = itou(gel(E,i));
-    vN[i] = idealHNF_norm_pval(x,gel(P,i), vZ[i]);
+    GEN p = gel(P,i);
+    vZ[i] = f0? Z_pval(xZ, p): itou(gel(E,i));
+    vN[i] = idealHNF_norm_pval(x,p, vZ[i]);
   }
   return P;
 }
+/* return P, primes dividing Nx and xZ = x\cap Z, set v_p(Nx), v_p(xZ);
+ * x integral in HNF */
+GEN
+idealHNF_Z_factor(GEN x, GEN *pvN, GEN *pvZ)
+{ return idealHNF_Z_factor_i(x, NULL, pvN, pvZ); }
 
 /* v_P(A)*f(P) <= Nval [e.g. Nval = v_p(Norm A)], Zval = v_p(A \cap Z).
  * Return v_P(A) */
@@ -566,16 +582,16 @@ idealHNF_val(GEN A, GEN P, long Nval, long Zval)
   }
   return v;
 }
-/* true nf, x integral ideal */
+/* true nf, x != 0 integral ideal in HNF, cx t_INT or NULL,
+ * FA integer factorization matrix or NULL. Return partial factorization of
+ * cx * x above primes in FA (complete factorization if !FA)*/
 static GEN
-idealHNF_factor(GEN nf, GEN x)
+idealHNF_factor_i(GEN nf, GEN x, GEN cx, GEN FA)
 {
   const long N = lg(x)-1;
   long i, j, k, l, v;
-  GEN vp, vN, vZ, vP, vE, cx;
+  GEN vN, vZ, vP, vE, vp = idealHNF_Z_factor_i(x, FA, &vN,&vZ);
 
-  x = Q_primitive_part(x, &cx);
-  vp = idealHNF_Z_factor(x, &vN,&vZ);
   l = lg(vp);
   i = cx? expi(cx)+1: 1;
   vP = cgetg((l+i-2)*N+1, t_COL);
@@ -591,7 +607,7 @@ idealHNF_factor(GEN nf, GEN x)
     }
     else
       L = idealprimedec_limit_f(nf,p,Nval);
-    for (j = 1; j < lg(L); j++)
+    for (j = 1; Nval && j < lg(L); j++) /* !Nval => only cx contributes */
     {
       GEN P = gel(L,j);
       pari_sp av = avma;
@@ -601,17 +617,16 @@ idealHNF_factor(GEN nf, GEN x)
       v += vc * pr_get_e(P); if (!v) continue;
       gel(vP,k) = P;
       gel(vE,k) = utoipos(v); k++;
-      if (!Nval) break; /* now only the content contributes */
     }
-    if (vc) for (j++; j<lg(L); j++)
+    if (vc) for (; j<lg(L); j++)
     {
       GEN P = gel(L,j);
       gel(vP,k) = P;
       gel(vE,k) = utoipos(vc * pr_get_e(P)); k++;
     }
   }
-  if (cx)
-  {
+  if (cx && !FA)
+  { /* complete factorization */
     GEN f = Z_factor(cx), cP = gel(f,1), cE = gel(f,2);
     long lc = lg(cP);
     for (i=1; i<lc; i++)
@@ -629,6 +644,23 @@ idealHNF_factor(GEN nf, GEN x)
   setlg(vP, k);
   setlg(vE, k); return mkmat2(vP, vE);
 }
+/* true nf, x integral ideal */
+static GEN
+idealHNF_factor(GEN nf, GEN x, ulong lim)
+{
+  GEN cx, F = NULL;
+  if (lim)
+  {
+    GEN P, E;
+    long l;
+    F = Z_factor_limit(gcoeff(x,1,1), lim);
+    P = gel(F,1); l = lg(P);
+    E = gel(F,2);
+    if (l > 1 && abscmpiu(gel(P,l-1), lim) >= 0) { setlg(P,l-1); setlg(E,l-1); }
+  }
+  x = Q_primitive_part(x, &cx);
+  return idealHNF_factor_i(nf, x, cx, F);
+}
 /* c * vector(#L,i,L[i].e), assume results fit in ulong */
 static GEN
 prV_e_muls(GEN L, long c)
@@ -640,19 +672,31 @@ prV_e_muls(GEN L, long c)
 }
 /* true nf, y in Q */
 static GEN
-Q_nffactor(GEN nf, GEN y)
+Q_nffactor(GEN nf, GEN y, ulong lim)
 {
   GEN f, P, E;
-  long lfa, i;
+  long l, i;
   if (typ(y) == t_INT)
   {
     if (!signe(y)) pari_err_DOMAIN("idealfactor", "ideal", "=",gen_0,y);
     if (is_pm1(y)) return trivial_fact();
   }
-  f = factor(Q_abs_shallow(y));
-  P = gel(f,1); lfa = lg(P);
+  y = Q_abs_shallow(y);
+  if (!lim) f = Q_factor(y);
+  else
+  {
+    f = Q_factor_limit(y, lim);
+    P = gel(f,1); l = lg(P);
+    E = gel(f,2);
+    for (i = l-1; i > 0; i--)
+    {
+      if (abscmpiu(gel(P,i), lim) < 0) break;
+      setlg(P,i); setlg(E,i);
+    }
+  }
+  P = gel(f,1); l = lg(P); if (l == 1) return f;
   E = gel(f,2);
-  for (i = 1; i < lfa; i++)
+  for (i = 1; i < l; i++)
   {
     gel(P,i) = idealprimedec(nf, gel(P,i));
     gel(E,i) = prV_e_muls(gel(P,i), itos(gel(E,i)));
@@ -664,29 +708,157 @@ Q_nffactor(GEN nf, GEN y)
 }
 
 GEN
-idealfactor(GEN nf, GEN x)
+idealfactor_limit(GEN nf, GEN x, ulong lim)
 {
   pari_sp av = avma;
   GEN fa, y;
   long tx = idealtyp(&x,&y);
 
   nf = checknf(nf);
-  if (tx == id_PRIME) retmkmat2(mkcolcopy(x), mkcol(gen_1));
+  if (tx == id_PRIME)
+  {
+    if (lim && abscmpiu(pr_get_p(x), lim) >= 0) return trivial_fact();
+    retmkmat2(mkcolcopy(x), mkcol(gen_1));
+  }
   if (tx == id_PRINCIPAL)
   {
     y = nf_to_scalar_or_basis(nf, x);
-    if (typ(y) != t_COL) return gerepilecopy(av, Q_nffactor(nf, y));
+    if (typ(y) != t_COL) return gerepilecopy(av, Q_nffactor(nf, y, lim));
   }
   y = idealnumden(nf, x);
-  fa = idealHNF_factor(nf, gel(y,1));
+  fa = idealHNF_factor(nf, gel(y,1), lim);
   if (!isint1(gel(y,2)))
-  {
-    GEN F = idealHNF_factor(nf, gel(y,2));
-    fa = famat_mul_shallow(fa, famat_inv_shallow(F));
-  }
+    fa = famat_div_shallow(fa, idealHNF_factor(nf, gel(y,2), lim));
   fa = gerepilecopy(av, fa);
   return sort_factor(fa, (void*)&cmp_prime_ideal, &cmp_nodata);
 }
+GEN
+idealfactor(GEN nf, GEN x) { return idealfactor_limit(nf, x, 0); }
+GEN
+gpidealfactor(GEN nf, GEN x, GEN lim)
+{
+  ulong L = 0;
+  if (lim)
+  {
+    if (typ(lim) != t_INT || signe(lim) < 0) pari_err_FLAG("idealfactor");
+    L = itou(lim);
+  }
+  return idealfactor_limit(nf, x, L);
+}
+
+/* true nf; A is assumed to be the n-th power of an integral ideal,
+ * return its n-th root; n > 1 */
+static long
+idealsqrtn_int(GEN nf, GEN A, long n, GEN *pB)
+{
+  GEN C, ram, vram, root;
+  long i, l;
+
+  if (typ(A) == t_INT) return Z_ispowerall(A, n, pB);
+  /* compute valuations at ramified primes */
+  ram = gel(idealfactor(nf, idealadd(nf, nf_get_diff(nf),A)), 1);
+  l = lg(ram); vram = cgetg(l, t_VECSMALL);
+  for (i = 1; i < l; i++)
+  {
+    long v = idealval(nf,A,gel(ram,i));
+    if (v % n) return 0;
+    vram[i] = v / n;
+  }
+  root = idealfactorback(nf, ram, vram, 0);
+  /* remove ramified primes */
+  if (isint1(root))
+    root = matid(nf_get_degree(nf));
+  else
+    A = idealdivexact(nf, A, idealpows(nf,root,n));
+  A = Q_primitive_part(A, &C);
+  if (C)
+  {
+    if (!Z_ispowerall(C,n,&C)) return 0;
+    if (pB) root = ZM_Z_mul(root, C);
+  }
+
+  /* compute final n-th root, at most degree(nf)-1 iterations */
+  for (i = 0;; i++)
+  {
+    GEN J, b, a = gcoeff(A,1,1); /* A \cap Z */
+    if (is_pm1(a)) break;
+    if (!Z_ispowerall(a,n,&b)) return 0;
+    J = idealadd(nf, b, A);
+    A = idealdivexact(nf, idealpows(nf,J,n), A);
+    if (pB) root = odd(i)? idealdivexact(nf, root, J): idealmul(nf, root, J);
+  }
+  if (pB) *pB = root;
+  return 1;
+}
+
+/* A is assumed to be the n-th power of an ideal in nf
+ returns its n-th root. */
+long
+idealispower(GEN nf, GEN A, long n, GEN *pB)
+{
+  pari_sp av = avma;
+  GEN v, N, D;
+  nf = checknf(nf);
+  if (n <= 0) pari_err_DOMAIN("idealispower", "n", "<=", gen_0, stoi(n));
+  if (n == 1) { if (pB) *pB = idealhnf(nf,A); return 1; }
+  v = idealnumden(nf,A);
+  if (gequal0(gel(v,1))) { avma = av; if (pB) *pB = cgetg(1,t_MAT); return 1; }
+  if (!idealsqrtn_int(nf, gel(v,1), n, pB? &N: NULL)) return 0;
+  if (!idealsqrtn_int(nf, gel(v,2), n, pB? &D: NULL)) return 0;
+  if (pB) *pB = gerepileupto(av, idealdiv(nf,N,D)); else avma = av;
+  return 1;
+}
+
+/* x t_INT or integral non-0 ideal in HNF */
+static GEN
+idealredmodpower_i(GEN nf, GEN x, ulong k, ulong B)
+{
+  GEN cx, y, U, N, F, Q;
+  long nF;
+  if (typ(x) == t_INT)
+  {
+    if (!signe(x) || is_pm1(x)) return gen_1;
+    F = Z_factor_limit(x, B);
+    gel(F,2) = gdiventgs(gel(F,2), k);
+    return ginv(factorback(F));
+  }
+  N = gcoeff(x,1,1); if (is_pm1(N)) return gen_1;
+  F = Z_factor_limit(N, B); nF=lg(gel(F,1))-1;
+  if (BPSW_psp(gcoeff(F,nF,1))) U = NULL;
+  else
+  {
+    GEN M = powii(gcoeff(F,nF,1), gcoeff(F,nF,2));
+    y = hnfmodid(x, M); /* coprime part to B! */
+    if (!idealispower(nf, y, k, &U)) U = NULL;
+    x = hnfmodid(x, diviiexact(N, M));
+    setlg(gel(F,1), nF); /* remove last entry (unfactored part) */
+    setlg(gel(F,2), nF);
+  }
+  /* x = B-smooth part of initial x */
+  x = Q_primitive_part(x, &cx);
+  F = idealHNF_factor_i(nf, x, cx, F);
+  gel(F,2) = gdiventgs(gel(F,2), k);
+  Q = idealfactorback(nf, gel(F,1), gel(F,2), 0);
+  if (U) Q = idealmul(nf,Q,U);
+  if (typ(Q) == t_INT) return Q;
+  y = idealred_elt(nf, idealHNF_inv_Z(nf, Q));
+  return gdiv(y, gcoeff(Q,1,1));
+}
+GEN
+idealredmodpower(GEN nf, GEN x, ulong n, ulong B)
+{
+  pari_sp av = avma;
+  GEN a, b;
+  nf = checknf(nf);
+  if (!n) pari_err_DOMAIN("idealredmodpower","n", "=", gen_0, gen_0);
+  x = idealnumden(nf, x);
+  a = gel(x,1);
+  if (isintzero(a)) { avma = av; return gen_1; }
+  a = idealredmodpower_i(nf, gel(x,1), n, B);
+  b = idealredmodpower_i(nf, gel(x,2), n, B);
+  if (!isint1(b)) a = nf_to_scalar_or_basis(nf, nfdiv(nf, a, b));
+  return gerepilecopy(av, a);
+}
 
 /* P prime ideal in idealprimedec format. Return valuation(A) at P */
 long
@@ -698,7 +870,7 @@ idealval(GEN nf, GEN A, GEN P)
 
   if (tx == id_PRINCIPAL) return nfval(nf,A,P);
   checkprid(P);
-  if (tx == id_PRIME) return pr_equal(nf, P, A)? 1: 0;
+  if (tx == id_PRIME) return pr_equal(P, A)? 1: 0;
   /* id_MAT */
   nf = checknf(nf);
   A = Q_primitive_part(A, &cA);
@@ -757,25 +929,39 @@ idealadd(GEN nf, GEN x, GEN y)
 static GEN
 trivial_merge(GEN x)
 { return (lg(x) == 1 || !is_pm1(gcoeff(x,1,1)))? NULL: gen_1; }
-GEN
-idealaddtoone_i(GEN nf, GEN x, GEN y)
+/* true nf */
+static GEN
+_idealaddtoone(GEN nf, GEN x, GEN y, long red)
 {
   GEN a;
   long tx = idealtyp(&x, &a/*junk*/);
   long ty = idealtyp(&y, &a/*junk*/);
+  long ea;
   if (tx != id_MAT) x = idealhnf_shallow(nf, x);
   if (ty != id_MAT) y = idealhnf_shallow(nf, y);
   if (lg(x) == 1)
     a = trivial_merge(y);
   else if (lg(y) == 1)
     a = trivial_merge(x);
-  else {
+  else
     a = hnfmerge_get_1(x, y);
-    if (a && typ(a) == t_COL) a = ZC_reducemodlll(a, idealHNF_mul(nf,x,y));
-  }
   if (!a) pari_err_COPRIME("idealaddtoone",x,y);
+  if (red && (ea = gexpo(a)) > 10)
+  {
+    GEN b = (typ(a) == t_COL)? a: scalarcol_shallow(a, nf_get_degree(nf));
+    b = ZC_reducemodlll(b, idealHNF_mul(nf,x,y));
+    if (gexpo(b) < ea) a = b;
+  }
   return a;
 }
+/* true nf */
+GEN
+idealaddtoone_i(GEN nf, GEN x, GEN y)
+{ return _idealaddtoone(nf, x, y, 1); }
+/* true nf */
+GEN
+idealaddtoone_raw(GEN nf, GEN x, GEN y)
+{ return _idealaddtoone(nf, x, y, 0); }
 
 GEN
 idealaddtoone(GEN nf, GEN x, GEN y)
@@ -891,8 +1077,8 @@ famat_mul_shallow(GEN f, GEN g)
 {
   if (typ(f) != t_MAT) f = to_famat_shallow(f,gen_1);
   if (typ(g) != t_MAT) g = to_famat_shallow(g,gen_1);
-  if (lg(f) == 1) return g;
-  if (lg(g) == 1) return f;
+  if (lgcols(f) == 1) return g;
+  if (lgcols(g) == 1) return f;
   return mkmat2(shallowconcat(gel(f,1), gel(g,1)),
                 shallowconcat(gel(f,2), gel(g,2)));
 }
@@ -903,6 +1089,17 @@ famat_mulpow_shallow(GEN f, GEN g, GEN e)
   return famat_mul_shallow(f, famat_pow_shallow(g, e));
 }
 
+GEN
+famat_mulpows_shallow(GEN f, GEN g, long e)
+{
+  if (e==0) return f;
+  return famat_mul_shallow(f, famat_pows_shallow(g, e));
+}
+
+GEN
+famat_div_shallow(GEN f, GEN g)
+{ return famat_mul_shallow(f, famat_inv_shallow(g)); }
+
 GEN
 to_famat(GEN x, GEN y) { retmkmat2(mkcolcopy(x), mkcolcopy(y)); }
 GEN
@@ -922,7 +1119,7 @@ static GEN
 famat_add(GEN f, GEN x)
 {
   GEN h = cgetg(3,t_MAT);
-  if (lg(f) == 1)
+  if (lgcols(f) == 1)
   {
     gel(h,1) = mkcolcopy(x);
     gel(h,2) = mkcol(gen_1);
@@ -946,8 +1143,8 @@ famat_mul(GEN f, GEN g)
     gel(h,2) = mkcol2(gen_1, gen_1);
   }
   if (typ(f) != t_MAT) return famat_add(g, f);
-  if (lg(f) == 1) return gcopy(g);
-  if (lg(g) == 1) return gcopy(f);
+  if (lgcols(f) == 1) return gcopy(g);
+  if (lgcols(g) == 1) return gcopy(f);
   h = cgetg(3,t_MAT);
   gel(h,1) = gconcat(gel(f,1), gel(g,1));
   gel(h,2) = gconcat(gel(f,2), gel(g,2));
@@ -958,8 +1155,8 @@ GEN
 famat_sqr(GEN f)
 {
   GEN h;
-  if (lg(f) == 1) return cgetg(1,t_MAT);
   if (typ(f) != t_MAT) return to_famat(f,gen_2);
+  if (lgcols(f) == 1) return gcopy(f);
   h = cgetg(3,t_MAT);
   gel(h,1) = gcopy(gel(f,1));
   gel(h,2) = gmul2n(gel(f,2),1);
@@ -969,33 +1166,43 @@ famat_sqr(GEN f)
 GEN
 famat_inv_shallow(GEN f)
 {
-  if (lg(f) == 1) return f;
   if (typ(f) != t_MAT) return to_famat_shallow(f,gen_m1);
+  if (lgcols(f) == 1) return f;
   return mkmat2(gel(f,1), ZC_neg(gel(f,2)));
 }
 GEN
 famat_inv(GEN f)
 {
-  if (lg(f) == 1) return cgetg(1,t_MAT);
   if (typ(f) != t_MAT) return to_famat(f,gen_m1);
+  if (lgcols(f) == 1) return gcopy(f);
   retmkmat2(gcopy(gel(f,1)), ZC_neg(gel(f,2)));
 }
 GEN
 famat_pow(GEN f, GEN n)
 {
-  if (lg(f) == 1) return cgetg(1,t_MAT);
   if (typ(f) != t_MAT) return to_famat(f,n);
+  if (lgcols(f) == 1) return gcopy(f);
   retmkmat2(gcopy(gel(f,1)), ZC_Z_mul(gel(f,2),n));
 }
 GEN
 famat_pow_shallow(GEN f, GEN n)
 {
   if (is_pm1(n)) return signe(n) > 0? f: famat_inv_shallow(f);
-  if (lg(f) == 1) return f;
   if (typ(f) != t_MAT) return to_famat_shallow(f,n);
+  if (lgcols(f) == 1) return f;
   return mkmat2(gel(f,1), ZC_Z_mul(gel(f,2),n));
 }
 
+GEN
+famat_pows_shallow(GEN f, long n)
+{
+  if (n==1) return f;
+  if (n==-1) return famat_inv_shallow(f);
+  if (typ(f) != t_MAT) return to_famat_shallow(f, stoi(n));
+  if (lgcols(f) == 1) return f;
+  return mkmat2(gel(f,1), ZC_z_mul(gel(f,2),n));
+}
+
 GEN
 famat_Z_gcd(GEN M, GEN n)
 {
@@ -1039,8 +1246,7 @@ famat_to_nf(GEN nf, GEN f)
 {
   GEN t, x, e;
   long i;
-  if (lg(f) == 1) return gen_1;
-
+  if (lgcols(f) == 1) return gen_1;
   x = gel(f,1);
   e = gel(f,2);
   t = nfpow(nf, gel(x,1), gel(e,1));
@@ -1055,7 +1261,7 @@ famat_reduce(GEN fa)
   GEN E, G, L, g, e;
   long i, k, l;
 
-  if (lg(fa) == 1) return fa;
+  if (lgcols(fa) == 1) return fa;
   g = gel(fa,1); l = lg(g);
   e = gel(fa,2);
   L = gen_indexsort(g, (void*)&cmp_universal, &cmp_nodata);
@@ -1089,7 +1295,7 @@ famatsmall_reduce(GEN fa)
 {
   GEN E, G, L, g, e;
   long i, k, l;
-  if (lg(fa) == 1) return fa;
+  if (lgcols(fa) == 1) return fa;
   g = gel(fa,1); l = lg(g);
   e = gel(fa,2);
   L = vecsmall_indexsort(g);
@@ -1125,7 +1331,7 @@ ZM_famat_limit(GEN fa, GEN limit)
   GEN E, G, g, e, r;
   long i, k, l, n, lG;
 
-  if (lg(fa) == 1) return fa;
+  if (lgcols(fa) == 1) return fa;
   g = gel(fa,1); l = lg(g);
   e = gel(fa,2);
   for(n=0, i=1; i<l; i++)
@@ -1271,7 +1477,7 @@ famat_makecoprime(GEN nf, GEN g, GEN e, GEN pr, GEN prk, GEN EX)
     }
     else
     {
-      (void)ZC_nfvalrem(nf, x, pr, &x); /* x *= (b/p)^v_pr(x) */
+      (void)ZC_nfvalrem(x, pr, &x); /* x *= (b/p)^v_pr(x) */
       x = ZC_hnfrem(FpC_red(x,prkZ), prk);
       if (dx) x = FpC_Fp_mul(x, Fp_inv(dx,prkZ), prkZ);
     }
@@ -1293,13 +1499,11 @@ famat_makecoprime(GEN nf, GEN g, GEN e, GEN pr, GEN prk, GEN EX)
   return famat_to_nf_modideal_coprime(nf, G, E, prk, EX);
 }
 
-/* prod g[i]^e[i] mod bid, assume (g[i], id) = 1 */
+/* prod g[i]^e[i] mod bid, assume (g[i], id) = 1 and 1 < lg(g) <= lg(e) */
 GEN
 famat_to_nf_moddivisor(GEN nf, GEN g, GEN e, GEN bid)
 {
-  GEN t, cyc;
-  if (lg(g) == 1) return gen_1;
-  cyc = bid_get_cyc(bid);
+  GEN t, cyc = bid_get_cyc(bid);
   if (lg(cyc) == 1)
     t = gen_1;
   else
@@ -1310,45 +1514,29 @@ famat_to_nf_moddivisor(GEN nf, GEN g, GEN e, GEN bid)
 GEN
 vecmul(GEN x, GEN y)
 {
-  long i,lx, tx = typ(x);
-  GEN z;
-  if (is_scalar_t(tx)) return gmul(x,y);
-  z = cgetg_copy(x, &lx);
-  for (i=1; i<lx; i++) gel(z,i) = vecmul(gel(x,i), gel(y,i));
-  return z;
+  if (is_scalar_t(typ(x))) return gmul(x,y);
+  pari_APPLY_same(vecmul(gel(x,i), gel(y,i)))
 }
 
 GEN
 vecinv(GEN x)
 {
-  long i,lx, tx = typ(x);
-  GEN z;
-  if (is_scalar_t(tx)) return ginv(x);
-  z = cgetg_copy(x, &lx);
-  for (i=1; i<lx; i++) gel(z,i) = vecinv(gel(x,i));
-  return z;
+  if (is_scalar_t(typ(x))) return ginv(x);
+  pari_APPLY_same(vecinv(gel(x,i)))
 }
 
 GEN
 vecpow(GEN x, GEN n)
 {
-  long i,lx, tx = typ(x);
-  GEN z;
-  if (is_scalar_t(tx)) return powgi(x,n);
-  z = cgetg_copy(x, &lx);
-  for (i=1; i<lx; i++) gel(z,i) = vecpow(gel(x,i), n);
-  return z;
+  if (is_scalar_t(typ(x))) return powgi(x,n);
+  pari_APPLY_same(vecpow(gel(x,i), n))
 }
 
 GEN
 vecdiv(GEN x, GEN y)
 {
-  long i,lx, tx = typ(x);
-  GEN z;
-  if (is_scalar_t(tx)) return gdiv(x,y);
-  z = cgetg_copy(x, &lx);
-  for (i=1; i<lx; i++) gel(z,i) = vecdiv(gel(x,i), gel(y,i));
-  return z;
+  if (is_scalar_t(typ(x))) return gdiv(x,y);
+  pari_APPLY_same(vecdiv(gel(x,i), gel(y,i)))
 }
 
 /* A ideal as a square t_MAT */
@@ -1394,7 +1582,7 @@ idealmul_aux(GEN nf, GEN x, GEN y, long tx, long ty)
           return idealhnf_principal(nf, nfmul(nf,x,y));
         case id_PRIME:
         {
-          GEN p = gel(y,1), pi = gel(y,2), cx;
+          GEN p = pr_get_p(y), pi = pr_get_gen(y), cx;
           if (pr_is_inert(y)) return RgM_Rg_mul(idealhnf_principal(nf,x),p);
 
           x = nf_to_scalar_or_basis(nf, x);
@@ -1402,9 +1590,9 @@ idealmul_aux(GEN nf, GEN x, GEN y, long tx, long ty)
           {
             case t_INT:
               if (!signe(x)) return cgetg(1,t_MAT);
-              return ZM_Z_mul(idealhnf_two(nf,y), absi(x));
+              return ZM_Z_mul(pr_hnf(nf,y), absi_shallow(x));
             case t_FRAC:
-              return RgM_Rg_mul(idealhnf_two(nf,y), Q_abs_shallow(x));
+              return RgM_Rg_mul(pr_hnf(nf,y), Q_abs_shallow(x));
           }
           /* t_COL */
           x = Q_primitive_part(x, &cx);
@@ -1418,11 +1606,11 @@ idealmul_aux(GEN nf, GEN x, GEN y, long tx, long ty)
       }
     case id_PRIME:
       if (ty==id_PRIME)
-      { y = idealhnf_two(nf,y); cy = NULL; }
+      { y = pr_hnf(nf,y); cy = NULL; }
       else
         y = Q_primitive_part(y, &cy);
       y = idealHNF_mul_two(nf,y,x);
-      return cy? RgM_Rg_mul(y,cy): y;
+      return cy? ZM_Q_mul(y,cy): y;
 
     default: /* id_MAT */
     {
@@ -1504,7 +1692,7 @@ idealsqr_aux(GEN nf, GEN x, long tx)
       {
         x = ZM_hnfmodid(m, gcdii(sqri(a), zkmultable_capZ(m)));
         if (cx) cx = gsqr(cx);
-        if (cx) x = RgM_Rg_mul(x, cx);
+        if (cx) x = ZM_Q_mul(x, cx);
       }
       return x;
   }
@@ -1647,8 +1835,8 @@ idealnumden(GEN nf, GEN x)
       x = nf_to_scalar_or_basis(nf, x);
       switch(typ(x))
       {
-        case t_INT: return gerepilecopy(av, mkvec2(absi(x),gen_1));
-        case t_FRAC:return gerepilecopy(av, mkvec2(absi(gel(x,1)), gel(x,2)));
+        case t_INT: return gerepilecopy(av, mkvec2(absi_shallow(x),gen_1));
+        case t_FRAC:return gerepilecopy(av, mkvec2(absi_shallow(gel(x,1)), gel(x,2)));
       }
       /* t_COL */
       x = Q_remove_denom(x, &d);
@@ -1715,7 +1903,7 @@ idealpowprime(GEN nf, GEN pr, GEN n, GEN *pc)
     }
     else
     {
-      m = absi(m);
+      m = absi_shallow(m);
       if (signe(r)) m = addiu(m,1);
       q = powii(p,m); /* m = ceil(|n|/e) */
       if (signe(n) >= 0) gen = nfpow(nf, pr_get_gen(pr), n);
@@ -1748,7 +1936,8 @@ idealmulpowprime(GEN nf, GEN x, GEN pr, GEN n)
   if (pr_is_inert(pr))
   {
     GEN q = powii(pr_get_p(pr), n);
-    return typ(x) == t_MAT? RgM_Rg_mul(x,q): scalarmat_shallow(gmul(x,q), N);
+    return typ(x) == t_MAT? RgM_Rg_mul(x,q)
+                          : scalarmat_shallow(gmul(Q_abs(x),q), N);
   }
 
   y = idealpowprime(nf, pr, n, &c);
@@ -1761,7 +1950,7 @@ idealmulpowprime(GEN nf, GEN x, GEN pr, GEN n)
     x = idealHNF_mul_two(nf,x,y);
   else
     x = idealhnf_two(nf,y);
-  if (cx) x = RgM_Rg_mul(x,cx);
+  if (cx) x = ZM_Q_mul(x,cx);
   return x;
 }
 GEN
@@ -1785,7 +1974,7 @@ idealpow_aux(GEN nf, GEN x, long tx, GEN n)
       if (pr_is_inert(x)) return scalarmat(powii(gel(x,1), n), N);
       x = idealpowprime(nf, x, n, &cx);
       x = idealhnf_two(nf,x);
-      return cx? RgM_Rg_mul(x, cx): x;
+      return cx? ZM_Q_mul(x, cx): x;
     default:
       if (is_pm1(n)) return (s < 0)? idealinv(nf, x): gcopy(x);
       n1 = (s < 0)? negi(n): n;
@@ -1809,7 +1998,7 @@ idealpow_aux(GEN nf, GEN x, long tx, GEN n)
           cx = cx ? gdiv(cx, xZ): ginv(xZ);
           x = idealHNF_inv_Z(nf,x);
         }
-        if (cx) x = RgM_Rg_mul(x, cx);
+        if (cx) x = ZM_Q_mul(x, cx);
       }
       return x;
   }
@@ -1946,7 +2135,7 @@ GEN
 idealdivexact(GEN nf, GEN x0, GEN y0)
 {
   pari_sp av = avma;
-  GEN x, y, yZ, Nx, Ny, Nz, cy, q, r;
+  GEN x, y, xZ, yZ, Nx, Ny, Nz, cy, q, r;
 
   nf = checknf(nf);
   x = idealhnf_shallow(nf, x0);
@@ -1955,6 +2144,8 @@ idealdivexact(GEN nf, GEN x0, GEN y0)
   if (lg(x) == 1) { avma = av; return cgetg(1, t_MAT); } /* numerator is zero */
   y = Q_primitive_part(y, &cy);
   if (cy) x = RgM_Rg_div(x,cy);
+  xZ = gcoeff(x,1,1); if (typ(xZ) != t_INT) err_divexact(x,y);
+  yZ = gcoeff(y,1,1); if (isint1(yZ)) return gerepilecopy(av, x);
   Nx = idealnorm(nf,x);
   Ny = idealnorm(nf,y);
   if (typ(Nx) != t_INT) err_divexact(x,y);
@@ -1969,15 +2160,19 @@ idealdivexact(GEN nf, GEN x0, GEN y0)
     Nz = diviiexact(Nz,p1);
     q = mulii(q,p1);
   }
-  /* Replace x/y  by  x+(Nx/Nz) / y+(Ny/Nz) */
-  x = ZM_hnfmodid(x, q);
-  /* y reduced to unit ideal ? */
-  if (Nz == Ny) return gerepileupto(av, x);
+  xZ = gcoeff(x,1,1); q = gcdii(q, xZ);
+  if (!equalii(xZ,q))
+  { /* Replace x/y  by  x+(Nx/Nz) / y+(Ny/Nz) */
+    x = ZM_hnfmodid(x, q);
+    /* y reduced to unit ideal ? */
+    if (Nz == Ny) return gerepileupto(av, x);
 
-  y = ZM_hnfmodid(y, diviiexact(Ny,Nz));
+    yZ = gcoeff(y,1,1); q = gcdii(diviiexact(Ny,Nz), yZ);
+    y = ZM_hnfmodid(y, q);
+  }
   yZ = gcoeff(y,1,1);
   y = idealHNF_mul(nf,x, idealHNF_inv_Z(nf,y));
-  return gerepileupto(av, RgM_Rg_div(y, yZ));
+  return gerepileupto(av, ZM_Z_divexact(y, yZ));
 }
 
 GEN
@@ -2161,7 +2356,7 @@ idealmin(GEN nf, GEN x, GEN vdir)
   switch( idealtyp(&x,&y) )
   {
     case id_PRINCIPAL: return gcopy(x);
-    case id_PRIME: x = idealhnf_two(nf,x); break;
+    case id_PRIME: x = pr_hnf(nf,x); break;
     case id_MAT: if (lg(x) == 1) return gen_0;
   }
   x = Q_remove_denom(x, &dx);
@@ -2263,6 +2458,31 @@ Z_cba(GEN a, GEN b)
   if (!is_pm1(t)) vectrunc_append(L, t);
   return L;
 }
+/* P = coprime base, extend it by b; TODO: quadratic for now */
+GEN
+ZV_cba_extend(GEN P, GEN b)
+{
+  long i, l = lg(P);
+  GEN w = cgetg(l+1, t_VEC);
+  for (i = 1; i < l; i++)
+  {
+    GEN v = Z_cba(gel(P,i), b);
+    long nv = lg(v)-1;
+    gel(w,i) = vecslice(v, 1, nv-1); /* those divide P[i] but not b */
+    b = gel(v,nv);
+  }
+  gel(w,l) = b; return shallowconcat1(w);
+}
+GEN
+ZV_cba(GEN v)
+{
+  long i, l = lg(v);
+  GEN P;
+  if (l <= 2) return v;
+  P = Z_cba(gel(v,1), gel(v,2));
+  for (i = 3; i < l; i++) P = ZV_cba_extend(P, gel(v,i));
+  return P;
+}
 
 /* write x = x1 x2, x2 maximal s.t. (x2,f) = 1, return x2 */
 GEN
@@ -2357,7 +2577,7 @@ idealprodprime(GEN nf, GEN L)
   long l = lg(L), i;
   GEN z;
   if (l == 1) return matid(nf_get_degree(nf));
-  z = idealhnf_two(nf, gel(L,1));
+  z = pr_hnf(nf, gel(L,1));
   for (i=2; i<l; i++) z = idealHNF_mul_two(nf,z, gel(L,i));
   return z;
 }
@@ -2376,6 +2596,16 @@ idealprod(GEN nf, GEN I)
   return z;
 }
 
+/* v_pr(idealprod(nf,I)) */
+long
+idealprodval(GEN nf, GEN I, GEN pr)
+{
+  long i, l = lg(I), v = 0;
+  for (i = 1; i < l; i++)
+    if (!equali1(gel(I,i))) v += idealval(nf, gel(I,i), pr);
+  return v;
+}
+
 /* assume L is a list of prime ideals. Return prod L[i]^e[i] */
 GEN
 factorbackprime(GEN nf, GEN L, GEN e)
@@ -2430,7 +2660,7 @@ prV_lcm_capZ(GEN L)
 
 /* Given a prime ideal factorization with possibly zero or negative
  * exponents, gives b such that v_p(b) = v_p(x) for all prime ideals pr | x
- * and v_pr(b)> = 0 for all other pr.
+ * and v_pr(b) >= 0 for all other pr.
  * For optimal performance, all [anti-]uniformizers should be precomputed,
  * but no support for this yet.
  *
@@ -2462,7 +2692,7 @@ idealapprfact_i(GEN nf, GEN x, int nored)
   if (!z) return gen_1;
   if (nored || typ(z) != t_COL) return z;
   e2 = cgetg(r, t_VEC);
-  for (i=1; i<r; i++) gel(e2,i) = addis(gel(e,i), 1);
+  for (i=1; i<r; i++) gel(e2,i) = addiu(gel(e,i), 1);
   x = factorbackprime(nf, L,e2);
   if (flagden) /* denominator */
   {
@@ -2492,240 +2722,6 @@ idealappr(GEN nf, GEN x) {
 GEN
 idealappr0(GEN nf, GEN x, long fl) { (void)fl; return idealappr(nf, x); }
 
-/* merge a^e b^f. Assume a and b sorted. Keep 0 exponents and *append* new
- * entries from b [ result not sorted ] */
-static void
-merge_fact(GEN *pa, GEN *pe, GEN b, GEN f)
-{
-  GEN A, E, a = *pa, e = *pe;
-  long k, i, la = lg(a), lb = lg(b), l = la+lb-1;
-
-  *pa = A = cgetg(l, t_COL);
-  *pe = E = cgetg(l, t_COL);
-  k = 1;
-  for (i=1; i<la; i++)
-  {
-    gel(A,i) = gel(a,i);
-    gel(E,i) = gel(e,i);
-    if (k < lb && gequal(gel(A,i), gel(b,k)))
-    {
-      gel(E,i) = addii(gel(E,i), gel(f,k));
-      k++;
-    }
-  }
-  for (; k < lb; i++,k++)
-  {
-    gel(A,i) = gel(b,k);
-    gel(E,i) = gel(f,k);
-  }
-  setlg(A, i);
-  setlg(E, i);
-}
-
-static int
-isprfact(GEN x)
-{
-  long i, l;
-  GEN L, E;
-  if (typ(x) != t_MAT || lg(x) != 3) return 0;
-  L = gel(x,1); l = lg(L);
-  E = gel(x,2);
-  for(i=1; i<l; i++)
-  {
-    checkprid(gel(L,i));
-    if (typ(gel(E,i)) != t_INT) return 0;
-  }
-  return 1;
-}
-
-/* initialize projectors mod pr[i]^e[i] for idealchinese */
-static GEN
-pr_init(GEN nf, GEN fa, GEN w, GEN dw)
-{
-  GEN L = gel(fa,1), E = gel(fa,2);
-  long r = lg(L);
-
-  if (w && lg(w) != r) pari_err_TYPE("idealchinese", w);
-  if (r > 1)
-  {
-    GEN U, F;
-    long i;
-    if (dw)
-    {
-      GEN p = gen_indexsort(L, (void*)&cmp_prime_ideal, cmp_nodata);
-      GEN fw = idealfactor(nf, dw); /* sorted */
-      L = vecpermute(L, p);
-      E = vecpermute(E, p);
-      w = vecpermute(w, p);
-      merge_fact(&L, &E, gel(fw,1), gel(fw,2));
-      /* L and E lenghtened, with factors of dw coming last */
-    }
-    else
-      E = leafcopy(E); /* do not destroy fa[2] */
-
-    for (i=1; i<r; i++)
-      if (signe(gel(E,i)) < 0) gel(E,i) = gen_0;
-    F = factorbackprime(nf, L, E);
-    U = cgetg(r, t_VEC);
-    for (i = 1; i < r; i++)
-    {
-      GEN u;
-      if (w && gequal0(gel(w,i))) u = gen_0; /* unused */
-      else
-      {
-        GEN pr = gel(L,i), e = gel(E,i), t;
-        t = idealdivpowprime(nf,F, pr, e);
-        u = hnfmerge_get_1(t, idealpow(nf, pr, e));
-        if (!u) pari_err_COPRIME("idealchinese", t,pr);
-      }
-      gel(U,i) = u;
-    }
-    F = idealpseudored(F, nf_get_roundG(nf));
-    fa = mkvec2(F, U);
-  }
-  else
-    fa = cgetg(1,t_VEC);
-  return fa;
-}
-
-static GEN
-pl_normalize(GEN nf, GEN pl)
-{
-  const char *fun = "idealchinese";
-  if (lg(pl)-1 != nf_get_r1(nf)) pari_err_TYPE(fun,pl);
-  switch(typ(pl))
-  {
-    case t_VEC:
-      RgV_check_ZV(pl,fun);
-      pl = ZV_to_zv(pl);
-      /* fall through */
-    case t_VECSMALL: break;
-    default: pari_err_TYPE(fun,pl);
-  }
-  return pl;
-}
-
-static int
-is_chineseinit(GEN x)
-{
-  GEN fa, pl;
-  long l;
-  if (typ(x) != t_VEC || lg(x)!=3) return 0;
-  fa = gel(x,1);
-  pl = gel(x,2);
-  if (typ(fa) != t_VEC || typ(pl) != t_VEC) return 0;
-  l = lg(fa);
-  if (l != 1)
-  {
-    if (l != 3 || typ(gel(fa,1)) != t_MAT || typ(gel(fa,2)) != t_VEC)
-      return 0;
-  }
-  l = lg(pl);
-  if (l != 1)
-  {
-    if (l != 5 || typ(gel(pl,1)) != t_MAT || typ(gel(pl,2)) != t_MAT
-               || typ(gel(pl,3)) != t_COL || typ(gel(pl,4)) != t_VECSMALL)
-      return 0;
-  }
-  return 1;
-}
-
-/* nf a true 'nf' */
-static GEN
-chineseinit_i(GEN nf, GEN fa, GEN w, GEN dw)
-{
-  const char *fun = "idealchineseinit";
-  GEN nz = NULL, pl = NULL;
-  switch(typ(fa))
-  {
-    case t_VEC:
-      if (is_chineseinit(fa))
-      {
-        if (dw) pari_err_DOMAIN(fun, "denom(y)", "!=", gen_1, w);
-        return fa;
-      }
-      if (lg(fa) != 3) pari_err_TYPE(fun, fa);
-      /* of the form [x,s] */
-      pl = pl_normalize(nf, gel(fa,2));
-      fa = gel(fa,1);
-      nz = vecsmall01_to_indices(pl);
-      if (is_chineseinit(fa)) { fa = gel(fa,1); break; /* keep fa, reset pl */ }
-      /* fall through */
-    case t_MAT: /* factorization? */
-      if (isprfact(fa)) { fa = pr_init(nf, fa, w, dw); break; }
-    default: pari_err_TYPE(fun,fa);
-  }
-
-  if (pl)
-  {
-    GEN C, Mr, MI, lambda, mlambda;
-    GEN F = (lg(fa) == 1)? NULL: gel(fa,1);
-    long i, r;
-    Mr = rowpermute(nf_get_M(nf), nz);
-    MI = F? RgM_mul(Mr, F): Mr;
-    lambda = gmul2n(matrixnorm(MI,DEFAULTPREC), -1);
-    mlambda = gneg(lambda);
-    r = lg(nz);
-    C = cgetg(r, t_COL);
-    for (i = 1; i < r; i++) gel(C,i) = pl[nz[i]] < 0? mlambda: lambda;
-    pl = mkvec4(MI, Mr, C, pl);
-  }
-  else
-    pl = cgetg(1,t_VEC);
-  return mkvec2(fa, pl);
-}
-
-/* Given a prime ideal factorization x, possibly with 0 or negative exponents,
- * and a vector w of elements of nf, gives b such that
- * v_p(b-w_p)>=v_p(x) for all prime ideals p in the ideal factorization
- * and v_p(b)>=0 for all other p, using the standard proof given in GTM 138. */
-GEN
-idealchinese(GEN nf, GEN x, GEN w)
-{
-  const char *fun = "idealchinese";
-  pari_sp av = avma;
-  GEN x1, x2, s, dw;
-
-  nf = checknf(nf);
-  if (!w) return gerepilecopy(av, chineseinit_i(nf,x,NULL,NULL));
-
-  if (typ(w) != t_VEC) pari_err_TYPE(fun,w);
-  w = Q_remove_denom(matalgtobasis(nf,w), &dw);
-  if (!is_chineseinit(x)) x = chineseinit_i(nf,x,w,dw);
-  /* x is a 'chineseinit' */
-  x1 = gel(x,1); s = NULL;
-  if (lg(x1) != 1)
-  {
-    GEN F = gel(x1,1), U = gel(x1,2);
-    long i, r = lg(w);
-    for (i=1; i<r; i++)
-      if (!gequal0(gel(w,i)))
-      {
-        GEN t = nfmuli(nf, gel(U,i), gel(w,i));
-        s = s? ZC_add(s,t): t;
-      }
-    if (s) s = ZC_reducemodmatrix(s, F);
-  }
-  if (!s) { s = zerocol(nf_get_degree(nf)); dw = NULL; }
-
-  x2 = gel(x,2);
-  if (lg(x2) != 1)
-  {
-    GEN pl = gel(x2,4);
-    if (!nfchecksigns(nf, s, pl))
-    {
-      GEN MI = gel(x2,1), Mr = gel(x2,2), C = gel(x2,3);
-      GEN t = RgC_sub(C, RgM_RgC_mul(Mr,s));
-      long e;
-      t = grndtoi(RgM_RgC_invimage(MI,t), &e);
-      if (lg(x1) != 1) { GEN F = gel(x1,1); t = ZM_ZC_mul(F, t); }
-      s = ZC_add(s, t);
-    }
-  }
-  if (dw) s = RgC_Rg_div(s,dw);
-  return gerepileupto(av, s);
-}
-
 static GEN
 mat_ideal_two_elt2(GEN nf, GEN x, GEN a)
 {
