@@ -156,7 +156,7 @@ polchebyshev(long n, long kind, long v)
     case 2: return polchebyshev2(n, v);
     default: pari_err_FLAG("polchebyshev");
   }
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 GEN
 polchebyshev_eval(long n, long kind, GEN x)
@@ -169,7 +169,7 @@ polchebyshev_eval(long n, long kind, GEN x)
     case 2: return polchebyshev2_eval(n, x);
     default: pari_err_FLAG("polchebyshev");
   }
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* Hermite polynomial H(n,x):  H(n+1) = 2x H(n) - 2n H(n-1)
@@ -489,35 +489,14 @@ matqpascal(long n, GEN q)
 /******************************************************************/
 
 GEN
-gprec(GEN x, long l)
+gprec(GEN x, long d)
 {
-  long lx, i;
-  GEN y;
-
-  if (l <= 0) pari_err_DOMAIN("gprec", "precision", "<=", gen_0, stoi(l));
-  switch(typ(x))
-  {
-    case t_REAL:
-      return rtor(x, ndec2prec(l));
-    case t_COMPLEX:
-      y = cgetg(3, t_COMPLEX);
-      gel(y,1) = gprec(gel(x,1),l);
-      gel(y,2) = gprec(gel(x,2),l);
-      break;
-    case t_POL: case t_SER:
-      y = cgetg_copy(x, &lx); y[1] = x[1];
-      for (i=2; i<lx; i++) gel(y,i) = gprec(gel(x,i),l);
-      break;
-    case t_POLMOD: case t_RFRAC: case t_VEC: case t_COL: case t_MAT:
-      y = cgetg_copy(x, &lx);
-      for (i=1; i<lx; i++) gel(y,i) = gprec(gel(x,i),l);
-      break;
-    default: y = gcopy(x);
-  }
-  return y;
+  pari_sp av = avma;
+  if (d <= 0) pari_err_DOMAIN("gprec", "precision", "<=", gen_0, stoi(d));
+  return gerepilecopy(av, gprec_w(x, ndec2prec(d)));
 }
 
-/* internal: precision given in word length (including codewords) */
+/* not GC-safe; precision given in word length (including codewords) */
 GEN
 gprec_w(GEN x, long pr)
 {
@@ -527,7 +506,7 @@ gprec_w(GEN x, long pr)
   switch(typ(x))
   {
     case t_REAL:
-      if (signe(x)) return rtor(x,pr);
+      if (signe(x)) return realprec(x) != pr? rtor(x,pr): x;
       i = -prec2nbits(pr);
       if (i < expo(x)) return real_0_bit(i);
       y = cgetr(2); y[1] = x[1]; return y;
@@ -548,9 +527,40 @@ gprec_w(GEN x, long pr)
   }
   return y;
 }
+/* not GC-safe; precision given in word length (including codewords) */
+GEN
+gprec_wensure(GEN x, long pr)
+{
+  long lx, i;
+  GEN y;
 
-/* internal: precision given in word length (including codewords), truncate
- * mantissa to precision 'pr' but never _increase_ it */
+  switch(typ(x))
+  {
+    case t_REAL:
+      if (signe(x)) return realprec(x) < pr? rtor(x,pr): x;
+      i = -prec2nbits(pr);
+      if (i < expo(x)) return real_0_bit(i);
+      y = cgetr(2); y[1] = x[1]; return y;
+    case t_COMPLEX:
+      y = cgetg(3, t_COMPLEX);
+      gel(y,1) = gprec_wensure(gel(x,1),pr);
+      gel(y,2) = gprec_wensure(gel(x,2),pr);
+      break;
+   case t_POL: case t_SER:
+      y = cgetg_copy(x, &lx); y[1] = x[1];
+      for (i=2; i<lx; i++) gel(y,i) = gprec_wensure(gel(x,i),pr);
+      break;
+    case t_POLMOD: case t_RFRAC: case t_VEC: case t_COL: case t_MAT:
+      y = cgetg_copy(x, &lx);
+      for (i=1; i<lx; i++) gel(y,i) = gprec_wensure(gel(x,i),pr);
+      break;
+    default: return x;
+  }
+  return y;
+}
+
+/* not GC-safe; precision given in word length (including codewords),
+ * truncate mantissa to precision 'pr' but never increase it */
 GEN
 gprec_wtrunc(GEN x, long pr)
 {
@@ -642,7 +652,7 @@ convol(GEN x, GEN y)
   ex = valp(x);
   ey = valp(y);
   if (ser_isexactzero(x))
-    return scalarser(gadd(RgX_get_0(x), RgX_get_0(y)), varn(x), maxss(ex,ey));
+    return scalarser(gadd(Rg_get_0(x), Rg_get_0(y)), varn(x), maxss(ex,ey));
   lx = lg(x) + ex; x -= ex;
   ly = lg(y) + ey; y -= ey;
   /* inputs shifted: x[i] and y[i] now correspond to monomials of same degree */
@@ -799,7 +809,7 @@ binomial(GEN n, long k)
   {
     if (signe(n) > 0)
     {
-      GEN z = subis(n,k);
+      GEN z = subiu(n,k);
       if (cmpis(z,k) < 0)
       {
         k = itos(z); avma = av;
@@ -820,7 +830,7 @@ binomial(GEN n, long k)
     else
     {
       y = cgetg(k+1,t_VEC);
-      for (i=1; i<=k; i++) gel(y,i) = subis(n,i-1);
+      for (i=1; i<=k; i++) gel(y,i) = subiu(n,i-1);
       y = ZV_prod(y);
     }
     y = diviiexact(y, mpfact(k));
@@ -838,9 +848,21 @@ binomial(GEN n, long k)
   return gerepileupto(av, gdiv(RgV_prod(y), mpfact(k)));
 }
 
+GEN
+binomial0(GEN x, GEN k)
+{
+  if (!k)
+  {
+    if (typ(x) != t_INT || signe(x) < 0) pari_err_TYPE("binomial", x);
+    return vecbinomial(itos(x));
+  }
+  if (typ(k) != t_INT) pari_err_TYPE("binomial", k);
+  return binomial(x, itos(k));
+}
+
 /* Assume n >= 0, return bin, bin[k+1] = binomial(n, k) */
 GEN
-vecbinome(long n)
+vecbinomial(long n)
 {
   long d, k;
   GEN C;
@@ -938,83 +960,7 @@ stirling(long n, long m, long flag)
     case 2: return stirling2((ulong)n,(ulong)m);
     default: pari_err_FLAG("stirling");
   }
-  return NULL; /*NOT REACHED*/
-}
-
-/***********************************************************************/
-/**                          PERMUTATIONS                             **/
-/***********************************************************************/
-GEN
-Z_to_perm(long n, GEN x)
-{
-  pari_sp av;
-  ulong i, r;
-  GEN v = cgetg(n+1, t_VECSMALL);
-  if (n==0) return v;
-  uel(v,n) = 1; av = avma;
-  if (signe(x) <= 0) x = modii(x, mpfact(n));
-  for (r=n-1; r>=1; r--)
-  {
-    ulong a;
-    x = diviu_rem(x, n+1-r, &a);
-    for (i=r+1; i<=(ulong)n; i++)
-      if (uel(v,i) > a) uel(v,i)++;
-    uel(v,r) = a+1;
-  }
-  avma = av; return v;
-}
-GEN
-numtoperm(long n, GEN x)
-{
-  long i;
-  GEN v;
-
-  if (n < 0) pari_err_DOMAIN("numtoperm", "n", "<", gen_0, stoi(n));
-  if (typ(x) != t_INT) pari_err_TYPE("numtoperm",x);
-  v = Z_to_perm(n, x); settyp(v, t_VEC);
-  for (i = 1; i <= n; i++) gel(v,i) = utoipos(uel(v,i));
-  return v;
-}
-
-/* destroys v */
-static GEN
-perm_to_Z_inplace(GEN v)
-{
-  long l = lg(v), i, r;
-  GEN x = gen_0;
-  for (i = 1; i < l; i++)
-  {
-    long vi = v[i];
-    if (vi <= 0) return NULL;
-    x = i==1 ? utoi(vi-1): addiu(muliu(x,l-i), vi-1);
-    for (r = i+1; r < l; r++)
-      if (v[r] > vi) v[r]--;
-  }
-  return x;
-}
-GEN
-perm_to_Z(GEN v)
-{
-  pari_sp av = avma;
-  GEN x = perm_to_Z_inplace(leafcopy(v));
-  if (!x) pari_err_TYPE("permtonum",v);
-  return gerepileuptoint(av, x);
-}
-GEN
-permtonum(GEN p)
-{
-  pari_sp av = avma;
-  GEN v, x;
-  switch(typ(p))
-  {
-    case t_VECSMALL: return perm_to_Z(p);
-    case t_VEC: case t_COL:
-      if (RgV_is_ZV(p)) { v = ZV_to_zv(p); break; }
-    default: pari_err_TYPE("permtonum",p); return NULL;
-  }
-  x = perm_to_Z_inplace(v);
-  if (!x) pari_err_TYPE("permtonum",p);
-  return gerepileuptoint(av, x);
+  return NULL; /*LCOV_EXCL_LINE*/
 }
 
 /*******************************************************************/
@@ -1026,7 +972,9 @@ permtonum(GEN p)
 GEN
 polrecip(GEN x)
 {
-  if (typ(x) != t_POL) pari_err_TYPE("polrecip",x);
+  long tx = typ(x);
+  if (is_scalar_t(tx)) return gcopy(x);
+  if (tx != t_POL) pari_err_TYPE("polrecip",x);
   return RgX_recip(x);
 }
 
@@ -1076,7 +1024,7 @@ polint_i(GEN X, GEN Y, GEN x, long n, GEN *ptdy)
   if (n == 1)
   {
     if (ptdy) *ptdy = gen_0;
-    return gmul(gel(Y,0), RgX_get_1(x));
+    return gmul(gel(Y,0), Rg_get_1(x));
   }
   if (!X)
   {
@@ -1171,7 +1119,7 @@ polint(GEN X, GEN Y, GEN t, GEN *ptdy)
     return gerepileupto(av, P);
   }
   /* numerical interpolation */
-  if (lx == 1) return RgX_get_0(t);
+  if (lx == 1) return Rg_get_0(t);
   return polint_i(X? X+1: NULL,Y+1,t,lx-1,ptdy);
 }
 
@@ -1219,7 +1167,7 @@ QXQ_reverse(GEN a, GEN T)
   if (typ(a) != t_POL || !signe(a)) err_reverse(a,T);
   if (gequalX(a)) return gcopy(a);
   y = RgXV_to_RgM(QXQ_powers(a,n-1,T), n);
-  y = RgM_solve(y, col_ei(n, 2));
+  y = QM_gauss(y, col_ei(n, 2));
   if (!y) err_reverse(a,T);
   return gerepilecopy(av, RgV_to_RgX(y, varn(T)));
 }
@@ -1228,15 +1176,14 @@ GEN
 modreverse(GEN x)
 {
   long v, n;
-  GEN T, a, y;
+  GEN T, a;
 
   if (typ(x)!=t_POLMOD) pari_err_TYPE("modreverse",x);
   T = gel(x,1); n = degpol(T); if (n <= 0) return gcopy(x);
   a = gel(x,2);
   v = varn(T);
-  y = cgetg(3,t_POLMOD);
-  gel(y,1) = (n==1)? gsub(pol_x(v), a): RgXQ_charpoly(a, T, v);
-  gel(y,2) = RgXQ_reverse(a, T); return y;
+  retmkpolmod(RgXQ_reverse(a, T),
+              (n==1)? gsub(pol_x(v), a): RgXQ_charpoly(a, T, v));
 }
 
 /********************************************************************/
@@ -1412,6 +1359,8 @@ sort_extract(GEN x, GEN y, long tx, long lx)
   return y;
 }
 
+static GEN
+triv_sort(long tx) { return tx == t_LIST? mklist(): cgetg(1, tx); }
 /* Sort the vector x, using cmp to compare entries. */
 GEN
 gen_sort_uniq(GEN x, void *E, int (*cmp)(void*,GEN,GEN))
@@ -1420,7 +1369,7 @@ gen_sort_uniq(GEN x, void *E, int (*cmp)(void*,GEN,GEN))
   GEN y;
 
   init_sort(&x, &tx, &lx);
-  if (lx==1) return tx == t_LIST? mklist(): cgetg(1, tx);
+  if (lx==1) return triv_sort(tx);
   y = gen_sortspec_uniq(x,lx-1,E,cmp);
   return sort_extract(x, y, tx, lg(y)); /* lg(y) <= lx */
 }
@@ -1432,7 +1381,7 @@ gen_sort(GEN x, void *E, int (*cmp)(void*,GEN,GEN))
   GEN y;
 
   init_sort(&x, &tx, &lx);
-  if (lx==1) return tx == t_LIST? mklist(): cgetg(1, tx);
+  if (lx==1) return triv_sort(tx);
   y = gen_sortspec(x,lx-1,E,cmp);
   return sort_extract(x, y, tx, lx);
 }
@@ -1488,12 +1437,9 @@ static int
 closurecmp(void *data, GEN x, GEN y)
 {
   pari_sp av = avma;
-  GEN z = closure_callgen2((GEN)data, x,y);
-  if (typ(z) != t_INT)
-    pari_err_TYPE("closurecmp, cmp. fun. must return an integer", z);
-  avma = av; return signe(z);
+  long s = gsigne(closure_callgen2((GEN)data, x,y));
+  avma = av; return s;
 }
-
 static void
 check_positive_entries(GEN k)
 {
@@ -1503,31 +1449,37 @@ check_positive_entries(GEN k)
 }
 
 typedef int (*CMP_FUN)(void*,GEN,GEN);
+/* return NULL if t_CLOSURE k is a "key" (arity 1) and not a sorting func */
 static CMP_FUN
 sort_function(void **E, GEN x, GEN k)
 {
   int (*cmp)(GEN,GEN) = &lexcmp;
+  long tx = typ(x);
   if (!k)
   {
     *E = (void*)((typ(x) == t_VECSMALL)? cmp_small: cmp);
     return &cmp_nodata;
   }
-  if (typ(x) == t_VECSMALL) pari_err_TYPE("sort_function", x);
+  if (tx == t_VECSMALL) pari_err_TYPE("sort_function", x);
   switch(typ(k))
   {
     case t_INT: k = mkvecsmall(itos(k));  break;
     case t_VEC: case t_COL: k = ZV_to_zv(k); break;
     case t_VECSMALL: break;
     case t_CLOSURE:
-     if (closure_arity(k) != 2 || closure_is_variadic(k))
-       pari_err_TYPE("sort_function, cmp. fun. needs exactly 2 arguments",k);
+     if (closure_is_variadic(k))
+       pari_err_TYPE("sort_function, variadic cmpf",k);
      *E = (void*)k;
-     return &closurecmp;
+     switch(closure_arity(k))
+     {
+       case 1: return NULL; /* wrt key */
+       case 2: return &closurecmp;
+       default: pari_err_TYPE("sort_function, cmpf arity != 1, 2",k);
+     }
     default: pari_err_TYPE("sort_function",k);
   }
   check_positive_entries(k);
-  *E = (void*)k;
-  return &veccmp;
+  *E = (void*)k; return &veccmp;
 }
 
 #define cmp_IND 1
@@ -1542,22 +1494,28 @@ vecsort0(GEN x, GEN k, long flag)
 
   if (flag < 0 || flag > (cmp_REV|cmp_LEX|cmp_IND|cmp_UNIQ))
     pari_err_FLAG("vecsort");
-  if (flag & cmp_UNIQ)
-    x = flag & cmp_IND? gen_indexsort_uniq(x, E, CMP): gen_sort_uniq(x, E, CMP);
+  if (!CMP)
+  { /* wrt key: precompute all values, O(n) calls instead of O(n log n) */
+    pari_sp av = avma;
+    GEN v, y;
+    long i, tx, lx;
+    init_sort(&x, &tx, &lx);
+    if (lx == 1) return flag&cmp_IND? cgetg(1,t_VECSMALL): triv_sort(tx);
+    v = cgetg(lx, t_VEC);
+    for (i = 1; i < lx; i++) gel(v,i) = closure_callgen1(k, gel(x,i));
+    y = vecsort0(v, NULL, flag | cmp_IND);
+    y = flag&cmp_IND? y: sort_extract(x, y, tx, lg(y));
+    return gerepileupto(av, y);
+  }
+  if (flag&cmp_UNIQ)
+    x = flag&cmp_IND? gen_indexsort_uniq(x, E, CMP): gen_sort_uniq(x, E, CMP);
   else
-    x = flag & cmp_IND? gen_indexsort(x, E, CMP): gen_sort(x, E, CMP);
-  if (flag & cmp_REV) { /* reverse order */
-    long j, lx;
-    GEN y;
-    if (typ(x)==t_LIST)
-    {
-      y = list_data(x);
-      if (!y) return x;
-    }
-    else
-      y = x;
-    lx = lg(y);
-    for (j=1; j<=(lx-1)>>1; j++) swap(gel(y,j), gel(y,lx-j));
+    x = flag&cmp_IND? gen_indexsort(x, E, CMP): gen_sort(x, E, CMP);
+  if (flag & cmp_REV)
+  { /* reverse order */
+    GEN y = x;
+    if (typ(x)==t_LIST) { y = list_data(x); if (!y) return x; }
+    vecreverse_inplace(y);
   }
   return x;
 }
@@ -1583,15 +1541,33 @@ vecsort(GEN x, GEN k)
   if (typ(k) != t_VECSMALL) pari_err_TYPE("vecsort",k);
   return gen_sort(x, (void*)k, &veccmp);
 }
+/* adapted from gen_search; don't export: keys of T[i] should be precomputed */
+static long
+key_search(GEN T, GEN x, GEN code)
+{
+  long u = lg(T)-1, i, l, s;
+
+  if (!u) return 0;
+  l = 1; x = closure_callgen1(code, x);
+  do
+  {
+    i = (l+u)>>1; s = lexcmp(x, closure_callgen1(code, gel(T,i)));
+    if (!s) return i;
+    if (s<0) u=i-1; else l=i+1;
+  } while (u>=l);
+  return 0;
+}
 long
 vecsearch(GEN v, GEN x, GEN k)
 {
   pari_sp av = avma;
   void *E;
-  int (*CMP)(void*,GEN,GEN) = sort_function(&E, x, k);
-  long r;
-  if (!is_matvec_t(typ(v))) pari_err_TYPE("vecsearch", v);
-  r = gen_search(v, x, 0, E, CMP);
+  int (*CMP)(void*,GEN,GEN) = sort_function(&E, v, k);
+  long r, tv = typ(v);
+  if (tv == t_VECSMALL)
+    x = (GEN)itos(x);
+  else if (!is_matvec_t(tv)) pari_err_TYPE("vecsearch", v);
+  r = CMP? gen_search(v, x, 0, E, CMP): key_search(v, x, k);
   avma = av; return r;
 }
 
@@ -1601,6 +1577,8 @@ GEN
 ZV_sort(GEN L) { return gen_sort(L, (void*)&cmpii, &cmp_nodata); }
 GEN
 ZV_sort_uniq(GEN L) { return gen_sort_uniq(L, (void*)&cmpii, &cmp_nodata); }
+void
+ZV_sort_inplace(GEN L) { gen_sort_inplace(L, (void*)&cmpii, &cmp_nodata,NULL); }
 
 /********************************************************************/
 /**                      SEARCH IN SORTED VECTOR                   **/
@@ -1624,10 +1602,10 @@ tablesearch(GEN T, GEN x, int (*cmp)(GEN,GEN))
 long
 gen_search(GEN T, GEN x, long flag, void *data, int (*cmp)(void*,GEN,GEN))
 {
-  long lx = lg(T), i, l, u, s;
+  long u = lg(T)-1, i, l, s;
 
-  if (lx==1) return flag? 1: 0;
-  l = 1; u = lx-1;
+  if (!u) return flag? 1: 0;
+  l = 1;
   do
   {
     i = (l+u)>>1; s = cmp(data, x, gel(T,i));
@@ -1855,7 +1833,7 @@ setsearch(GEN T, GEN y, long flag)
     if (list_typ(T) != t_LIST_RAW) pari_err_TYPE("setsearch",T);
     T = list_data(T); lx = T? lg(T): 1; break;
     default: pari_err_TYPE("setsearch",T);
-      return 0; /*not reached*/
+      return 0; /*LCOV_EXCL_LINE*/
   }
   if (lx==1) return flag? 1: 0;
   return gen_search(T,y,flag,(void*)cmp_universal,cmp_nodata);
@@ -1888,11 +1866,11 @@ setunion(GEN x, GEN y)
   return gerepilecopy(av, z);
 }
 /* in case of equal keys in x,y, take the key from x */
-GEN
-ZV_union_shallow(GEN x, GEN y)
+static GEN
+ZV_union_shallow_t(GEN x, GEN y, long t)
 {
   long i, j, k, lx = lg(x), ly = lg(y);
-  GEN z = cgetg(lx + ly - 1, t_VEC);
+  GEN z = cgetg(lx + ly - 1, t);
   i = j = k = 1;
   while (i<lx && j<ly)
   {
@@ -1911,6 +1889,13 @@ ZV_union_shallow(GEN x, GEN y)
   setlg(z, k); return z;
 }
 
+GEN
+ZV_union_shallow(GEN x, GEN y)
+{ return ZV_union_shallow_t(x, y, t_VEC); }
+
+GEN
+ZC_union_shallow(GEN x, GEN y)
+{ return ZV_union_shallow_t(x, y, t_COL); }
 
 GEN
 setintersect(GEN x, GEN y)
