@@ -204,11 +204,11 @@ factorquad(struct buch_quad *B, GEN f, long nFB, ulong limp)
   for (i=1; lgefint(x) > 3; i++)
   {
     ulong p = uel(FB,i), r;
-    GEN q = diviu_rem(x, p, &r);
+    GEN q = absdiviu_rem(x, p, &r);
     if (!r)
     {
       long k = 0;
-      do { k++; x = q; q = diviu_rem(x, p, &r); } while (!r);
+      do { k++; x = q; q = absdiviu_rem(x, p, &r); } while (!r);
       lo++; P[lo] = p; E[lo] = k;
     }
     if (isless_iu(q,p)) {
@@ -236,7 +236,7 @@ END:
   if (X > B->limhash) return 0;
   if (X != 1 && X <= limp)
   {
-    if (B->badprim && ugcd(X, umodiu(B->badprim,X)) > 1) return 0;
+    if (B->badprim && ugcdui(X, B->badprim) > 1) return 0;
     lo++; P[lo] = X; E[lo] = 1;
     X = 1;
   }
@@ -357,7 +357,7 @@ is_bad(GEN D, ulong p)
     if (r && signe(D) < 0) r = 8-r;
     return (r < 4);
   }
-  r = (remii(D, sqru(p)) == gen_0); /* p^2 | D ? */
+  r = dvdii(D, sqru(p)); /* p^2 | D ? */
   avma = av; return r;
 }
 
@@ -886,8 +886,8 @@ real_be_honest(struct buch_quad *B)
 static GEN
 gcdreal(GEN a,GEN b)
 {
-  if (!signe(a)) return mpabs(b);
-  if (!signe(b)) return mpabs(a);
+  if (!signe(a)) return mpabs_shallow(b);
+  if (!signe(b)) return mpabs_shallow(a);
   if (typ(a)==t_INT)
   {
     if (typ(b)==t_INT) return gcdii(a,b);
@@ -905,7 +905,7 @@ gcdreal(GEN a,GEN b)
     if (e > 0) return NULL;
     r = subrr(a, mulir(q,b)); a = b; b = r;
   }
-  return absr(a);
+  return mpabs_shallow(a);
 }
 
 static int
@@ -917,7 +917,7 @@ get_R(struct buch_quad *B, GEN C, long sreg, GEN z, GEN *ptR)
 
   if (B->PRECREG)
   {
-    R = mpabs(gel(C,1));
+    R = mpabs_shallow(gel(C,1));
     for (i=2; i<=sreg; i++)
     {
       R = gcdreal(gel(C,i), R);
