@@ -78,17 +78,23 @@ static long
 zpsol(GEN T, GEN p, long nu, GEN pnu, GEN x0)
 {
   long i, res;
-  pari_sp av = avma;
+  pari_sp av = avma, btop;
   GEN x, pnup;
 
   res = absequaliu(p,2)? lemma7(T,nu,x0): lemma6(T,p,nu,x0);
   if (res== 1) return 1;
   if (res==-1) return 0;
   x = x0; pnup = mulii(pnu,p);
+  btop = avma;
   for (i=0; i < itos(p); i++)
   {
     x = addii(x,pnu);
     if (zpsol(T,p,nu+1,pnup,x)) { avma = av; return 1; }
+    if (gc_needed(btop, 2))
+    {
+      x = gerepileupto(btop, x);
+      if (DEBUGMEM > 1) pari_warn(warnmem, "zpsol: %ld/%Ps",i,p);
+    }
   }
   avma = av; return 0;
 }
@@ -115,8 +121,8 @@ quad_char(GEN nf, GEN t, GEN pr)
   t = nf_to_Fq(nf,t,modpr);
   if (T)
   {
-    ord = subis( pr_norm(pr), 1 ); /* |(O_K / pr)^*| */
-    ordp= subis( p, 1);            /* |F_p^*|        */
+    ord = subiu( pr_norm(pr), 1 ); /* |(O_K / pr)^*| */
+    ordp= subiu( p, 1);            /* |F_p^*|        */
     t = Fq_pow(t, diviiexact(ord, ordp), T,p); /* in F_p^* */
     if (typ(t) == t_POL)
     {
@@ -151,7 +157,7 @@ psquarenf(GEN nf,GEN x,GEN pr,GEN modpr)
     if (v&1) return 0;
     v = (Z_quad_char(x, pr) == 1);
   } else {
-    v = ZC_nfvalrem(nf, x, pr, &x);
+    v = ZC_nfvalrem(x, pr, &x);
     if (v&1) return 0;
     v = (quad_char(nf, x, modpr) == 1);
   }
@@ -160,9 +166,9 @@ psquarenf(GEN nf,GEN x,GEN pr,GEN modpr)
 
 /* Is  x a square in (ZK / pr^(1+2e))^* ?  pr | 2 */
 static long
-check2(GEN nf, GEN x, GEN zinit)
+check2(GEN nf, GEN x, GEN sprk)
 {
-  GEN zlog = ideallog(nf, x, zinit);
+  GEN zlog = zlog_pr(nf, x, sprk);
   long i, l = lg(zlog);
   for (i=1; i<l; i++) /* all elementary divisors are even (1+2e > 1) */
     if (mpodd(gel(zlog,i))) return 0;
@@ -171,17 +177,17 @@ check2(GEN nf, GEN x, GEN zinit)
 
 /* pr | 2. Return 1 if x in Z_K is square in Z_{K_pr}, 0 otherwise */
 static int
-psquare2nf_i(GEN nf,GEN x,GEN pr,GEN zinit)
+psquare2nf_i(GEN nf,GEN x,GEN pr,GEN sprk)
 {
   long v = nfvalrem(nf, x, pr, &x);
   /* now (x,pr) = 1 */
-  return v == LONG_MAX || (!odd(v) && check2(nf,x,zinit));
+  return v == LONG_MAX || (!odd(v) && check2(nf,x,sprk));
 }
 static int
-psquare2nf(GEN nf,GEN x,GEN pr,GEN zinit)
+psquare2nf(GEN nf,GEN x,GEN pr,GEN sprk)
 {
   pari_sp av = avma;
-  long v = psquare2nf_i(nf,x,pr,zinit);
+  long v = psquare2nf_i(nf,x,pr,sprk);
   avma = av; return v;
 }
 
@@ -205,12 +211,12 @@ lemma6nf(GEN nf, GEN T, GEN pr, long nu, GEN x, GEN modpr)
 }
 /* pr above 2 */
 static long
-lemma7nf(GEN nf, GEN T, GEN pr, long nu, GEN x, GEN zinit)
+lemma7nf(GEN nf, GEN T, GEN pr, long nu, GEN x, GEN sprk)
 {
   long res, la, mu, q;
   GEN gpx, gx = nfpoleval(nf, T, x);
 
-  if (psquare2nf(nf,gx,pr,zinit)) return 1;
+  if (psquare2nf(nf,gx,pr,sprk)) return 1;
 
   gpx = nfpoleval(nf, RgX_deriv(T), x);
   /* gx /= pi^la, pi a pr-uniformizer */
@@ -234,11 +240,11 @@ lemma7nf(GEN nf, GEN T, GEN pr, long nu, GEN x, GEN zinit)
   if (q == 1) return res;
 
   /* is gx a square mod pi^q ? FIXME : highly inefficient */
-  zinit = Idealstarprk(nf, pr, q, nf_INIT);
-  if (!check2(nf, gx, zinit)) res = -1;
+  sprk = zlog_pr_init(nf, pr, q);
+  if (!check2(nf, gx, sprk)) res = -1;
   return res;
 }
-/* zinit either a bid (pr | 2) or a modpr structure (pr | p odd).
+/* zinit either a sprk (pr | 2) or a modpr structure (pr | p odd).
    pnu = pi^nu, pi a uniformizer */
 static long
 zpsolnf(GEN nf,GEN T,GEN pr,long nu,GEN pnu,GEN x0,GEN repr,GEN zinit)
@@ -299,7 +305,7 @@ nf_hyperell_locally_soluble(GEN nf,GEN T,GEN pr)
   checkprid(pr); nf = checknf(nf);
   if (absequaliu(pr_get_p(pr), 2))
   { /* tough case */
-    zinit = Idealstarprk(nf, pr, 1+2*pr_get_e(pr), nf_INIT);
+    zinit = zlog_pr_init(nf, pr, 1+2*pr_get_e(pr));
     if (psquare2nf(nf,constant_coeff(T),pr,zinit)) return 1;
     if (psquare2nf(nf, leading_coeff(T),pr,zinit)) return 1;
   }
@@ -330,10 +336,9 @@ den_remove(GEN nf, GEN a)
     case t_COL:
       a = Q_remove_denom(a, &da);
       if (da) a = ZC_Z_mul(a, da);
-      a = coltoliftalg(nf, a);
-      return a;
+      return nf_to_scalar_or_alg(nf, a);
     default: pari_err_TYPE("nfhilbert",a);
-      return NULL;/*not reached*/
+      return NULL;/*LCOV_EXCL_LINE*/
   }
 }
 
@@ -370,7 +375,7 @@ nfhilbertp(GEN nf, GEN a, GEN b, GEN pr)
   if (!odd(va) && !odd(vb)) { avma = av; return 1; }
   /* Trick: pretend the exponent is 2, result is OK up to squares ! */
   t = famat_makecoprime(nf, mkvec2(a,b), mkvec2s(vb, -va),
-                        pr, idealhnf_two(nf, pr), gen_2);
+                        pr, pr_hnf(nf, pr), gen_2);
   if (typ(t) == t_INT) {
     if (odd(va) && odd(vb)) t = negi(t);
     /* t = (-1)^(v(a)v(b)) a^v(b) b^(-v(a)) */
@@ -496,7 +501,7 @@ bnfsunit0(GEN bnf, GEN S, long flag, long prec)
     long l;
     ZV_snf_trunc(D); l = lg(D);
     card = ZV_prod(D);
-    A = cgetg(l, t_VEC); pow = ZM_inv(u,gen_1);
+    A = cgetg(l, t_VEC); pow = ZM_inv(u, NULL);
     for(i = 1; i < l; i++) gel(A,i) = idealfactorback(nf, gen, gel(pow,i), 1);
     gel(res,5) = mkvec3(card, D, A);
   }
@@ -535,7 +540,7 @@ bnfsunit0(GEN bnf, GEN S, long flag, long prec)
       v = gel(v,2); if (flag == nf_GEN) v = nf_to_scalar_or_alg(nf, v);
       gel(sunit,i) = v;
    }
-    den = ZM_det_triangular(H); H = ZM_inv(H,den);
+    H = ZM_inv(H,&den);
     A = shallowconcat(H, ZM_neg(ZM_mul(H,B))); /* top part of inverse * den */
     /* HNF in split form perm + (H B) [0 Id missing] */
     gel(res,1) = sunit;
@@ -585,7 +590,7 @@ make_unit(GEN nf, GEN bnfS, GEN *px)
   for (i=1; i<ls; i++)
   {
     GEN P = gel(S,i), p = pr_get_p(P);
-    if ( remii(N, p) == gen_0 )
+    if (dvdii(N, p))
     {
       v[i] = nfval(nf,xb,P);
       (void)Z_pvalrem(N0, p, &N0);
@@ -605,14 +610,14 @@ make_unit(GEN nf, GEN bnfS, GEN *px)
   v = shallowconcat(v, zc_to_ZC(p1)); /* append bottom of p1 (= [0 Id] part) */
 
   gen = gel(bnfS,1);
-  xp = cgetg(1, t_MAT);
+  xp = trivial_fact();
   for (i=1; i<ls; i++)
   {
     GEN e = gel(v,i);
     if (!signe(e)) continue;
     xp = famat_mulpow_shallow(xp, gel(gen,i), negi(e));
   }
-  if (lg(xp) > 1) *px = famat_mulpow_shallow(xp, xb, gen_1);
+  if (lgcols(xp) != 1) *px = famat_mulpow_shallow(xp, xb, gen_1);
   return v;
 }
 
@@ -684,10 +689,9 @@ static GEN
 nfX_eltup(GEN nf, GEN rnfeq, GEN x)
 {
   long i, l;
-  GEN zknf, czknf, y = cgetg_copy(x, &l);
-  y[1] = x[1]; nf_nfzk(nf, rnfeq, &zknf, &czknf);
-  for (i=2; i<l; i++) gel(y,i) = nfeltup(nf, gel(x,i), zknf, czknf);
-  return y;
+  GEN y = cgetg_copy(x, &l), zknf = nf_nfzk(nf, rnfeq);
+  for (i=2; i<l; i++) gel(y,i) = nfeltup(nf, gel(x,i), zknf);
+  y[1] = x[1]; return y;
 }
 
 static hashtable *
@@ -739,7 +743,7 @@ rnfisnorminit(GEN T, GEN relpol, int galois)
   for(i=1; i<l; i++)
   {
     GEN g = gel(gen,i);
-    if (ugcd(umodiu(gel(cyc,i), drel), drel) == 1) break;
+    if (ugcdiu(gel(cyc,i), drel) == 1) break;
     Zfa_append(gcoeff(g,1,1), H, NULL);
   }
   if (!galois)
