@@ -12,6 +12,7 @@ with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 #include "pari.h"
+#include "paripriv.h"
 
 /* Return 1 if the point Q is a Weierstrass (2-torsion) point of the
  * curve E, return 0 otherwise */
@@ -65,6 +66,30 @@ RgX_homogenous_evalpow(GEN P, GEN A, GEN B)
   return mkvec2(s, gel(B,d+1));
 }
 
+static GEN
+RgXQX_homogenous_evalpow(GEN P, GEN A, GEN B, GEN T)
+{
+  pari_sp av = avma;
+  long i, d = degpol(P), v = varn(A);
+  GEN s;
+  if (signe(P)==0) return mkvec2(pol_0(v), pol_1(v));
+  s = scalarpol_shallow(gel(P, d+2), v);
+  for (i = d-1; i >= 0; i--)
+  {
+    s = RgX_add(RgXQX_mul(s, A, T), RgXQX_RgXQ_mul(gel(B,d+1-i), gel(P,i+2), T));
+    if (gc_needed(av,1))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"RgX_homogenous_eval(%ld)",i);
+      s = gerepileupto(av, s);
+    }
+  }
+  s = gerepileupto(av, s);
+  return mkvec2(s, gel(B,d+1));
+}
+
+/* x must be nonzero */
+INLINE long _degree(GEN x) { return typ(x)==t_POL ? degpol(x): 0; }
+
 /* Given isogenies F:E' -> E and G:E'' -> E', return the composite
  * isogeny F o G:E'' -> E */
 static GEN
@@ -85,7 +110,8 @@ ellcompisog(GEN F, GEN G)
   K = RgX_normalize(RgX_div(K, RgX_gcd(K,deriv(K,0))));
   K2 = gsqr(K); K3 = gmul(K, K2);
   F0 = polcoeff0(gel(F,2), 0, vy); F1 = polcoeff0(gel(F,2), 1, vy);
-  d = maxss(maxss(degpol(gel(F,1)),degpol(gel(F,3))),maxss(degpol(F0),degpol(F1)));
+  d = maxss(maxss(degpol(gel(F,1)),_degree(gel(F,3))),
+            maxss(_degree(F0),_degree(F1)));
   Gp = gpowers(Gh2, d);
   f  = RgX_homogenous_evalpow(gel(F,1), gel(G,1), Gp);
   g0 = RgX_homogenous_evalpow(F0, gel(G,1), Gp);
@@ -100,6 +126,92 @@ ellcompisog(GEN F, GEN G)
   return gerepilecopy(av, mkvec3(f,g,K));
 }
 
+static GEN
+to_RgX(GEN P, long vx)
+{
+  return typ(P) == t_POL ? lift(P): scalarpol_shallow(lift(P), vx);
+}
+
+static GEN
+divy(GEN P0, GEN P1, GEN Q, GEN T, long vy)
+{
+  GEN DP0, P0r = Q_remove_denom(P0, &DP0), P0D;
+  GEN DP1, P1r = Q_remove_denom(P1, &DP1), P1D;
+  GEN DQ, Qr = Q_remove_denom(Q, &DQ), P2;
+  P0D = RgXQX_div(P0r, Qr, T);
+  if (DP0) P0D = gdiv(P0D, DP0);
+  P1D = RgXQX_div(P1r, Qr, T);
+  if (DP1) P1D = gdiv(P1D, DP1);
+  P2 = gadd(gmul(P1D, pol_x(vy)), P0D);
+  if (DQ) P2 = gmul(P2, DQ);
+  return P2;
+}
+
+static GEN
+ellnfcompisog(GEN nf, GEN F, GEN G)
+{
+  pari_sp av = avma;
+  GEN Fv, Gh, Gh2, Gh3, f, g, gd, h, h21, h22, h31, h32, den;
+  GEN K, K2, K3, F0, F1, G0, G1, g0, g1, Gp;
+  GEN num0, num1, gn0, gn1;
+  GEN g0d, g01, k3h32;
+  GEN T, res;
+  pari_timer ti;
+  long v, vx, vy, d;
+  if (!nf) return ellcompisog(F, G);
+  T = nf_get_pol(nf);
+  timer_start(&ti);
+  checkellisog(F);
+  checkellisog(G);
+  get_isog_vars(F, &vx, &vy);
+  v = fetch_var_higher();
+  Fv = shallowcopy(gel(F,3)); setvarn(Fv, v);
+  Gh = lift(gel(G,3)); Gh2 = RgXQX_sqr(Gh, T); Gh3 = RgXQX_mul(Gh, Gh2, T);
+  res = to_RgX(polresultant0(Fv, deg1pol(gmul(gneg(Gh2),gmodulo(gen_1,T)),gel(G,1), v), v, 0),vx);
+  delete_var();
+  K = Q_remove_denom(RgXQX_mul(res, Gh, T), NULL);
+  if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: resultant");
+  K = RgXQX_div(K, nfgcd(K, deriv(K,0), T, NULL), T);
+  K = RgX_normalize(K);
+  if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: nfgcd");
+  K2 = RgXQX_sqr(K, T); K3 = RgXQX_mul(K, K2, T);
+  F0 = to_RgX(polcoeff0(gel(F,2), 0, vy), vx);
+  F1 = to_RgX(polcoeff0(gel(F,2), 1, vy), vx);
+  G0 = to_RgX(polcoeff0(gel(G,2), 0, vy), vx);
+  G1 = to_RgX(polcoeff0(gel(G,2), 1, vy), vx);
+  d = maxss(maxss(degpol(gel(F,1)),degpol(gel(F,3))),maxss(degpol(F0),degpol(F1)));
+  Gp = RgXQX_powers(Gh2, d, T);
+  f  = RgXQX_homogenous_evalpow(to_RgX(gel(F,1),vx), gel(G,1), Gp, T);
+  g0 = RgXQX_homogenous_evalpow(F0, to_RgX(gel(G,1),vx), Gp, T);
+  g1 = RgXQX_homogenous_evalpow(F1, to_RgX(gel(G,1),vx), Gp, T);
+  h  = RgXQX_homogenous_evalpow(to_RgX(gel(F,3),vx), gel(G,1), Gp, T);
+  if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: evalpow");
+  h21 = RgXQX_sqr(gel(h,1),T);
+  h22 = RgXQX_sqr(gel(h,2),T);
+  h31 = RgXQX_mul(gel(h,1), h21,T);
+  h32 = RgXQX_mul(gel(h,2), h22,T);
+  if (DEBUGLEVEL) timer_printf(&ti,"h");
+  f  = RgXQX_div(RgXQX_mul(RgXQX_mul(K2, gel(f,1), T), h22, T),
+                           RgXQX_mul(gel(f,2), h21, T), T);
+  if (DEBUGLEVEL) timer_printf(&ti,"f");
+  den = RgXQX_mul(Gh3, gel(g1,2), T);
+  if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: den");
+  g0d = RgXQX_mul(gel(g0,1),den, T);
+  g01 = RgXQX_mul(gel(g1,1),gel(g0,2),T);
+  num0 = RgX_add(g0d, RgXQX_mul(G0,g01, T));
+  num1 = RgXQX_mul(G1,g01, T);
+  if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: num");
+  k3h32 = RgXQX_mul(K3,h32,T);
+  gn0 = RgXQX_mul(num0, k3h32, T);
+  gn1 = RgXQX_mul(num1, k3h32, T);
+  if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: gn");
+  gd = RgXQX_mul(RgXQX_mul(gel(g0,2), den, T), h31, T);
+  if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: gd");
+  g = divy(gn0, gn1, gd, T, vy);
+  if (DEBUGLEVEL) timer_printf(&ti,"ellnfcompisog: divy");
+  return gerepilecopy(av, gmul(mkvec3(f,g,K),gmodulo(gen_1,T)));
+}
+
 /* Given an isogeny phi from ellisogeny() and a point P in the domain of phi,
  * return phi(P) */
 GEN
@@ -403,7 +515,7 @@ isog_abscissa(GEN E, GEN kerp, GEN h0, GEN x, GEN two_tors)
   f2 = gel(two_tors, 3);
   h2 = gel(two_tors, 5);
 
-  /* Combine f0 and f2 into the final abcissa of the isogeny. */
+  /* Combine f0 and f2 into the final abscissa of the isogeny. */
   t1 = RgX_mul(x, RgX_sqr(kerp));
   t2 = RgX_mul(f2, RgX_sqr(h0));
   t3 = RgX_mul(f0, RgX_sqr(h2));
@@ -501,13 +613,14 @@ isogeny_from_kernel_poly(GEN E, GEN kerp, long only_image, long vx, long vy)
   /* isogeny degree: 2*degpol(kerp)+1-degpol(kerh) */
   m = degpol(kerq);
 
-  kerp = RgX_Rg_div(kerp, leading_coeff(kerp));
-  kerq = RgX_Rg_div(kerq, leading_coeff(kerq));
-  kerh = RgX_Rg_div(kerh, leading_coeff(kerh));
+  kerp = RgX_normalize(kerp);
+  kerq = RgX_normalize(kerq);
+  kerh = RgX_normalize(kerh);
   switch(degpol(kerh))
   {
   case 0:
-    two_tors = mkvec5(gen_0, gen_0, pol_0(vx), pol_0(vx), pol_1(vx));
+    two_tors = only_image? mkvec2(gen_0, gen_0):
+      mkvec5(gen_0, gen_0, pol_0(vx), pol_0(vx), pol_1(vx));
     break;
   case 1:
     two_tors = contrib_weierstrass_pt(E, kerh, only_image,vx,vy);
@@ -556,8 +669,10 @@ ellisogeny(GEN E, GEN G, long only_image, long vx, long vy)
   checkell(E);j = ell_get_j(E);
   if (vx < 0) vx = 0;
   if (vy < 0) vy = 1;
-  if (varncmp(vx, vy) >= 0) pari_err_PRIORITY("ellisogeny", pol_x(vx), "<=", vy);
-  if (varncmp(vy, gvar(j)) >= 0) pari_err_PRIORITY("ellisogeny", j, ">=", vy);
+  if (varncmp(vx, vy) >= 0)
+    pari_err_PRIORITY("ellisogeny", pol_x(vx), "<=", vy);
+  if (!only_image && varncmp(vy, gvar(j)) >= 0)
+    pari_err_PRIORITY("ellisogeny", j, ">=", vy);
   switch(typ(G))
   {
   case t_VEC:
@@ -565,13 +680,18 @@ ellisogeny(GEN E, GEN G, long only_image, long vx, long vy)
     if (!ell_is_inf(G))
     {
       GEN x =  gel(G,1), y = gel(G,2);
-      if (varncmp(vy, gvar(x)) >= 0) pari_err_PRIORITY("ellisogeny", x, ">=", vy);
-      if (varncmp(vy, gvar(y)) >= 0) pari_err_PRIORITY("ellisogeny", y, ">=", vy);
+      if (!only_image)
+      {
+        if (varncmp(vy, gvar(x)) >= 0)
+          pari_err_PRIORITY("ellisogeny", x, ">=", vy);
+        if (varncmp(vy, gvar(y)) >= 0)
+          pari_err_PRIORITY("ellisogeny", y, ">=", vy);
+      }
     }
     z = isogeny_from_kernel_point(E, G, only_image, vx, vy);
     break;
   case t_POL:
-    if (varncmp(vy, gvar(constant_coeff(G))) >= 0)
+    if (!only_image && varncmp(vy, gvar(constant_coeff(G))) >= 0)
       pari_err_PRIORITY("ellisogeny", constant_coeff(G), ">=", vy);
     z = isogeny_from_kernel_poly(E, G, only_image, vx, vy);
     break;
@@ -639,49 +759,42 @@ ellisog_by_Kohel(GEN a4, GEN a6, long n, GEN ker, GEN kert, long flag)
 static GEN
 ellisog_by_roots(GEN a4, GEN a6, long n, GEN z, long flag)
 {
-  return ellisog_by_Kohel(a4, a6, n, deg1pol(gen_1, gneg(z), 0),
-                                  deg1pol(gen_1, gmulsg(n, z), 0), flag);
+  GEN k = deg1pol_shallow(gen_1, gneg(z), 0);
+  GEN kt= deg1pol_shallow(gen_1, gmulsg(n,z), 0);
+  return ellisog_by_Kohel(a4, a6, n, k, kt, flag);
 }
 
+/* n = 2 or 3 */
 static GEN
 a4a6_divpol(GEN a4, GEN a6, long n)
 {
-  switch(n)
-  {
-    case 2:
-      return mkpoln(4, gen_1, gen_0, a4, a6);
-    case 3:
-      return mkpoln(5, utoi(3), gen_0, gmulgs(a4,6) , gmulgs(a6,12),
-                       gneg(gsqr(a4)));
-  }
-  return NULL;
+  if (n == 2) return mkpoln(4, gen_1, gen_0, a4, a6);
+  return mkpoln(5, utoi(3), gen_0, gmulgs(a4,6) , gmulgs(a6,12),
+                   gneg(gsqr(a4)));
 }
 
 static GEN
-ellisograph_Kohel_iso(GEN nf, GEN e, long n, GEN z, long flag)
+ellisograph_Kohel_iso(GEN nf, GEN e, long n, GEN z, GEN *pR, long flag)
 {
   long i, r;
-  GEN R, V;
-  GEN c4 = gel(e,1), c6 = gel(e, 2);
+  GEN R, V, c4 = gel(e,1), c6 = gel(e,2);
   GEN a4 = gdivgs(c4, -48), a6 = gdivgs(c6, -864);
   GEN P = a4a6_divpol(a4, a6, n);
   R = nfroots(nf, z ? RgX_div_by_X_x(P, z, NULL): P);
-  r = lg(R);
-  V = cgetg(r, t_VEC);
-  for (i=1; i < r; i++)
-    gel(V, i) = ellisog_by_roots(a4, a6, n, gel(R, i), flag);
-  return mkvec2(V, R);
+  if (pR) *pR = R;
+  r = lg(R); V = cgetg(r, t_VEC);
+  for (i=1; i < r; i++) gel(V,i) = ellisog_by_roots(a4, a6, n, gel(R,i), flag);
+  return V;
 }
 
 static GEN
 ellisograph_Kohel_r(GEN nf, GEN e, long n, GEN z, long flag)
 {
-  GEN W = ellisograph_Kohel_iso(nf, e, n, z, flag);
-  GEN iso = gel(W, 1), R = gel(W, 2);
+  GEN R, iso = ellisograph_Kohel_iso(nf, e, n, z, &R, flag);
   long i, r = lg(iso);
   GEN V = cgetg(r, t_VEC);
   for (i=1; i < r; i++)
-    gel(V, i) = ellisograph_Kohel_r(nf, gel(iso, i), n, gmulgs(gel(R, i), -n), flag);
+    gel(V,i) = ellisograph_Kohel_r(nf, gel(iso,i), n, gmulgs(gel(R,i), -n), flag);
   return mkvec2(e, V);
 }
 
@@ -793,8 +906,8 @@ ellisograph_iso(GEN nf, GEN e, ulong p, GEN P, GEN oj, long flag)
 {
   long i, r;
   GEN Pj, R, V;
-  GEN j = gel(e, 3);
-  Pj = poleval(P, j);
+  if (!P) return ellisograph_Kohel_iso(nf, e, p, oj, NULL, flag);
+  Pj = poleval(P, gel(e,3));
   R = nfroots(nf,oj ? RgX_div_by_X_x(Pj, oj, NULL):Pj);
   r = lg(R);
   V = cgetg(r, t_VEC);
@@ -806,12 +919,10 @@ ellisograph_iso(GEN nf, GEN e, ulong p, GEN P, GEN oj, long flag)
 static GEN
 ellisograph_r(GEN nf, GEN e, ulong p, GEN P, GEN oj, long flag)
 {
-  GEN iso = ellisograph_iso(nf, e, p, P, oj, flag);
-  GEN j = gel(e, 3);
+  GEN j = gel(e,3), iso = ellisograph_iso(nf, e, p, P, oj, flag);
   long i, r = lg(iso);
   GEN V = cgetg(r, t_VEC);
-  for (i=1; i < r; i++)
-    gel(V, i) = ellisograph_r(nf, gel(iso, i), p, P, j, flag);
+  for (i=1; i < r; i++) gel(V,i) = ellisograph_r(nf, gel(iso,i), p, P, j, flag);
   return mkvec2(e, V);
 }
 
@@ -826,8 +937,7 @@ ellisograph_a4a6(GEN E, long flag)
 static GEN
 ellisograph_dummy(GEN E, long n, GEN jt, GEN jtt, GEN s0, long flag)
 {
-  GEN c4 = ell_get_c4(E), c6 = ell_get_c6(E);
-  GEN c6c4 = gdiv(c6, c4);
+  GEN c4 = ell_get_c4(E), c6 = ell_get_c6(E), c6c4 = gdiv(c6, c4);
   GEN jtp = gmul(c6c4, gdivgs(gmul(jt, jtt), -n));
   GEN iso = ellisog_by_jt(c4, c6, jt, jtp, gmul(s0, c6c4), n, flag);
   GEN v = mkvec2(iso, cgetg(1, t_VEC));
@@ -835,20 +945,28 @@ ellisograph_dummy(GEN E, long n, GEN jt, GEN jtt, GEN s0, long flag)
 }
 
 static GEN
-ellisograph_p(GEN nf, GEN E, ulong p, long flag)
+isograph_p(GEN nf, GEN e, ulong p, GEN P, long flag)
 {
   pari_sp av = avma;
-  GEN iso, e = ellisograph_a4a6(E, flag);
-  if (p > 3)
-  {
-    GEN P = polmodular_ZXX(p, 0, 0, 1);
+  GEN iso;
+  if (P)
     iso = ellisograph_r(nf, e, p, P, NULL, flag);
-  }
   else
     iso = ellisograph_Kohel_r(nf, e, p, NULL, flag);
   return gerepilecopy(av, iso);
 }
 
+static GEN
+get_polmodular(ulong p)
+{ return p > 3 ? polmodular_ZXX(p,0,0,1): NULL; }
+static GEN
+ellisograph_p(GEN nf, GEN E, ulong p, long flag)
+{
+  GEN e = ellisograph_a4a6(E, flag);
+  GEN P = get_polmodular(p);
+  return isograph_p(nf, e, p, P, flag);
+}
+
 static long
 etree_nbnodes(GEN T)
 {
@@ -860,15 +978,15 @@ etree_nbnodes(GEN T)
 }
 
 static long
-etree_listr(GEN T, GEN V, long n, GEN u, GEN ut)
+etree_listr(GEN nf, GEN T, GEN V, long n, GEN u, GEN ut)
 {
   GEN E = gel(T, 1), F = gel(T,2);
   long i, l = lg(F);
   GEN iso, isot = NULL;
   if (lg(E) == 6)
   {
-    iso  = ellisogenyapply(gel(E,4), u);
-    isot = ellisogenyapply(ut, gel(E,5));
+    iso  = ellnfcompisog(nf,gel(E,4), u);
+    isot = ellnfcompisog(nf,ut, gel(E,5));
     gel(V, n) = mkvec5(gel(E,1), gel(E,2), gel(E,3), iso, isot);
   } else
   {
@@ -876,16 +994,16 @@ etree_listr(GEN T, GEN V, long n, GEN u, GEN ut)
     iso = u;
   }
   for (i = 1; i < l; i++)
-    n = etree_listr(gel(F, i), V, n + 1, iso, isot);
+    n = etree_listr(nf, gel(F, i), V, n + 1, iso, isot);
   return n;
 }
 
 static GEN
-etree_list(GEN T)
+etree_list(GEN nf, GEN T)
 {
   long n = etree_nbnodes(T);
   GEN V = cgetg(n+1, t_VEC);
-  (void) etree_listr(T, V, 1, trivial_isogeny(), trivial_isogeny());
+  (void) etree_listr(nf, T, V, 1, trivial_isogeny(), trivial_isogeny());
   return V;
 }
 
@@ -929,83 +1047,57 @@ etree_distmat(GEN T)
   return M;
 }
 
-static GEN
-list_to_crv(GEN L)
-{
-  long i, l;
-  GEN V = cgetg_copy(L, &l);
-  for(i=1; i < l; i++)
-  {
-    GEN Li = gel(L, i);
-    GEN e = mkvec2(gdivgs(gel(Li,1), -48), gdivgs(gel(Li,2), -864));
-    gel(V, i) = lg(Li)==6 ? mkvec3(e, gel(Li,4), gel(Li,5)): e;
-  }
-  return V;
-}
-
 static GEN
 distmat_pow(GEN E, ulong p)
 {
-  long i, j, n = lg(E)-1;
-  GEN M = cgetg(n+1, t_MAT);
-  for(i = 1; i <= n; i++)
+  long i, j, l = lg(E);
+  GEN M = cgetg(l, t_MAT);
+  for(i = 1; i < l; i++)
   {
-    gel(M,i) = cgetg(n+1, t_VEC);
-    for(j = 1; j <= n; j++)
-      gmael(M,i,j) = powuu(p,mael(E,i,j));
+    gel(M,i) = cgetg(l, t_COL);
+    for(j = 1; j < l; j++) gmael(M,i,j) = powuu(p,mael(E,i,j));
   }
   return M;
 }
 
 /* Assume there is a single p-isogeny */
-
 static GEN
 isomatdbl(GEN nf, GEN L, GEN M, ulong p, GEN T2, long flag)
 {
   long i, j, n = lg(L) -1;
-  GEN P = p > 3 ? polmodular_ZXX(p, 0, 0, 1): NULL;
-  GEN V = cgetg(2*n+1, t_VEC);
-  GEN N = cgetg(2*n+1, t_MAT);
-  for(i=1; i <= n; i++)
-    gel(V, i) = gel(L, i);
-  for(i=1; i <= n; i++)
+  GEN P = get_polmodular(p), V = cgetg(2*n+1, t_VEC), N = cgetg(2*n+1, t_MAT);
+  for (i=1; i <= n; i++)
   {
-    GEN e = gel(L, i);
-    GEN F, E;
+    GEN F, E, e = gel(L,i);
     if (i == 1)
       F = gmael(T2, 2, 1);
     else
     {
-      if (p > 3)
-        F = ellisograph_iso(nf, e, p, P, NULL, flag);
-      else
-        F = gel(ellisograph_Kohel_iso(nf, e, p, NULL, flag), 1);
+      F = ellisograph_iso(nf, e, p, P, NULL, flag);
       if (lg(F) != 2) pari_err_BUG("isomatdbl");
     }
     E = gel(F, 1);
     if (flag)
-      gel(V, i+n) = mkvec3(gel(E,1), gel(E,2), gel(E,3));
+      E = mkvec3(gel(E,1), gel(E,2), gel(E,3));
     else
     {
-      GEN iso = ellisogenyapply(gel(E,4), gel(e, 4));
-      GEN isot = ellisogenyapply(gel(e,5), gel(E, 5));
-      gel(V, i+n) = mkvec5(gel(E,1), gel(E,2), gel(E,3), iso, isot);
+      GEN iso = ellnfcompisog(nf, gel(E,4), gel(e, 4));
+      GEN isot = ellnfcompisog(nf, gel(e,5), gel(E, 5));
+      E = mkvec5(gel(E,1), gel(E,2), gel(E,3), iso, isot);
     }
+    gel(V, i)   = e;
+    gel(V, i+n) = E;
   }
-  for(i=1; i <= 2*n; i++)
-    gel(N, i) = cgetg(2*n+1, t_COL);
-  for(i=1; i <= n; i++)
-    for(j=1; j <= n; j++)
+  for (i=1; i <= 2*n; i++) gel(N, i) = cgetg(2*n+1, t_COL);
+  for (i=1; i <= n; i++)
+    for (j=1; j <= n; j++)
     {
       gcoeff(N,i,j) = gcoeff(N,i+n,j+n) = gcoeff(M,i,j);
       gcoeff(N,i,j+n) = gcoeff(N,i+n,j) = muliu(gcoeff(M,i,j), p);
     }
-  return mkvec2(list_to_crv(V), N);
+  return mkvec2(V, N);
 }
 
-INLINE GEN
-mkfracss(long x, long y) { retmkfrac(stoi(x),stoi(y)); }
-
 static ulong
 ellQ_exceptional_iso(GEN j, GEN *jt, GEN *jtp, GEN *s0)
 {
@@ -1067,21 +1159,17 @@ ellQ_exceptional_iso(GEN j, GEN *jt, GEN *jtp, GEN *s0)
   return 0;
 }
 
+static GEN
+nfmkisomat(GEN nf, ulong p, GEN T)
+{ return mkvec2(etree_list(nf,T), distmat_pow(etree_distmat(T),p)); }
 static GEN
 mkisomat(ulong p, GEN T)
-{
-  pari_sp av = avma;
-  GEN L = list_to_crv(etree_list(T));
-  GEN M = distmat_pow(etree_distmat(T), p);
-  return gerepilecopy(av, mkvec2(L, M));
-}
-
+{ return nfmkisomat(NULL, p, T); }
 static GEN
 mkisomatdbl(ulong p, GEN T, ulong p2, GEN T2, long flag)
 {
-  GEN L = etree_list(T);
-  GEN M = distmat_pow(etree_distmat(T), p);
-  return isomatdbl(NULL, L, M, p2, T2, flag);
+  GEN v = mkisomat(p,T);
+  return isomatdbl(NULL, gel(v,1), gel(v,2), p2, T2, flag);
 }
 
 /*
@@ -1093,60 +1181,496 @@ Volume 15, Issue 2, October 1982, Pages 199-202
 http://www.sciencedirect.com/science/article/pii/0022314X82900257
 */
 
+enum { _2 = 1, _3 = 2, _5 = 4, _7 = 8, _13 = 16 };
+static ulong
+ellQ_goodl(GEN E)
+{
+  forprime_t T;
+  long i, CM = ellQ_get_CM(E);
+  ulong mask = 31;
+  GEN disc = ell_get_disc(E);
+  pari_sp av = avma;
+  u_forprime_init(&T, 17UL,ULONG_MAX);
+  for(i=1; mask && i<=20; i++)
+  {
+    ulong p = u_forprime_next(&T);
+    if (umodiu(disc,p)==0) i--;
+    else
+    {
+      long t = ellap_CM_fast(E, p, CM), D = t*t-4*p;
+      if (t%2) mask &= ~_2;
+      if ((mask & _3) && kross(D,3)==-1)  mask &= ~_3;
+      if ((mask & _5) && kross(D,5)==-1)  mask &= ~_5;
+      if ((mask & _7) && kross(D,7)==-1)  mask &= ~_7;
+      if ((mask &_13) && kross(D,13)==-1) mask &= ~_13;
+    }
+  }
+  avma = av; return mask;
+}
+
+static long
+ellQ_goodl_l(GEN E, long l)
+{
+  forprime_t T;
+  long i;
+  GEN disc = ell_get_disc(E);
+  pari_sp av = avma;
+  u_forprime_init(&T, 17UL,ULONG_MAX);
+  for(i=1; i<=20; i++)
+  {
+    ulong p = u_forprime_next(&T);
+    if (umodiu(disc,p)==0) { i--; continue; }
+    else
+    {
+      long t = itos(ellap(E, utoi(p)));
+      if (l==2)
+      {
+        if (t%2==1) return 0;
+      }
+      else
+      {
+        long D = t*t-4*p;
+        if (kross(D,l)==-1) return 0;
+      }
+      avma = av;
+    }
+  }
+  return 1;
+}
+
+static ulong
+ellnf_goodl_l(GEN E, GEN v)
+{
+  forprime_t T;
+  long i;
+  GEN nf = ellnf_get_nf(E);
+  GEN disc = ell_get_disc(E);
+  long lv = lg(v);
+  ulong w = 0UL;
+  pari_sp av = avma;
+  u_forprime_init(&T, 17UL,ULONG_MAX);
+  for(i=1; i<=20; i++)
+  {
+    ulong p = u_forprime_next(&T);
+    GEN pr = idealprimedec(nf, utoi(p));
+    long j, k, lv = lg(v), g = lg(pr)-1;
+    for (j=1; j<=g; j++)
+    {
+      GEN prj = gel(pr, j);
+      if (idealval(nf,disc,prj) > 0) {i--; continue;}
+      else
+      {
+        long t = itos(ellap(E, prj));
+        for(k = 1; k < lv; k++)
+        {
+          long l = v[k];
+          if (l==2)
+          {
+            if (t%2==1) w |= 1<<(k-1);
+          }
+          else
+          {
+            GEN D = subii(sqrs(t),shifti(pr_norm(prj),2));
+            if (krois(D,l)==-1) w |= 1<<(k-1);
+          }
+        }
+      }
+    }
+    avma = av;
+  }
+  return w^((1UL<<(lv-1))-1);
+}
+
+static GEN
+ellnf_charpoly(GEN E, GEN pr)
+{
+  return deg2pol_shallow(gen_1, negi(ellap(E,pr)), pr_norm(pr), 0);
+}
+
+static GEN
+RgX_homogenize(GEN P, long v)
+{
+  GEN Q = leafcopy(P);
+  long i, l = lg(P), d = degpol(P);
+  for (i = 2; i < l; i++) gel(Q,i) = monomial(gel(Q,i), d--, v);
+  return Q;
+}
+
+static GEN
+starlaw(GEN p, GEN q)
+{
+  GEN Q = RgX_homogenize(RgX_recip(q), 1);
+  return ZX_ZXY_resultant(p, Q);
+}
+
+static GEN
+startor(GEN p, long r)
+{
+  GEN xr = pol_xn(r, 0);
+  GEN psir = gsub(xr, gen_1);
+  return gsubstpol(starlaw(p, psir),xr,pol_x(0));
+}
+
+static GEN
+ellnf_get_degree(GEN E, GEN p)
+{
+  GEN nf = ellnf_get_nf(E);
+  long d = nf_get_degree(nf);
+  GEN dec = idealprimedec(nf, p);
+  long i, l = lg(dec), k;
+  GEN R, starl = deg1pol_shallow(gen_1, gen_m1, 0);
+  for(i=1; i < l; i++)
+  {
+    GEN pr = gel(dec,i);
+    GEN q = ellnf_charpoly(E, pr);
+    starl = starlaw(starl, startor(q, 12*pr_get_e(pr)));
+  }
+  R = p;
+  for(k=0; 2*k<=d; k++)
+    R = mulii(R, poleval(starl,powiu(p,12*k)));
+  return R;
+}
+
+/*
+Based on a GP script by Nicolas Billerey itself
+based on Th\'eor\`emes 2.4 and 2.8 of the following article:
+N. Billerey, Crit\`eres d'irr\'eductibilit\'e pour les
+repr\'esentations des courbes elliptiques,
+Int. J. Number Theory 7 (2011), no. 4, 1001-1032.
+*/
+
+static GEN
+ellnf_prime_degree(GEN E)
+{
+  forprime_t T;
+  long i;
+  GEN nf = ellnf_get_nf(E);
+  GEN disc = ell_get_disc(E);
+  GEN P, B = gen_0, rB;
+  GEN bad = mulii(nfnorm(nf, disc),nf_get_disc(nf));
+  u_forprime_init(&T, 5UL,ULONG_MAX);
+  for(i=1; i<=20; i++)
+  {
+    ulong p = u_forprime_next(&T);
+    if (dvdiu(bad, p)) {i--; continue;}
+    B = gcdii(B, ellnf_get_degree(E, utoi(p)));
+    if (Z_issquareall(B,&rB)) B=rB;
+  }
+  if (signe(B)==0) pari_err_IMPL("ellisomat, CM case");
+  P = vec_to_vecsmall(gel(Z_factor(B),1));
+  return shallowextract(P, utoi(ellnf_goodl_l(E, P)));
+}
+
 static GEN
 ellQ_isomat(GEN E, long flag)
 {
-  GEN K = NULL;
-  GEN T2, T3, T5, T7, T13;
+  GEN K = NULL, T2 = NULL, T3 = NULL, T5, T7, T13;
+  ulong good;
   long n2, n3, n5, n7, n13;
-  GEN jt, jtp, s0;
-  GEN c4 = ell_get_c4(E), c6 = ell_get_c6(E), j = ell_get_j(E);
+  GEN jt, jtp, s0, j = ell_get_j(E);
   long l = ellQ_exceptional_iso(j, &jt, &jtp, &s0);
   if (l)
   {
 #if 1
-      return mkisomat(l, ellisograph_dummy(E, l, jt, jtp, s0, flag));
+    return mkisomat(l, ellisograph_dummy(E, l, jt, jtp, s0, flag));
 #else
-      return mkisomat(l, ellisograph_p(K, E, l), flag);
+    return mkisomat(l, ellisograph_p(K, E, l), flag);
 #endif
   }
-  T2 = ellisograph_p(K, E, 2, flag);
-  n2 = etree_nbnodes(T2);
-  if (n2>4 || gequalgs(j, 1728) || gequalgs(j, 287496))
-    return mkisomat(2, T2);
-  T3 = ellisograph_p(K, E, 3, flag);
-  n3 = etree_nbnodes(T3);
-  if (n3>1 && n2==2) return mkisomatdbl(3,T3,2,T2, flag);
-  if (n3==2 && n2>1)  return mkisomatdbl(2,T2,3,T3, flag);
-  if (n3>2 || gequal0(j)) return mkisomat(3, T3);
-  T5 = ellisograph_p(K, E, 5, flag);
-  n5 = etree_nbnodes(T5);
-  if (n5>1 && n2>1) return mkisomatdbl(2,T2,5,T5, flag);
-  if (n5>1 && n3>1) return mkisomatdbl(3,T3,5,T5, flag);
-  if (n5>1) return mkisomat(5, T5);
-  T7 = ellisograph_p(K, E, 7, flag);
-  n7 = etree_nbnodes(T7);
-  if (n7>1 && n2>1) return mkisomatdbl(2,T2,7,T7, flag);
-  if (n7>1 && n3>1) return mkisomatdbl(3,T3,7,T7, flag);
-  if (n7>1) return mkisomat(7,T7);
+  good = ellQ_goodl(ellintegralmodel(E,NULL));
+  if (good & _2)
+  {
+    T2 = ellisograph_p(K, E, 2, flag);
+    n2 = etree_nbnodes(T2);
+    if (n2>4 || gequalgs(j, 1728) || gequalgs(j, 287496))
+      return mkisomat(2, T2);
+  } else n2 = 1;
+  if (good & _3)
+  {
+    T3 = ellisograph_p(K, E, 3, flag);
+    n3 = etree_nbnodes(T3);
+    if (n3>1 && n2==2) return mkisomatdbl(3,T3,2,T2, flag);
+    if (n3==2 && n2>1)  return mkisomatdbl(2,T2,3,T3, flag);
+    if (n3>2 || gequal0(j)) return mkisomat(3, T3);
+  } else n3 = 1;
+  if (good & _5)
+  {
+    T5 = ellisograph_p(K, E, 5, flag);
+    n5 = etree_nbnodes(T5);
+    if (n5>1 && n2>1) return mkisomatdbl(2,T2,5,T5, flag);
+    if (n5>1 && n3>1) return mkisomatdbl(3,T3,5,T5, flag);
+    if (n5>1) return mkisomat(5, T5);
+  } else n5 = 1;
+  if (good & _7)
+  {
+    T7 = ellisograph_p(K, E, 7, flag);
+    n7 = etree_nbnodes(T7);
+    if (n7>1 && n2>1) return mkisomatdbl(2,T2,7,T7, flag);
+    if (n7>1 && n3>1) return mkisomatdbl(3,T3,7,T7, flag);
+    if (n7>1) return mkisomat(7,T7);
+  } else n7 = 1;
   if (n2>1) return mkisomat(2,T2);
   if (n3>1) return mkisomat(3,T3);
-  T13 = ellisograph_p(K, E, 13, flag);
-  n13 = etree_nbnodes(T13);
-  if (n13>1) return mkisomat(13,T13);
-  if (flag)
-    retmkvec2(list_to_crv(mkvec(mkvec3(c4, c6, j))), matid(1));
-  else
-    retmkvec2(list_to_crv(mkvec(mkvec5(c4, c6, j, isogeny_a4a6(E), invisogeny_a4a6(E)))), matid(1));
+  if (good & _13)
+  {
+    T13 = ellisograph_p(K, E, 13, flag);
+    n13 = etree_nbnodes(T13);
+    if (n13>1) return mkisomat(13,T13);
+  } else n13 = 1;
+  return mkvec2(mkvec(ellisograph_a4a6(E,flag)), matid(1));
+}
+
+static long
+fill_LM(GEN LM, GEN L, GEN M, GEN z, long k)
+{
+  GEN Li = gel(LM,1), Mi1 = gmael(LM,2,1);
+  long j, m = lg(Li);
+  for (j = 2; j < m; j++)
+  {
+    GEN d = gel(Mi1,j);
+    gel(L, k) = gel(Li,j);
+    gel(M, k) = z? mulii(d,z): d;
+    k++;
+  }
+  return k;
+}
+static GEN
+ellnf_isocrv(GEN nf, GEN E, GEN v, GEN PE, long flag)
+{
+  long i, l, lv, n, k;
+  GEN L, M, LE = cgetg_copy(v,&lv), e = ellisograph_a4a6(E, flag);
+  for (i = n = 1; i < lv; i++)
+  {
+    ulong p = uel(v,i);
+    GEN T = isograph_p(nf, e, p, gel(PE,i), flag);
+    GEN LM = nfmkisomat(nf, p, T);
+    gel(LE,i) = LM;
+    n *= lg(gel(LM,1)) - 1;
+  }
+  L = cgetg(n+1,t_VEC); gel(L,1) = e;
+  M = cgetg(n+1,t_COL); gel(M,1) = gen_1;
+  for (i = 1, k = 2; i < lv; i++)
+  {
+    ulong p = uel(v,i);
+    GEN P = gel(PE,i);
+    long kk = k;
+    k = fill_LM(gel(LE,i), L, M, NULL, k);
+    for (l = 2; l < kk; l++)
+    {
+      GEN T = isograph_p(nf, gel(L,l), p, P, flag);
+      GEN LMe = nfmkisomat(nf, p, T);
+      k = fill_LM(LMe, L, M, gel(M,l), k);
+    }
+  }
+  return mkvec2(L, M);
+}
+
+static long
+nfispower(GEN nf, long d, GEN a, GEN b)
+{
+  GEN N;
+  if (gequal(a,b)) return 1;
+  N = nfroots(nf, gsub(monomial(b, d, 0), monomial(a,0,0)));
+  return lg(N) > 1;
+}
+
+static long
+isomat_eq(GEN nf, GEN e1, GEN e2)
+{
+  if (gequal(e1,e2)) return 1;
+  if (!gequal(gel(e1,3), gel(e2,3))) return 0;
+  if (gequal0(gel(e1,3)))
+    return nfispower(nf,6,gel(e1,2),gel(e2,2));
+  if (gequalgs(gel(e1,3),1728))
+    return nfispower(nf,4,gel(e1,1),gel(e2,1));
+  return nfispower(nf,2,gmul(gel(e1,1),gel(e2,2)),gmul(gel(e1,2),gel(e2,1)));
+}
+
+static long
+isomat_find(GEN nf, GEN e, GEN L)
+{
+  long i, l = lg(L);
+  for (i=1; i<l; i++)
+    if (isomat_eq(nf, e, gel(L,i))) return i;
+  pari_err_BUG("isomat_find"); return 0; /* LCOV_EXCL_LINE */
+}
+
+static GEN
+isomat_perm(GEN nf, GEN E, GEN L)
+{
+  long i, l = lg(E);
+  GEN v = cgetg(l, t_VECSMALL);
+  for (i=1; i<l; i++)
+    uel(v, i) = isomat_find(nf, gel(E,i), L);
+  return v;
+}
+
+static GEN
+ellnf_modpoly(GEN v)
+{
+  long i, l = lg(v);
+  GEN P = cgetg(l, t_VEC);
+  for(i = 1; i < l; i++) gel(P, i) = get_polmodular(v[i]);
+  return P;
+}
+
+static GEN
+ellnf_isomat(GEN E, long flag)
+{
+  GEN nf = ellnf_get_nf(E);
+  GEN v = ellnf_prime_degree(E);
+  GEN P = ellnf_modpoly(v);
+  GEN LM = ellnf_isocrv(nf, E, v, P, flag), L = gel(LM,1), M = gel(LM,2);
+  long i, l = lg(L);
+  GEN R = cgetg(l, t_MAT);
+  gel(R,1) = M;
+  for(i = 2; i < l; i++)
+  {
+    GEN Li = gel(L,i);
+    GEN e = mkvec2(gdivgs(gel(Li,1), -48), gdivgs(gel(Li,2), -864));
+    GEN LMi = ellnf_isocrv(nf, ellinit(e, nf, DEFAULTPREC), v, P, 1);
+    GEN LLi = gel(LMi, 1), Mi = gel(LMi, 2);
+    GEN r = isomat_perm(nf, L, LLi);
+    gel(R,i) = vecpermute(Mi, r);
+  }
+  return mkvec2(L, R);
+}
+
+static GEN
+list_to_crv(GEN L)
+{
+  long i, l;
+  GEN V = cgetg_copy(L, &l);
+  for (i=1; i < l; i++)
+  {
+    GEN Li = gel(L,i);
+    GEN e = mkvec2(gdivgs(gel(Li,1), -48), gdivgs(gel(Li,2), -864));
+    gel(V,i) = lg(Li)==6 ? mkvec3(e, gel(Li,4), gel(Li,5)): e;
+  }
+  return V;
 }
 
 GEN
-ellisomat(GEN E, long flag)
+ellisomat(GEN E, long p, long flag)
 {
   pari_sp av = avma;
-  GEN LM;
-  checkell_Q(E);
+  GEN r = NULL, nf = NULL;
+  long good = 1;
   if (flag < 0 || flag > 1) pari_err_FLAG("ellisomat");
-  LM = ellQ_isomat(E, flag);
-  return gerepilecopy(av, LM);
+  if (p < 0) pari_err_PRIME("ellisomat", utoi(p));
+  if (p == 1) { flag = 1; p = 0; } /* for backward compatibility */
+  checkell(E);
+  switch(ell_get_type(E))
+  {
+    case t_ELL_Q:
+      if (p) good = ellQ_goodl_l(E, p);
+      break;
+    case t_ELL_NF:
+      if (p) good = ellnf_goodl_l(E, mkvecsmall(p));
+      nf = ellnf_get_nf(E);
+      break;
+    default: pari_err_TYPE("ellisomat",E);
+  }
+  if (!good) r = mkvec2(mkvec(ellisograph_a4a6(E, flag)),matid(1));
+  else
+  {
+    if (p)
+      r = nfmkisomat(nf, p, ellisograph_p(nf, E, p, flag));
+    else
+      r = nf? ellnf_isomat(E, flag): ellQ_isomat(E, flag);
+    gel(r,1) = list_to_crv(gel(r,1));
+  }
+  return gerepilecopy(av, r);
+}
+
+static GEN
+get_isomat(GEN v)
+{
+  GEN M, vE, wE;
+  long i, l;
+  if (typ(v) != t_VEC) return NULL;
+  if (checkell_i(v))
+  {
+    v = ellisomat(v,0,1);
+    wE = gel(v,1); l = lg(wE);
+    M  = gel(v,2);
+  }
+  else
+  {
+    if (lg(v) != 3) return NULL;
+    vE = gel(v,1); l = lg(vE);
+    M  = gel(v,2);
+    if (typ(M) != t_MAT || !RgM_is_ZM(M)) return NULL;
+    if (typ(vE) != t_VEC || l == 1) return NULL;
+    if (lg(gel(vE,1)) == 3) wE = shallowcopy(vE);
+    else
+    { /* [[a4,a6],f,g] */
+      wE = cgetg_copy(vE,&l);
+      for (i = 1; i < l; i++) gel(wE,i) = gel(gel(vE,i),1);
+    }
+  }
+  /* wE a vector of [a4,a6] */
+  for (i = 1; i < l; i++)
+  {
+    GEN e = ellinit(gel(wE,i), gen_1, 0), E = ellminimalmodel(e, NULL);
+    obj_free(e); gel(wE,i) = E;
+  }
+  return mkvec2(wE, M);
+}
+
+GEN
+ellweilcurve(GEN E, GEN *ms)
+{
+  pari_sp av = avma;
+  GEN vE = get_isomat(E), vL, Wx, W, XPM, Lf, Cf;
+  long i, l;
+
+  if (!vE) pari_err_TYPE("ellweilcurve",E);
+  vE = gel(vE,1); l = lg(vE);
+  Wx = msfromell(vE, 0);
+  W = gel(Wx,1);
+  XPM = gel(Wx,2);
+  /* lattice attached to the Weil curve in the isogeny class */
+  Lf = mslattice(W, gmael(XPM,1,3));
+  Cf = ginv(Lf); /* left-inverse */
+  vL = cgetg(l, t_VEC);
+  for (i=1; i < l; i++)
+  {
+    GEN c, Ce, Le = gmael(XPM,i,3);
+    Ce = Q_primitive_part(RgM_mul(Cf, Le), &c);
+    Ce = ZM_snf(Ce);
+    if (c) { Ce = ZC_Q_mul(Ce,c); settyp(Ce,t_VEC); }
+    gel(vL,i) = Ce;
+  }
+  for (i = 1; i < l; i++) obj_free(gel(vE,i));
+  vE = mkvec2(vE, vL);
+  if (!ms) return gerepilecopy(av, vE);
+  *ms = Wx; gerepileall(av, 2, &vE, ms); return vE;
+}
+
+GEN
+ellisotree(GEN E)
+{
+  pari_sp av = avma;
+  GEN L = get_isomat(E), vE, adj, M;
+  long i, j, n;
+  if (!L) pari_err_TYPE("ellisotree",E);
+  vE = gel(L,1);
+  adj = gel(L,2);
+  n = lg(vE)-1; L = cgetg(n+1, t_VEC);
+  for (i = 1; i <= n; i++) gel(L,i) = ellR_area(gel(vE,i), LOWDEFAULTPREC);
+  M = zeromatcopy(n,n);
+  for (i = 1; i <= n; i++)
+    for (j = i+1; j <= n; j++)
+    {
+      GEN p = gcoeff(adj,i,j);
+      if (!isprime(p)) continue;
+      /* L[i] / L[j] = p or 1/p; p iff E[i].lattice \subset E[j].lattice */
+      if (gcmp(gel(L,i), gel(L,j)) > 0)
+        gcoeff(M,i,j) = p;
+      else
+        gcoeff(M,j,i) = p;
+    }
+  for (i = 1; i <= n; i++) obj_free(gel(vE,i));
+  return gerepilecopy(av, mkvec2(vE,M));
 }
