@@ -140,60 +140,9 @@ matsize(GEN x)
     case t_MAT: return mkvec2s(L? nbrows(x): 0, L);
   }
   pari_err_TYPE("matsize",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
-/*******************************************************************/
-/*                                                                 */
-/*                  Conversion t_POL --> t_SER                     */
-/*                                                                 */
-/*******************************************************************/
-static GEN
-greffe_aux(GEN x, long l, long lx, long v)
-{
-  GEN y = cgetg(l,t_SER);
-  long i;
-  if (l <= 2) pari_err_BUG("RgX_to_ser (l <= 2)");
-  y[1] = x[1]; setvalp(y, v);
-  x += v; lx -= v;
-  if (lx > l) {
-    for (i = 2; i < l; i++) gel(y,i) = gel(x,i);
-  } else {
-    for (i = 2; i <lx; i++) gel(y,i) = gel(x,i);
-    for (     ; i < l; i++) gel(y,i) = gen_0;
-  }
-  return normalize(y);
-}
-/* enlarge/truncate t_POL x to a t_SER with lg l */
-GEN
-RgX_to_ser(GEN x, long l)
-{
-  if (lg(x) == 2) return zeroser(varn(x), l-2);
-  return greffe_aux(x, l, lg(x), RgX_val(x));
-}
-GEN
-RgX_to_ser_inexact(GEN x, long l)
-{
-  long i, lx = lg(x);
-  int first = 1;
-  /* analogous to RgX_valrem + normalize */
-  if (lx == 2) return zeroser(varn(x), l-2);
-  i = 2;
-  while (i<lx && gequal0(gel(x,i))) {
-    if (first && !isexactzero(gel(x,i)))
-    {
-      pari_warn(warner,"normalizing a series with 0 leading term");
-      first = 0;
-    }
-    i++;
-  }
-  return greffe_aux(x, l, lx, i - 2);
-}
-GEN
-rfrac_to_ser(GEN x, long l)
-{
-  return gdiv(gel(x,1), RgX_to_ser(gel(x,2), l));
-}
 /*******************************************************************/
 /*                                                                 */
 /*                 CONVERSION GEN --> long                         */
@@ -220,7 +169,7 @@ gtolong(GEN x)
       if (gequal0(gel(x,3))) return gtolong(gel(x,2)); break;
   }
   pari_err_TYPE("gtolong",x);
-  return 0; /* not reached */
+  return 0; /* LCOV_EXCL_LINE */
 }
 
 /*******************************************************************/
@@ -487,7 +436,7 @@ gequalm1(GEN x)
       return s < 0 ? absrnz_equal1(x): 0;
     }
     case t_INTMOD:
-      av=avma; y=equalii(addsi(1,gel(x,2)), gel(x,1)); avma=av; return y;
+      av=avma; y=equalii(addui(1,gel(x,2)), gel(x,1)); avma=av; return y;
 
     case t_FRAC:
       return 0;
@@ -502,7 +451,7 @@ gequalm1(GEN x)
       return gequalm1(gel(x,2)) && gequal0(gel(x,3));
 
     case t_PADIC:
-      av=avma; y=equalii(addsi(1,gel(x,4)), gel(x,3)); avma=av; return y;
+      av=avma; y=equalii(addui(1,gel(x,4)), gel(x,3)); avma=av; return y;
 
     case t_POLMOD:
       av=avma; p1 = gaddgs(gel(x,2), 1);
@@ -692,14 +641,14 @@ gcmp(GEN x, GEN y)
     case t_INFINITY: return inf_get_sign(x);
   }
   pari_err_TYPE2("comparison",x,y);
-  return 0;/*not reached*/
+  return 0;/*LCOV_EXCL_LINE*/
 }
 
 int
 gcmpsg(long s, GEN y)
 {
-  long ty = typ(y);
-  switch(ty) {
+  switch(typ(y))
+  {
     case t_INT:  return cmpsi(s,y);
     case t_REAL: return cmpsr(s,y);
     case t_FRAC: {
@@ -707,10 +656,14 @@ gcmpsg(long s, GEN y)
       GEN n = gel(y,1), d = gel(y,2);
       int f = cmpii(mulsi(s,d), n); avma = av; return f;
     }
-    case t_STR: return -1;
+    case t_QUAD: {
+      pari_sp av = avma;
+      int f = gsigne(gsubsg(s, y)); avma = av; return f;
+    }
+    case t_INFINITY: return -inf_get_sign(y);
   }
   pari_err_TYPE2("comparison",stoi(s),y);
-  return 0; /* not reached */
+  return 0; /* LCOV_EXCL_LINE */
 }
 
 static long
@@ -1144,6 +1097,20 @@ gequalsg(long s, GEN x)
   int f = gequal(stoi(s), x);
   avma = av; return f;
 }
+
+/* a and b are t_INT, t_FRAC, t_REAL or t_COMPLEX of those. Check whether
+ * a-b is invertible */
+int
+cx_approx_equal(GEN a, GEN b)
+{
+  pari_sp av = avma;
+  GEN d;
+  int r;
+  if (a == b) return 1;
+  d = gsub(a,b);
+  r = (gequal0(d) || (typ(d) == t_COMPLEX && gequal0(cxnorm(d))));
+  avma = av; return r;
+}
 /*******************************************************************/
 /*                                                                 */
 /*                          VALUATION                              */
@@ -1307,7 +1274,7 @@ gvaluation(GEN x, GEN p)
       return minval(x,p);
   }
   pari_err_OP("valuation", x,p);
-  return 0; /* not reached */
+  return 0; /* LCOV_EXCL_LINE */
 }
 GEN
 gpvaluation(GEN x, GEN p)
@@ -1406,13 +1373,13 @@ Z_lval(GEN x, ulong p)
   for(vx = 0;;)
   {
     ulong r;
-    GEN q = diviu_rem(x, p, &r);
+    GEN q = absdiviu_rem(x, p, &r);
     if (r) break;
     vx++; x = q;
     if (vx == VAL_DC_THRESHOLD) {
       if (p == 1) pari_err_DOMAIN("Z_lval", "p", "=", gen_1, gen_1);
       vx += Z_pvalrem_DC(x, sqru(p), &x) << 1;
-      q = diviu_rem(x, p, &r); if (!r) vx++;
+      q = absdiviu_rem(x, p, &r); if (!r) vx++;
       break;
     }
   }
@@ -1435,13 +1402,13 @@ Z_lvalrem(GEN x, ulong p, GEN *py)
   for(vx = 0;;)
   {
     ulong r;
-    GEN q = diviu_rem(x, p, &r);
+    GEN q = absdiviu_rem(x, p, &r);
     if (r) break;
     vx++; x = q;
     if (vx == VAL_DC_THRESHOLD) {
       if (p == 1) pari_err_DOMAIN("Z_lvalrem", "p", "=", gen_1, gen_1);
       vx += Z_pvalrem_DC(x, sqru(p), &x) << 1;
-      q = diviu_rem(x, p, &r); if (!r) { vx++; x = q; }
+      q = absdiviu_rem(x, p, &r); if (!r) { vx++; x = q; }
       break;
     }
   }
@@ -1484,7 +1451,7 @@ Z_lvalrem_stop(GEN *n, ulong p, int *stop)
     if (v) *n = utoipos(r);
     return v;
   }
-  av = avma; v = 0; q = diviu_rem(*n, p, &r);
+  av = avma; v = 0; q = absdiviu_rem(*n, p, &r);
   if (r) avma = av;
   else
   {
@@ -1493,10 +1460,10 @@ Z_lvalrem_stop(GEN *n, ulong p, int *stop)
       if (v == VAL_DC_THRESHOLD)
       {
         v += Z_pvalrem_DC(N,sqru(p),&N) << 1;
-        q = diviu_rem(N, p, &r); if (!r) { v++; N = q; }
+        q = absdiviu_rem(N, p, &r); if (!r) { v++; N = q; }
         break;
       }
-      q = diviu_rem(N, p, &r);
+      q = absdiviu_rem(N, p, &r);
     } while (!r);
     *n = N;
   }
@@ -1591,7 +1558,7 @@ gen_z_divides(GEN x, ulong q, long imin)
     ulong r;
     GEN xi = gel(x,i);
     if (!signe(xi)) { gel(y,i) = xi; continue; }
-    gel(y,i) = diviu_rem(xi, q, &r);
+    gel(y,i) = absdiviu_rem(xi, q, &r);
     if (r) { avma = (pari_sp)(y+l); return NULL; }
     affectsign_safe(xi, &gel(y,i));
   }
@@ -1652,7 +1619,7 @@ gen_lval(GEN x, ulong p, long imin)
   for(v = 0;; v++)
     for (i = imin; i < lx; i++)
     {
-      ulong r; gel(y,i) = diviu_rem(gel(y,i), p, &r);
+      ulong r; gel(y,i) = absdiviu_rem(gel(y,i), p, &r);
       if (r) { avma = av; return v; }
     }
 }
@@ -1697,7 +1664,7 @@ ZV_Z_dvd(GEN v, GEN p)
   pari_sp av = avma;
   long i, l = lg(v);
   for (i=1; i<l; i++)
-    if (remii(gel(v,i), p) != gen_0) { avma = av; return 0; }
+    if (!dvdii(gel(v,i), p)) { avma = av; return 0; }
   avma = av; return 1;
 }
 
@@ -1743,7 +1710,7 @@ gen_lvalrem(GEN x, ulong p, GEN *px, long imin)
 
     for (i = imin; i < lx; i++)
     {
-      ulong r; gel(y,i) = diviu_rem(gel(x,i), p, &r);
+      ulong r; gel(y,i) = absdiviu_rem(gel(x,i), p, &r);
       if (r) { *px = x; return v; }
       affectsign_safe(gel(x,i), &gel(y,i));
     }
@@ -1820,9 +1787,8 @@ gneg(GEN x)
       gel(y,2) = gneg(gel(x,2));
       break;
 
-    case t_POLMOD: y=cgetg(3,t_POLMOD);
-      gel(y,1) = RgX_copy(gel(x,1));
-      gel(y,2) = gneg(gel(x,2)); break;
+    case t_POLMOD:
+      retmkpolmod(gneg(gel(x,2)), RgX_copy(gel(x,1)));
 
     case t_RFRAC:
       y = cgetg(3, t_RFRAC);
@@ -1856,7 +1822,7 @@ gneg(GEN x)
     case t_INFINITY: return inf_get_sign(x) == 1? mkmoo(): mkoo();
     default:
       pari_err_TYPE("gneg",x);
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   return y;
 }
@@ -1895,9 +1861,8 @@ gneg_i(GEN x)
       gel(y,3) = gel(x,3);
       gel(y,4) = signe(gel(x,4))? subii(gel(x,3),gel(x,4)): gen_0; break;
 
-    case t_POLMOD: y=cgetg(3,t_POLMOD);
-      gel(y,1) = gel(x,1);
-      gel(y,2) = gneg_i(gel(x,2)); break;
+    case t_POLMOD:
+      retmkpolmod(gneg_i(gel(x,2)), RgX_copy(gel(x,1)));
 
     case t_FFELT: return FF_neg_i(x);
 
@@ -1923,7 +1888,7 @@ gneg_i(GEN x)
 
     default:
       pari_err_TYPE("gneg_i",x);
-      return NULL; /* not reached */
+      return NULL; /* LCOV_EXCL_LINE */
   }
   return y;
 }
@@ -1998,16 +1963,16 @@ gabs(GEN x, long prec)
       return y;
   }
   pari_err_TYPE("gabs",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
-gmax(GEN x, GEN y) { return gcopy(gcmp(x,y)<0? y: x); }
+gmax(GEN x, GEN y) { return gcopy(gmax_shallow(x,y)); }
 GEN
 gmaxgs(GEN x, long s) { return (gcmpsg(s,x)>=0)? stoi(s): gcopy(x); }
 
 GEN
-gmin(GEN x, GEN y) { return gcopy(gcmp(x,y)<0? x: y); }
+gmin(GEN x, GEN y) { return gcopy(gmin_shallow(x,y)); }
 GEN
 gmings(GEN x, long s) { return (gcmpsg(s,x)>0)? gcopy(x): stoi(s); }
 
@@ -2029,7 +1994,7 @@ vecindexmax(GEN x)
       return vecsmall_indexmax(x);
     default: pari_err_TYPE("vecindexmax",x);
   }
-  /* NOT REACHED */
+  /* LCOV_EXCL_LINE */
   return 0;
 }
 long
@@ -2050,30 +2015,26 @@ vecindexmin(GEN x)
       return vecsmall_indexmin(x);
     default: pari_err_TYPE("vecindexmin",x);
   }
-  /* NOT REACHED */
+  /* LCOV_EXCL_LINE */
   return 0;
 }
 
-
 GEN
 vecmax0(GEN x, GEN *pi)
 {
-  long i0, j0, i, j;
-  GEN s;
+  long i, lx = lg(x), tx = typ(x);
+  if (!is_matvec_t(tx) && tx != t_VECSMALL) return gcopy(x);
+  if (lx==1) pari_err_DOMAIN("vecmax", "empty argument", "=", x,x);
   switch(typ(x))
   {
     case t_VEC: case t_COL:
-      i = vecindexmax(x);
-      if (pi) *pi = utoipos(i);
-      return gcopy(gel(x, i));
+      i = vecindexmax(x); if (pi) *pi = utoipos(i);
+      return gcopy(gel(x,i));
     case t_MAT: {
-      long lx2, lx = lg(x);
-      if (lx==1 || (lx2 = lgcols(x)) == 1)
-      {
-        pari_err_DOMAIN("vecmax", "empty argument", "=", x,x);
-        return NULL;/*not reached*/
-      }
-      s = gcoeff(x,i0=1,j0=1); i = 2;
+      long j, i0 = 1, j0 = 1, lx2 = lgcols(x);
+      GEN s;
+      if (lx2 == 1) pari_err_DOMAIN("vecmax", "empty argument", "=", x,x);
+      s = gcoeff(x,i0,j0); i = 2;
       for (j=1; j<lx; j++,i=1)
       {
         GEN c = gel(x,j);
@@ -2084,32 +2045,27 @@ vecmax0(GEN x, GEN *pi)
       return gcopy(s);
     }
     case t_VECSMALL:
-      i = vecsmall_indexmax(x);
-      if (pi) *pi = utoipos(i);
+      i = vecsmall_indexmax(x); if (pi) *pi = utoipos(i);
       return stoi(x[i]);
-    default:
-      return gcopy(x);
   }
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 GEN
 vecmin0(GEN x, GEN *pi)
 {
-  long i0, j0, i, j;
-  GEN s;
+  long i, lx = lg(x), tx = typ(x);
+  if (!is_matvec_t(tx) && tx != t_VECSMALL) return gcopy(x);
+  if (lx==1) pari_err_DOMAIN("vecmin", "empty argument", "=", x,x);
   switch(typ(x))
   {
     case t_VEC: case t_COL:
-      i = vecindexmin(x);
-      if (pi) *pi = utoipos(i);
-      return gcopy(gel(x, i));
+      i = vecindexmin(x); if (pi) *pi = utoipos(i);
+      return gcopy(gel(x,i));
     case t_MAT: {
-      long lx2, lx = lg(x);
-      if (lx==1 || (lx2 = lgcols(x)) == 1)
-      {
-        pari_err_DOMAIN("vecmin", "empty argument", "=", x,x);
-        return NULL;/*not reached*/
-      }
-      s = gcoeff(x,i0=1,j0=1); i = 2;
+      long j, i0 = 1, j0 = 1, lx2 = lgcols(x);
+      GEN s;
+      if (lx2 == 1) pari_err_DOMAIN("vecmin", "empty argument", "=", x,x);
+      s = gcoeff(x,i0,j0); i = 2;
       for (j=1; j<lx; j++,i=1)
       {
         GEN c = gel(x,j);
@@ -2120,12 +2076,10 @@ vecmin0(GEN x, GEN *pi)
       return gcopy(s);
     }
     case t_VECSMALL:
-      i = vecsmall_indexmin(x);
-      if (pi) *pi = utoipos(i);
+      i = vecsmall_indexmin(x); if (pi) *pi = utoipos(i);
       return stoi(x[i]);
-    default:
-      return gcopy(x);
   }
+  return NULL;/*LCOV_EXCL_LINE*/
 }
 
 GEN
@@ -2468,7 +2422,7 @@ cvtop2(GEN x, GEN y)
     case t_QUAD:    return qtop(x, p, d);
   }
   pari_err_TYPE("cvtop2",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /* assume is_const_t(tx) */
@@ -2521,7 +2475,7 @@ cvtop(GEN x, GEN p, long d)
     case t_QUAD: return qtop(x, p, d);
   }
   pari_err_TYPE("cvtop",x);
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
@@ -2545,7 +2499,7 @@ gcvtop(GEN x, GEN p, long r)
 }
 
 long
-gexpo(GEN x)
+gexpo_safe(GEN x)
 {
   long tx = typ(x), lx, e, f, i;
 
@@ -2579,8 +2533,20 @@ gexpo(GEN x)
       for (i=1; i<lx; i++) { e=gexpo(gel(x,i)); if (e>f) f=e; }
       return f;
   }
-  pari_err_TYPE("gexpo",x);
-  return 0; /* not reached */
+  return -1-(long)HIGHEXPOBIT;
+}
+long
+gexpo(GEN x)
+{
+  long e = gexpo_safe(x);
+  if (e < -(long)HIGHEXPOBIT) pari_err_TYPE("gexpo",x);
+  return e;
+}
+GEN
+gpexponent(GEN x)
+{
+  long e = gexpo(x);
+  return e == -(long)HIGHEXPOBIT? mkmoo(): stoi(e);
 }
 
 long
@@ -2709,7 +2675,7 @@ gsigne(GEN x)
     case t_INFINITY: return inf_get_sign(x);
   }
   pari_err_TYPE("gsigne",x);
-  return 0; /* not reached */
+  return 0; /* LCOV_EXCL_LINE */
 }
 
 /*******************************************************************/
