@@ -105,7 +105,7 @@ ZpXXQ_invsqrt(GEN S, GEN T, ulong p, long e)
     q = powuu(p,n); q2 = powuu(p,n2);
     f = RgX_sub(FpXXQ_mul(FpXX_red(S, q), FpXXQ_sqr(a, T, q), T, q), pol_1(v));
     fq = ZXX_Z_divexact(f, q2);
-    q22 = shifti(addis(q2,1),-1);
+    q22 = shifti(addiu(q2,1),-1);
     afq = FpXX_Fp_mul(FpXXQ_mul(a, fq, T, q2), q22, q2);
     a = RgX_sub(a, ZXX_Z_mul(afq, q2));
     if (gc_needed(av2,1))
@@ -230,7 +230,7 @@ diff_red(GEN s, GEN A, long m, GEN T, GEN p)
     GEN rS = cgetg(l+1,t_VEC);
     for (i = l-1; i >=0 ; i--)
       gel(rS,i+1) = to_ZX(gel(sQ, 1+v+l-i), vT);
-    rS = FpX_fromdigits(rS,T,p);
+    rS = FpXV_FpX_fromdigits(rS,T,p);
     gel(qS,2) = FpX_add(FpX_mul(rS, T, p), gel(qS, 2), p);
     if (DEBUGLEVEL>1) timer_printf(&ti,"redadd");
   }
@@ -269,24 +269,12 @@ ZX_to_padic(GEN P, GEN q)
 }
 
 static GEN
-ZXC_to_padic(GEN C, GEN q)
-{
-  long i, l = lg(C);
-  GEN V = cgetg(l,t_COL);
-  for(i = 1; i < l; i++)
-    gel(V, i) = ZX_to_padic(gel(C, i), q);
-  return V;
-}
+ZXC_to_padic(GEN x, GEN q)
+{ pari_APPLY_type(t_COL, ZX_to_padic(gel(x, i), q)) }
 
 static GEN
-ZXM_to_padic(GEN M, GEN q)
-{
-  long i, l = lg(M);
-  GEN V = cgetg(l,t_MAT);
-  for(i = 1; i < l; i++)
-    gel(V, i) = ZXC_to_padic(gel(M, i), q);
-  return V;
-}
+ZXM_to_padic(GEN x, GEN q)
+{ pari_APPLY_same(ZXC_to_padic(gel(x, i), q)) }
 
 static GEN
 ZlX_hyperellpadicfrobenius(GEN H, ulong p, long n)
@@ -303,7 +291,7 @@ ZlX_hyperellpadicfrobenius(GEN H, ulong p, long n)
   if (n < 1)
     pari_err_DOMAIN("hyperellpadicfrobenius","n","<", gen_1, utoi(n));
   k = get_basis(p, d);
-  N = n + logint(stoi(2*n), stoi(p)) + 1;
+  N = n + ulogint(2*n, p) + 1;
   pN1 = powuu(p,N+1);
   Q = RgX_to_FpX(H, pN1);
   if (dvdiu(leading_coeff(Q),p)) is_sing(H, p);
@@ -431,7 +419,7 @@ ZpXQXXQ_invsqrt(GEN F, GEN S, GEN T, ulong p, long e)
     av3 = avma;
     f = RgX_sub(ZpXQXXQ_mul(F, ZpXQXXQ_sqr(a, S, T, q, pp, n), S, T, q, pp, n), pol_1(v));
     fq = gerepileupto(av3, RgX_Rg_divexact(f, q2));
-    q22 = shifti(addis(q2,1),-1);
+    q22 = shifti(addiu(q2,1),-1);
     afq = FpXXX_Fp_mul(ZpXQXXQ_mul(a, fq, S, T, q2, pp, n2), q22, q2);
     a = RgX_sub(a, RgX_Rg_mul(afq, q2));
     if (gc_needed(av2,1))
@@ -524,7 +512,7 @@ Fq_diff_red(GEN s, GEN A, long m, GEN S, GEN T, GEN q, GEN p, long e)
     GEN rS = cgetg(l+1,t_VEC);
     for (i = l-1; i >=0 ; i--)
       gel(rS,i+1) = gel(sQ, 1+v+l-i);
-    rS = FpXQX_fromdigits(rS,S,T,q);
+    rS = FpXQXV_FpXQX_fromdigits(rS, S, T, q);
     gel(qS,2) = FpXX_add(FpXQX_mul(rS, S, T, q), gel(qS, 2), q);
     if (DEBUGLEVEL>1) timer_printf(&ti,"redadd");
   }
@@ -578,7 +566,7 @@ ZlXQX_hyperellpadicfrobenius(GEN H, GEN T, ulong p, long n)
   if (n < 1)
     pari_err_DOMAIN("hyperellpadicfrobenius","n","<", gen_1, utoi(n));
   k = get_basis(p, d); pp = utoi(p);
-  N = n + logint(stoi(2*n), pp) + 1;
+  N = n + ulogint(2*n, p) + 1;
   q = powuu(p,n); N1 = N+1;
   pN1 = powuu(p,N1); T = FpX_get_red(T, pN1);
   Q = RgX_to_FqX(H, T, pN1);
@@ -635,23 +623,6 @@ F2x_genus2charpoly_naive(GEN P, GEN Q)
   return mkvecsmalln(6, 0UL, 4, 2*a, (b+2*c+a*a)>>1, a, 1UL);
 }
 
-static GEN
-Flx_translate1(GEN P, ulong p)
-{
-  long i, k, n = degpol(P);
-  GEN R = Flx_copy(P);
-  for (i=1; i<=n; i++)
-    for (k=n-i; k<n; k++)
-      uel(R,k+2) = Fl_add(uel(R,k+2), uel(R,k+3), p);
-  return R;
-}
-
-static GEN
-Flx_diff1(GEN P, ulong p)
-{
-  return Flx_sub(Flx_translate1(P, p), P, p);
-}
-
 static GEN
 Flx_difftable(GEN P, ulong p)
 {
@@ -814,7 +785,7 @@ hyperellcharpoly(GEN PQ)
     GEN b = get_basis(p, d) == 3 ? gen_1 : q;
     GEN pn = powuu(p, n);
     R = FpX_div_by_X_x(R, eps? b: negi(b), pn, NULL);
-    R = FpX_center(R, pn, shifti(pn,-1));
+    R = FpX_center_i(R, pn, shifti(pn,-1));
   }
   R = charpoly_funceq(R, q);
   return gerepilecopy(av, R);
