@@ -707,45 +707,6 @@ lllgramint(GEN x) { return lllall(x, LLL_IM | LLL_GRAM); }
 GEN
 lllgramkerim(GEN x) { return lllall(x, LLL_ALL | LLL_GRAM); }
 
-static GEN
-rescale_to_int(GEN x)
-{
-  long e, i,j, lx, hx, emin;
-  GEN D = gen_1;
-  int exact = 1;
-
-  lx = lg(x); if (lx == 1) return x;
-  hx = lgcols(x);
-  emin = HIGHEXPOBIT;
-  for (j = 1; j < lx; j++)
-    for (i = 1; i < hx; i++)
-    {
-      GEN c = gcoeff(x,i,j);
-      switch(typ(c))
-      {
-        case t_REAL:
-          exact = 0;
-          if (!signe(c)) continue;
-          e = expo(c) - bit_prec(c);
-          break;
-        case t_INT:
-          if (!signe(c)) continue;
-          e = expi(c) + 32;
-          break;
-        case t_FRAC:
-          e = expi(gel(c,1)) - expi(gel(c,2)) + 32;
-          if (exact) D = lcmii(D, gel(c,2));
-          break;
-        default:
-          pari_err_TYPE("rescale_to_int",c);
-          return NULL; /* not reached */
-      }
-      if (e < emin) emin = e;
-    }
-  if (exact) return D == gen_1 ? x: Q_muli_to_int(x, D);
-  return grndtoi(gmul2n(x, -emin), &e);
-}
-
 GEN
 lllfp(GEN x, double D, long flag)
 {
@@ -753,7 +714,7 @@ lllfp(GEN x, double D, long flag)
   pari_sp av = avma;
   GEN h;
   if (n <= 1) return lll_trivial(x,flag);
-  h = ZM_lll(rescale_to_int(x), D, flag);
+  h = ZM_lll(RgM_rescale_to_int(x), D, flag);
   return gerepilecopy(av, h);
 }
 
@@ -776,7 +737,7 @@ qflll0(GEN x, long flag)
     case 8: return lllgen(x);
     default: pari_err_FLAG("qflll");
   }
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 GEN
@@ -792,7 +753,7 @@ qflllgram0(GEN x, long flag)
     case 8: return lllgramgen(x);
     default: pari_err_FLAG("qflllgram");
   }
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
 
 /********************************************************************/
@@ -803,7 +764,8 @@ qflllgram0(GEN x, long flag)
 static GEN
 kerint0(GEN M)
 {
-  GEN U, H = ZM_hnfall(M,&U,1);
+  /* return ZM_lll(M, LLLDFT, LLL_KER); */
+  GEN U, H = ZM_hnflll(M,&U,1);
   long d = lg(M)-lg(H);
   if (!d) return cgetg(1, t_MAT);
   return ZM_lll(vecslice(U,1,d), LLLDFT, LLL_INPLACE);
@@ -828,5 +790,5 @@ matkerint0(GEN M, long flag)
     case 1: return gerepilecopy(av, kerint0(M));
     default: pari_err_FLAG("matkerint");
   }
-  return NULL; /* not reached */
+  return NULL; /* LCOV_EXCL_LINE */
 }
