@@ -1135,7 +1135,7 @@ nextK:
       i = 1; curdeg = deg[ind[1]];
       if (DEBUGLEVEL > 2)
       {
-        err_printf("\n"); timer_printf(&ti, "to find factor %Ps",y);
+        err_printf("\n"); timer_printf(&ti, "to find factor %Ps",gel(fa,cnt-1));
         err_printf("remaining modular factor(s): %ld\n", lfamod);
       }
       continue;
@@ -1420,9 +1420,8 @@ nf_LLL_cmbf(nfcmbf_t *T, long rec)
   ZERO = zeromat(n0, dnf);
 
   av = avma;
-  TT = cgetg(n0+1, t_VEC);
+  TT = const_vec(n0, NULL);
   Tra  = cgetg(n0+1, t_MAT);
-  for (i=1; i<=n0; i++) TT[i] = 0;
   CM_L = scalarmat_s(C, n0);
   /* tmax = current number of traces used (and computed so far) */
   for(tmax = 0;; tmax++)
@@ -1446,7 +1445,7 @@ nf_LLL_cmbf(nfcmbf_t *T, long rec)
       bestlift_init((L->k)<<1, T->nf, Btra, L);
       polred = ZqX_normalize(T->polbase, lP, L);
       famod = ZqX_liftfact(polred, famod, L->Tpk, L->pk, L->p, L->k);
-      for (i=1; i<=n0; i++) TT[i] = 0;
+      for (i=1; i<=n0; i++) gel(TT,i) = NULL;
     }
     for (i=1; i<=n0; i++)
     {
@@ -1565,8 +1564,10 @@ nf_combine_factors(nfcmbf_t *T, GEN polred, long klim)
     GEN v;
     if (l > 1)
     {
+      GEN den;
       T->pol = gel(res,l);
-      T->polbase = RgX_to_nfX(T->nf, T->pol);
+      T->polbase = Q_remove_denom(RgX_to_nfX(T->nf, T->pol), &den);
+      if (den) { T->Br = gmul(T->Br, den); T->pol = RgX_Rg_mul(T->pol, den); }
     }
     v = nf_LLL_cmbf(T, maxK);
     /* remove last elt, possibly unfactored. Add all new ones. */
