@@ -50,7 +50,7 @@ rem_pol_scal(GEN x, GEN y)
 {
   pari_sp av = avma;
   if (gequal0(y)) pari_err_INV("grem",y);
-  return gerepileupto(av, simplify(gmul(RgX_get_0(x),y)));
+  return gerepileupto(av, simplify(gmul(Rg_get_0(x),y)));
 }
 /* x "scalar", y pol, return x % y with type info */
 static GEN
@@ -59,7 +59,7 @@ rem_scal_pol(GEN x, GEN y)
   if (degpol(y))
   {
     if (!signe(y)) pari_err_INV("grem",y);
-    return gmul(x, RgX_get_1(y));
+    return gmul(x, Rg_get_1(y));
   }
   y = gel(y,2); return rem_scal_scal(x,y);
 }
@@ -93,10 +93,10 @@ poldivrem(GEN x, GEN y, GEN *pr)
       *pr = z; return gdiv(x,y);
     }
     if (!signe(y)) pari_err_INV("poldivrem",y);
-    if (!pr || pr == ONLY_DIVIDES) return gequal0(x)? RgX_get_0(y): NULL;
-    z = gmul(x, RgX_get_1(y));
+    if (!pr || pr == ONLY_DIVIDES) return gequal0(x)? Rg_get_0(y): NULL;
+    z = gmul(x, Rg_get_1(y));
     if (pr == ONLY_REM) return z;
-    *pr = z; return RgX_get_0(y);
+    *pr = z; return Rg_get_0(y);
   }
   return RgX_divrem(x,y,pr);
 }
@@ -112,7 +112,7 @@ gdeuc(GEN x, GEN y)
   { /* x "scalar" */
     if (!signe(y)) pari_err_INV("gdeuc",y);
     if (!degpol(y)) return gdiv(x, gel(y,2)); /* constant */
-    return RgX_get_0(y);
+    return Rg_get_0(y);
   }
   return RgX_div(x,y);
 }
@@ -135,8 +135,8 @@ grem(GEN x, GEN y)
         case t_POLMOD: return polmod_mod(x,y);
         case t_POL: return RgX_rem(x,y);
         case t_RFRAC:
-          av = avma; z = gmul(gel(x,1), RgXQ_inv(gel(x,2),y));
-          return gerepileupto(av, grem(z,y));
+          av = avma; z = RgXQ_inv(RgX_rem(gel(x,2), y), y);
+          return gerepileupto(av, grem(gmul(gel(x,1), z), y));
         case t_SER:
           if (RgX_is_monomial(y))
           {
@@ -247,7 +247,7 @@ ZV_to_ZpV(GEN z, GEN p, long r)
 {
   long i, l = lg(z);
   GEN Z = cgetg(l, typ(z)), q = powiu(p, r);
-  for (i=1; i<lg(z); i++) gel(Z,i) = Z_to_Zp(gel(z,i),p,q,r);
+  for (i=1; i<l; i++) gel(Z,i) = Z_to_Zp(gel(z,i),p,q,r);
   return Z;
 }
 /* shallow */
@@ -256,7 +256,7 @@ ZX_to_ZpX(GEN z, GEN p, GEN q, long r)
 {
   long i, l = lg(z);
   GEN Z = cgetg(l, t_POL); Z[1] = z[1];
-  for (i=2; i<lg(z); i++) gel(Z,i) = Z_to_Zp(gel(z,i),p,q,r);
+  for (i=2; i<l; i++) gel(Z,i) = Z_to_Zp(gel(z,i),p,q,r);
   return Z;
 }
 /* return (x + O(p^r)) normalized (multiply by a unit such that leading coeff
@@ -350,7 +350,7 @@ Zp_appr(GEN f, GEN a)
 
   f = QpX_to_ZX(f, p);
   if (degpol(f) <= 0) pari_err_CONSTPOL("Zp_appr");
-  (void)ZX_gcd_all(f, ZX_deriv(f), &f);
+  f = ZX_radical(f);
   a = padic_to_Q(a);
   if (signe(FpX_eval(f,a,p))) { avma = av; return cgetg(1,t_COL); }
   z = ZX_Zp_root(f, a, p, prec);
@@ -363,7 +363,7 @@ ZX_Zp_roots(GEN f, GEN p, long prec)
   GEN y, z, rac;
   long lx, i, j, k;
 
-  (void)ZX_gcd_all(f, ZX_deriv(f), &f);
+  f = ZX_radical(f);
   rac = FpX_roots(f, p);
   lx = lg(rac); if (lx == 1) return rac;
   y = cgetg(degpol(f)+1,t_COL);
@@ -521,13 +521,31 @@ cmp_padic(GEN x, GEN y)
   return cmpii(gel(x,4), gel(y,4));
 }
 
+/* replace p^e by p*...*p [ factors are not known to be equal, only close at
+ * input accuracy ] */
+static GEN
+famat_flatten(GEN fa)
+{
+  GEN y, P = gel(fa,1), E = gel(fa,2);
+  long i, l = lg(E);
+  y = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
+  {
+    GEN p = gel(P,i);
+    long e = itou(gel(E,i));
+    gel(y,i) = const_vec(e, p);
+  }
+  y = shallowconcat1(y); settyp(y, t_COL);
+  return mkmat2(y, const_col(lg(y)-1, gen_1));
+}
+
 GEN
 factorpadic(GEN f, GEN p, long r)
 {
   pari_sp av = avma;
   GEN y, ppow;
   long v, n;
-  int reverse = 0;
+  int reverse = 0, exact;
 
   if (typ(f)!=t_POL) pari_err_TYPE("factorpadic",f);
   if (typ(p)!=t_INT) pari_err_TYPE("factorpadic",p);
@@ -535,7 +553,8 @@ factorpadic(GEN f, GEN p, long r)
   if (!signe(f)) return prime_fact(f);
   if (!degpol(f)) return trivial_fact();
 
-  v = RgX_valrem(f, &f);
+  exact = RgX_is_QX(f); /* before RgX_valrem which may lose type information */
+  v = RgX_valrem_inexact(f, &f);
   ppow = powiu(p,r);
   n = degpol(f);
   if (!n)
@@ -564,5 +583,6 @@ factorpadic(GEN f, GEN p, long r)
     GEN X = ZX_to_ZpX(pol_x(varn(f)), p, ppow, r);
     y = famat_mulpow_shallow(y, X, utoipos(v));
   }
+  if (!exact) y = famat_flatten(y);
   return gerepilecopy(av, sort_factor_pol(y, cmp_padic));
 }
