@@ -31,9 +31,8 @@ typedef struct {
 static void
 init_MR_Jaeschke(MR_Jaeschke_t *S, GEN n)
 {
-  if (signe(n) < 0) n = absi(n);
-  S->n = n;
-  S->t = addsi(-1,n);
+  S->n = n = absi_shallow(n);
+  S->t = subiu(n,1);
   S->r1 = vali(S->t);
   S->t1 = shifti(S->t, -S->r1);
   S->sqrt1 = cgeti(lg(n)); S->sqrt1[1] = evalsigne(0)|evallgefint(2);
@@ -222,27 +221,16 @@ Fl_MR_Jaeschke(ulong n, long k)
 }
 
 int
-MR_Jaeschke(GEN n, long k)
+MR_Jaeschke(GEN n)
 {
-  pari_sp av2, av = avma;
-  const ulong pr[] =
-    { 0, 2,3,5,7,11,13,17,19,23,29, 31,73, 2,13,23,1662803UL, };
-  const ulong *p;
-  long i;
+  pari_sp av = avma;
   MR_Jaeschke_t S;
 
-  if (lgefint(n) == 3) return Fl_MR_Jaeschke(uel(n,2), k);
-
+  if (lgefint(n) == 3) return Fl_MR_Jaeschke(uel(n,2), 17);
   if (!mod2(n)) return 0;
-  if      (k == 16) { p = pr+13; k = 4; } /* 2,13,23,1662803 */
-  else if (k == 17) { p = pr+11; k = 2; } /* 31,73 */
-  else p = pr; /* 2,3,5,... */
-  init_MR_Jaeschke(&S, n); av2 = avma;
-  for (i=1; i<=k; i++)
-  {
-    if (bad_for_base(&S, utoipos(p[i]))) { avma = av; return 0; }
-    avma = av2;
-  }
+  av = avma; init_MR_Jaeschke(&S, n);
+  if (bad_for_base(&S, utoipos(31))) { avma = av; return 0; }
+  if (bad_for_base(&S, utoipos(73))) { avma = av; return 0; }
   avma = av; return 1;
 }
 
@@ -302,6 +290,33 @@ LucasMod(GEN n, ulong P, GEN N)
  * v_{k+2} = P v_{k+1} - v_k, v_0 = 2, v_1 = P.
  * Assume n > 0 */
 static ulong
+u_LucasMod_pre(ulong n, ulong P, ulong N, ulong NI)
+{
+  ulong v, v1, m;
+  long j;
+
+  if (n == 1) return P;
+  j = 1 + bfffo(n); /* < BIL */
+  v = P; v1 = P*P - 2;
+  m = n<<j; j = BITS_IN_LONG - j;
+  for (; j; m<<=1,j--)
+  { /* v = v_k, v1 = v_{k+1} */
+    if (m & HIGHBIT)
+    { /* set v = v_{2k+1}, v1 = v_{2k+2} */
+      v = Fl_sub(Fl_mul_pre(v,v1,N,NI), P, N);
+      v1= Fl_sub(Fl_sqr_pre(v1,N,NI), 2UL, N);
+    }
+    else
+    {/* set v = v_{2k}, v1 = v_{2k+1} */
+      v1= Fl_sub(Fl_mul_pre(v,v1,N,NI),P, N);
+      v = Fl_sub(Fl_sqr_pre(v,N,NI), 2UL, N);
+    }
+  }
+  return v;
+}
+
+/* !(N & HIGHMASK) */
+static ulong
 u_LucasMod(ulong n, ulong P, ulong N)
 {
   ulong v, v1, m;
@@ -315,13 +330,13 @@ u_LucasMod(ulong n, ulong P, ulong N)
   { /* v = v_k, v1 = v_{k+1} */
     if (m & HIGHBIT)
     { /* set v = v_{2k+1}, v1 = v_{2k+2} */
-      v = Fl_sub(Fl_mul(v,v1,N), P, N);
-      v1= Fl_sub(Fl_sqr(v1,N), 2UL, N);
+      v = Fl_sub((v*v1) % N, P, N);
+      v1= Fl_sub((v1*v1)% N, 2UL, N);
     }
     else
     {/* set v = v_{2k}, v1 = v_{2k+1} */
-      v1= Fl_sub(Fl_mul(v,v1,N),P, N);
-      v = Fl_sub(Fl_sqr(v,N), 2UL, N);
+      v1= Fl_sub((v*v1) % N ,P, N);
+      v = Fl_sub((v*v) % N, 2UL, N);
     }
   }
   return v;
@@ -332,7 +347,6 @@ uislucaspsp(ulong n)
 {
   long i, v;
   ulong b, z, m = n + 1;
-
   for (b=3, i=0;; b+=2, i++)
   {
     ulong c = b*b - 4; /* = 1 mod 4 */
@@ -341,13 +355,28 @@ uislucaspsp(ulong n)
   }
   if (!m) return 0; /* neither 2^32-1 nor 2^64-1 are Lucas-pp */
   v = vals(m); m >>= v;
-  z = u_LucasMod(m, b, n);
-  if (z == 2 || z == n-2) return 1;
-  for (i=1; i<v; i++)
+  if (n & HIGHMASK)
   {
-    if (!z) return 1;
-    z = Fl_sub(Fl_sqr(z,n), 2UL, n);
-    if (z == 2) return 0;
+    ulong ni = get_Fl_red(n);
+    z = u_LucasMod_pre(m, b, n, ni);
+    if (z == 2 || z == n-2) return 1;
+    for (i=1; i<v; i++)
+    {
+      if (!z) return 1;
+      z = Fl_sub(Fl_sqr_pre(z,n,ni), 2UL, n);
+      if (z == 2) return 0;
+    }
+  }
+  else
+  {
+    z = u_LucasMod(m, b, n);
+    if (z == 2 || z == n-2) return 1;
+    for (i=1; i<v; i++)
+    {
+      if (!z) return 1;
+      z = Fl_sub((z*z) % n, 2UL, n);
+      if (z == 2) return 0;
+    }
   }
   return 0;
 }
@@ -367,7 +396,7 @@ IsLucasPsP(GEN N)
     if (b == 129 && Z_issquare(N)) return 0; /* avoid oo loop if N = m^2 */
     if (krouu(umodiu(N,c), c) < 0) break;
   }
-  m = addis(N,1); v = vali(m); m = shifti(m,-v);
+  m = addiu(N,1); v = vali(m); m = shifti(m,-v);
   z = LucasMod(m, b, N);
   if (absequaliu(z, 2)) return 1;
   if (equalii(z, subiu(N,2))) return 1;
@@ -390,13 +419,13 @@ static int
 iu_coprime(GEN N, ulong u)
 {
   const ulong n = umodiu(N, u);
-  return (n == 1 || gcduodd(n, u) == 1);
+  return (n == 1 || ugcd(n, u) == 1);
 }
 /* assume u odd, u > 1 */
 static int
 uu_coprime(ulong n, ulong u)
 {
-  return gcduodd(n, u) == 1;
+  return ugcd(n, u) == 1;
 }
 
 /* composite strong 2-pseudoprime < 1016801 whose prime divisors are > 101 */
@@ -493,9 +522,7 @@ uisprime(ulong n)
       !uu_coprime(n, 1673450759UL) ||
       !uu_coprime(n, 4269855901UL)) return 0;
 #endif
-  if (n < 10427) return 1;
-  if (n < 1016801) return !is_2_prp_101(n) && u_2_prp(n);
-  return uBPSW_psp(n);
+  return uisprime_101(n);
 }
 
 /* assume no prime divisor <= 101 */
@@ -503,7 +530,7 @@ int
 uisprime_101(ulong n)
 {
   if (n < 10427) return 1;
-  if (n < 1016801) return !is_2_prp_101(n) && u_2_prp(n);
+  if (n < 1016801) return u_2_prp(n) && !is_2_prp_101(n);
   return uBPSW_psp(n);
 }
 
@@ -610,7 +637,7 @@ BPSW_isprime_small(GEN x)
 static ulong
 pl831(GEN N, GEN p)
 {
-  GEN b, c, g, Nmunp = diviiexact(addis(N,-1), p);
+  GEN b, c, g, Nmunp = diviiexact(subiu(N,1), p);
   pari_sp av = avma;
   ulong a;
   for(a = 2;; a++, avma = av)
@@ -619,7 +646,7 @@ pl831(GEN N, GEN p)
     if (!equali1(b)) break;
   }
   c = Fp_pow(b,p,N);
-  g = gcdii(addis(b,-1), N); /* 0 < g < N */
+  g = gcdii(subiu(b,1), N); /* 0 < g < N */
   return (equali1(c) && equali1(g))? a: 0;
 }
 
@@ -640,7 +667,7 @@ BLS_test(GEN N, GEN f)
 }
 
 /* BPSW_psp(N) && !BPSW_isprime_small(N). Decide between Pocklington-Lehmer
- * and APRCL. Return a vector of (small) primes such that PL-witnesses
+ * and APRCL/ECPP. Return a vector of (small) primes such that PL-witnesses
  * guarantee the primality of N. Return NULL if PL is likely too expensive.
  * Return gen_0 if BLS test finds N to be composite */
 static GEN
@@ -669,22 +696,6 @@ BPSW_try_PL(GEN N)
 }
 
 static GEN isprimePL(GEN N);
-static GEN PL_certificate(GEN N, GEN F);
-/* Assume N a BPSW pseudoprime. Return 0 if not prime, and a primality label
- * otherwise: 1 (small), 2 (APRCL), or PL-certificate  */
-static GEN
-check_prime(GEN N)
-{
-  GEN P;
-  if (BPSW_isprime_small(N)) return gen_1;
-  /* PL for small N: APRCL is faster but we prefer a certificate */
-  if (expi(N) <= 250) return isprimePL(N);
-  P = BPSW_try_PL(N);
-  /* if PL likely too expensive: give up certificate and use APRCL */
-  if (!P) return isprimeAPRCL(N)? gen_2: gen_0;
-  return typ(P) != t_INT? PL_certificate(N,P): gen_0;
-}
-
 /* F is a vector whose entries are primes. For each of them, find a PL
  * witness. Return 0 if caller lied and F contains a composite */
 static long
@@ -695,31 +706,78 @@ PL_certify(GEN N, GEN F)
     if (! pl831(N, gel(F,i))) return 0;
   return 1;
 }
-/* F is a vector whose entries are *believed* to be primes. For each of them,
- * recording a witness and recursive primality certificate */
+/* F is a vector whose entries are *believed* to be primes (BPSW_psp).
+ * For each of them, recording a witness and recursive primality certificate */
 static GEN
 PL_certificate(GEN N, GEN F)
 {
   long i, l = lg(F);
-  GEN W = cgetg(l,t_COL);
-  GEN R = cgetg(l,t_COL);
-  for(i=1; i<l; i++)
+  GEN C;
+  if (BPSW_isprime_small(N)) return N;
+  C = cgetg(l, t_VEC);
+  for (i = 1; i < l; i++)
   {
-    GEN p = gel(F,i);
-    ulong witness = pl831(N,p);
-    if (!witness) return gen_0;
-    gel(W,i) = utoipos(witness);
-    gel(R,i) = check_prime(p);
-    if (isintzero(gel(R,i)))
+    GEN p = gel(F,i), C0;
+    ulong w;
+
+    if (BPSW_isprime_small(p)) { gel(C,i) = p; continue; }
+    w = pl831(N,p); if (!w) return gen_0;
+    C0 = isprimePL(p);
+    if (isintzero(C0))
     { /* composite in prime factorisation ! */
       err_printf("Not a prime: %Ps", p);
       pari_err_BUG("PL_certificate [false prime number]");
     }
+    gel(C,i) = mkvec3(p,utoipos(w), C0);
   }
-  return mkmat3(F, W, R);
+  return mkvec2(N, C);
 }
+/* M a t_MAT */
+static int
+PL_isvalid(GEN v)
+{
+  GEN C, F, F2, N, N1, U;
+  long i, l;
+  switch(typ(v))
+  {
+    case t_INT: return BPSW_isprime_small(v) && BPSW_psp(v);
+    case t_VEC: if (lg(v) == 3) break;/*fall through */
+    default: return 0;
+  }
+  N = gel(v,1);
+  C = gel(v,2);
+  if (typ(N) != t_INT || signe(N) <= 0 || typ(C) != t_VEC) return 0;
+  U = N1 = subiu(N,1);
+  l = lg(C);
+  for (i = 1; i < l; i++)
+  {
+    GEN p = gel(C,i), a = NULL, C0 = NULL, ap;
+    long vp;
+    if (typ(p) != t_INT)
+    {
+      if (lg(p) != 4) return 0;
+      a = gel(p,2); C0 = gel(p,3); p = gel(p,1);
+      if (typ(p) != t_INT || typ(a) != t_INT || !PL_isvalid(C0)) return 0;
+    }
+    vp = Z_pvalrem(U, p, &U); if (!vp) return 0;
+    if (!a)
+    {
+      if (!BPSW_isprime_small(p)) return 0;
+      continue;
+    }
+    if (!equalii(gel(C0,1), p)) return 0;
+    ap = Fp_pow(a, diviiexact(N1,p), N);
+    if (!equali1(gcdii(subiu(ap,1), N)) || !equali1(Fp_pow(ap, p, N))) return 0;
+  }
+  F = diviiexact(N1, U); /* factored part of N-1 */
+  F2= sqri(F);
+  if (cmpii(F2, N) > 0) return 1;
+  if (cmpii(mulii(F2,F), N) <= 0) return 0;
+  return BLS_test(N,F);
+}
+
 /* Assume N is a strong BPSW pseudoprime, Pocklington-Lehmer primality proof.
- * Return gen_0 (non-prime), gen_1 (small prime), matrix (large prime)
+ * Return gen_0 (non-prime), N (small prime), matrix (large prime)
  *
  * The matrix has 3 columns, [a,b,c] with
  * a[i] prime factor of N-1,
@@ -728,18 +786,10 @@ PL_certificate(GEN N, GEN F)
 static GEN
 isprimePL(GEN N)
 {
-  pari_sp ltop = avma;
   GEN cbrtN, N_1, F, f;
-
-  if (typ(N) != t_INT) pari_err_TYPE("isprimePL",N);
-  switch(cmpis(N,2))
-  {
-    case -1:return gen_0;
-    case 0: return gen_1;
-  }
-  /* N > 2 */
-  cbrtN = sqrtnint(N, 3);
-  N_1 = addis(N,-1);
+  if (BPSW_isprime_small(N)) return N;
+  cbrtN = sqrtnint(N,3);
+  N_1 = subiu(N,1);
   F = Z_factor_until(N_1, sqri(cbrtN));
   f = factorback(F); /* factored part of N-1, f^3 > N */
   if (DEBUGLEVEL>3)
@@ -750,8 +800,9 @@ isprimePL(GEN N)
   }
   /* if N-1 is only N^(1/3)-smooth, BLS test */
   if (!equalii(f,N_1) && cmpii(sqri(f),N) <= 0 && !BLS_test(N,f))
-  { avma = ltop; return gen_0; } /* Failed, N is composite */
-  return gerepilecopy(ltop, PL_certificate(N, gel(F,1)));
+    return gen_0; /* Failed, N is composite */
+  F = gel(F,1); settyp(F, t_VEC);
+  return PL_certificate(N, F);
 }
 
 /* assume N a BPSW pseudoprime, in particular, it is odd > 2. Prove N prime */
@@ -763,21 +814,29 @@ BPSW_isprime(GEN N)
   GEN P;
   if (BPSW_isprime_small(N)) return 1;
   av = avma; P = BPSW_try_PL(N);
-  if (!P)
-    t = isprimeAPRCL(N); /* not smooth enough */
+  if (!P) /* not smooth enough */
+    t = expi(N) < 768? isprimeAPRCL(N): isprimeECPP(N);
   else
     t = (typ(P) == t_INT)? 0: PL_certify(N,P);
   avma = av; return t;
 }
 
+static long
+_isprimePL(GEN x)
+{
+  pari_sp av = avma;
+  if (!BPSW_psp(x)) return 0;
+  x = isprimePL(x); avma = av; return !isintzero(x);
+}
 GEN
 gisprime(GEN x, long flag)
 {
   switch (flag)
   {
     case 0: return map_proto_lG(isprime,x);
-    case 1: return map_proto_G(isprimePL,x);
+    case 1: return map_proto_lG(_isprimePL,x);
     case 2: return map_proto_lG(isprimeAPRCL,x);
+    case 3: return map_proto_lG(isprimeECPP,x);
   }
   pari_err_FLAG("gisprime");
   return NULL;
@@ -786,6 +845,85 @@ gisprime(GEN x, long flag)
 long
 isprime(GEN x) { return BPSW_psp(x) && BPSW_isprime(x); }
 
+GEN
+primecert(GEN x, long flag)
+{
+  if (!BPSW_psp(x)) return gen_0;
+  switch(flag)
+  {
+    case 0: return ecpp(x);
+    case 1: { pari_sp av = avma; return gerepilecopy(av, isprimePL(x)); }
+  }
+  pari_err_FLAG("primecert");
+  return NULL;/*LCOV_EXCL_LINE*/
+}
+
+enum { c_VOID = 0, c_ECPP, c_N1 };
+static long
+cert_type(GEN c)
+{
+  switch(typ(c))
+  {
+    case t_INT: return c_ECPP;
+    case t_VEC:
+      if (lg(c) == 3 && typ(gel(c,1)) == t_INT) return c_N1;
+      return c_ECPP;
+  }
+  return c_VOID;
+}
+
+long
+primecertisvalid(GEN c)
+{
+  switch(typ(c))
+  {
+    case t_INT: return BPSW_isprime_small(c) && BPSW_psp(c);
+    case t_VEC:
+      return cert_type(c) == c_ECPP? ecppisvalid(c): PL_isvalid(c);
+  }
+  return 0;
+}
+
+static long
+check_eccpcertentry(GEN c)
+{
+  GEN v;
+  long i,l = lg(c);
+  if (typ(c)!=t_VEC || l!=6) return 0;
+  for(i=1; i<=4; i++)
+    if (typ(gel(c,i))!=t_INT) return 0;
+  v = gel(c,5);
+  if(typ(v)!=t_VEC) return 0;
+  for(i=1; i<=2; i++)
+    if (typ(gel(v,i))!=t_INT) return 0;
+  return 1;
+}
+
+static long
+check_eccpcert(GEN c)
+{
+  long i, l;
+  switch(typ(c))
+  {
+    case t_INT: return signe(c) >= 0;
+    case t_VEC: break;
+    default: return 0;
+  }
+  l = lg(c); if (l == 1) return 0;
+  for (i = 1; i < l; i++)
+    if (check_eccpcertentry(gel(c,i))==0) return 0;
+  return 1;
+}
+
+GEN
+primecertexport(GEN c, long flag)
+{
+  if (cert_type(c) != c_ECPP) pari_err_IMPL("N-1 certificate");
+  if (!check_eccpcert(c))
+    pari_err_TYPE("primecertexport - invalid certificate", c);
+  return ecppexport(c, flag);
+}
+
 /***********************************************************************/
 /**                                                                   **/
 /**                          PRIME NUMBERS                            **/
@@ -828,11 +966,38 @@ static struct {
   {  4000000483UL,189961831},
   {  4222234741UL,200000000},
 #if BITS_IN_LONG == 64
+  { 5336500537UL,   250000000L},
+  { 6461335109UL,   300000000L},
+  { 7594955549UL,   350000000L},
+  { 8736028057UL,   400000000L},
+  { 9883692017UL,   450000000L},
   { 11037271757UL,  500000000L},
+  { 13359555403UL,  600000000L},
+  { 15699342107UL,  700000000L},
+  { 18054236957UL,  800000000L},
+  { 20422213579UL,  900000000L},
   { 22801763489UL, 1000000000L},
   { 47055833459UL, 2000000000L},
+  { 71856445751UL, 3000000000L},
+  { 97011687217UL, 4000000000L},
   {122430513841UL, 5000000000L},
+  {148059109201UL, 6000000000L},
+  {173862636221UL, 7000000000L},
   {200000000507UL, 8007105083L},
+  {225898512559UL, 9000000000L},
+  {252097800623UL,10000000000L},
+  {384489816343UL,15000000000L},
+  {518649879439UL,20000000000L},
+  {654124187867UL,25000000000L},
+  {790645490053UL,30000000000L},
+  {928037044463UL,35000000000L},
+ {1066173339601UL,40000000000L},
+ {1344326694119UL,50000000000L},
+ {1624571841097UL,60000000000L},
+ {1906555030411UL,70000000000L},
+ {2190026988349UL,80000000000L},
+ {2474799787573UL,90000000000L},
+ {2760727302517UL,100000000000L}
 #endif
 };
 static const int prime_table_len = numberof(prime_table);
@@ -982,7 +1147,7 @@ randomprime(GEN N)
       break;
     default:
       pari_err_TYPE("randomprime", N);
-      return NULL; /*notreached*/
+      return NULL; /*LCOV_EXCL_LINE*/
   }
   av2 = avma;
   for (;;)
@@ -1264,6 +1429,24 @@ primes_upto_zv(ulong b)
 /**                       PRIVATE PRIME TABLE                         **/
 /**                                                                   **/
 /***********************************************************************/
+
+static GEN global_primetab;
+void
+pari_init_primetab(void)  { global_primetab = NULL; }
+void
+pari_pthread_init_primetab(void) { global_primetab = primetab; }
+void
+pari_thread_init_primetab(void)
+{
+  if (global_primetab)
+  {
+    long i, l = lg(global_primetab);
+    primetab = cgetg_block(l, t_VEC);
+    for (i = 1; i < l; i++)
+      gel(primetab,i) = gclone(gel(global_primetab,i));
+  } else primetab = cgetg_block(1, t_VEC);
+}
+
 /* delete dummy NULL entries */
 static void
 cleanprimetab(GEN T)
@@ -1331,9 +1514,9 @@ addp(GEN *T, GEN p)
   l = lg(p);
   if (l != lg(*T))
   {
-    GEN old = *T, t = cgetalloc(t_VEC, l);
+    GEN old = *T, t = cgetg_block(l, t_VEC);
     for (i = 1; i < l; i++) gel(t,i) = gel(p,i);
-    *T = t; free(old);
+    *T = t; gunclone(old);
   }
   avma = av; return *T;
 }
