@@ -1444,7 +1444,8 @@ gsqrt(GEN x, long prec)
       y = cgetg(3,t_COMPLEX); av = avma;
 
       r = cxnorm(x);
-      if (typ(r) == t_INTMOD) pari_err_IMPL("sqrt(complex of t_INTMODs)");
+      if (typ(r) == t_INTMOD || typ(r) == t_PADIC)
+        pari_err_IMPL("sqrt(complex of t_INTMODs)");
       r = gsqrt(r, prec); /* t_REAL, |a+Ib| */
       if (!signe(r))
         u = v = gerepileuptoleaf(av, sqrtr(r));
@@ -1868,7 +1869,7 @@ sqrtnof1(ulong n, long prec)
     z = gdiv(y, gaddgs(gmulsg(n1, y), n2));
     shiftc_inplace(z,1);
     x = gmul(x, gsubsg(1, z));
-    if (mask == 1) return gerepileupto(av, gprec_w(x,prec));
+    if (mask == 1) return gerepilecopy(av, gprec_w(x,prec));
     eold = enew;
   }
 }
