@@ -17,15 +17,13 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include <windows.h>
 #include <time.h>
 #include "pari.h"
-#include "paripriv.h"
 #include "rect.h"
 
 static void SetForeground(void *data, long col)
 {
-  int r,g,b;
   HPEN hOldPen;
-  color_to_rgb(gel(GP_DATA->colormap,col+1), &r, &g, &b);
-
+  int r, g, b; long_to_rgb(col, &r, &g, &b);
+  SetDCPenColor((HDC)data,RGB(r,g,b));
   hOldPen = SelectObject((HDC)data, CreatePen(PS_SOLID, 1, RGB(r,g,b)));
   if( hOldPen ) DeleteObject(hOldPen);
 }
@@ -49,6 +47,19 @@ static void DrawRectangle(void *data, long x, long y, long w, long h)
   DrawLine(data, x,y+h,x+w,y+h);
 }
 
+static void FillRectangle(void *data, long x, long y, long w, long h)
+{
+  RECT rc;
+  COLORREF color;
+  HBRUSH brush;
+  rc.left = x; rc.right  = x+w;
+  rc.top  = y; rc.bottom = y+h;
+  color = GetDCPenColor((HDC) data);
+  brush = CreateSolidBrush(color);
+  FillRect((HDC)data, &rc, brush);
+  DeleteObject(brush);
+}
+
 static void DrawPoints(void *data, long nb, struct plot_points *p)
 {
   long i;
@@ -69,12 +80,13 @@ static void DrawString(void *data, long x, long y, char *text, long numtext)
   TextOut((HDC)data, x, y, text, numtext);
 }
 
-void rectdraw0(long *w, long *x, long *y, long lw)
+static void
+draw(PARI_plot *T, GEN w, GEN x, GEN y)
 {
   char tmppath[MAX_PATH], fname[MAX_PATH];
   struct plot_eng plotWin32;
   HDC hEmf;
-  int bg_r,bg_g,bg_b;
+  int r, g, b;
 
   GetTempPath(sizeof(tmppath), tmppath);
   sprintf(fname, "%s\\gp-ploth-%lx.emf", tmppath, time(NULL)/(24*60*60)*1000+GetTickCount());
@@ -82,44 +94,48 @@ void rectdraw0(long *w, long *x, long *y, long lw)
   hEmf = CreateEnhMetaFile(GetDC(NULL), fname, NULL, NULL);
   SetMapMode(hEmf, MM_TEXT);
   SelectObject(hEmf, GetStockObject(DEFAULT_GUI_FONT));
-  color_to_rgb(gel(GP_DATA->colormap,1), &bg_r, &bg_g, &bg_b);
-  SetBkColor(hEmf, RGB(bg_r,bg_g,bg_b));
+  color_to_rgb(gel(GP_DATA->colormap,1), &r,&g,&b);
+  SetBkColor(hEmf, RGB(r,g,b));
   SetBkMode(hEmf, OPAQUE);
 
   plotWin32.sc=&SetForeground;
   plotWin32.pt=&DrawPoint;
   plotWin32.ln=&DrawLine;
   plotWin32.bx=&DrawRectangle;
+  plotWin32.fb=&FillRectangle;
   plotWin32.mp=&DrawPoints;
   plotWin32.ml=&DrawLines;
   plotWin32.st=&DrawString;
-  plotWin32.pl=&pari_plot;
+  plotWin32.pl=T;
   plotWin32.data=(void*)hEmf;
 
-  gen_rectdraw0(&plotWin32, w, x, y, lw, 1, 1);
+  gen_draw(&plotWin32, w, x, y, 1, 1);
   DeleteEnhMetaFile(CloseEnhMetaFile(hEmf));
 
   ShellExecute(NULL,NULL,fname,NULL,NULL,SW_SHOWDEFAULT);
 }
 
-void
-PARI_get_plot(void)
+INLINE void
+gp_get_display_sizes(long *dwidth, long *dheight, long *fwidth, long *fheight)
 {
   HDC hdc;
   TEXTMETRIC tm;
-  if (pari_plot.init) return;      /* pari_plot is already set */
 
-  pari_plot.init    = 1;
-  pari_plot.width   = GetSystemMetrics(SM_CXSCREEN)*4/5;
-  pari_plot.height  = GetSystemMetrics(SM_CYSCREEN)*4/5;
-  pari_plot.hunit   = pari_plot.width/100;
-  pari_plot.vunit   = pari_plot.height/100;
+  *dwidth  = GetSystemMetrics(SM_CXSCREEN);
+  *dheight = GetSystemMetrics(SM_CYSCREEN);
 
   hdc = GetDC(0);
   SelectObject(hdc, GetStockObject(DEFAULT_GUI_FONT));
   GetTextMetrics(hdc, &tm);
   ReleaseDC(0,hdc);
 
-  pari_plot.fwidth  = tm.tmAveCharWidth;
-  pari_plot.fheight = tm.tmHeight;
+  *fwidth  = tm.tmAveCharWidth;
+  *fheight = tm.tmHeight;
+}
+
+void
+gp_get_plot(PARI_plot *T)
+{
+  gp_get_plot_generic(T,gp_get_display_sizes);
+  T->draw = &draw;
 }
