@@ -437,7 +437,7 @@ truncr(GEN x)
   long s, e, d, m, i;
   GEN y;
   if ((s=signe(x)) == 0 || (e=expo(x)) < 0) return gen_0;
-  d = nbits2prec(e+1); m = remsBIL(e);
+  d = nbits2lg(e+1); m = remsBIL(e);
   if (d > lg(x)) pari_err_PREC( "truncr (precision loss in truncation)");
 
   y=cgeti(d); y[1] = evalsigne(s) | evallgefint(d);
@@ -461,7 +461,7 @@ floorr(GEN x)
   GEN y;
   if (signe(x) >= 0) return truncr(x);
   if ((e=expo(x)) < 0) return gen_m1;
-  d = nbits2prec(e+1); m = remsBIL(e);
+  d = nbits2lg(e+1); m = remsBIL(e);
   lx=lg(x); if (d>lx) pari_err_PREC( "floorr (precision loss in truncation)");
   y = cgeti(d+1);
   if (++m == BITS_IN_LONG)
@@ -563,12 +563,12 @@ umodiu(GEN y, ulong x)
 
 /* return |y| \/ x */
 GEN
-diviu_rem(GEN y, ulong x, ulong *rem)
+absdiviu_rem(GEN y, ulong x, ulong *rem)
 {
   long ly;
   GEN z;
 
-  if (!x) pari_err_INV("diviu_rem",gen_0);
+  if (!x) pari_err_INV("absdiviu_rem",gen_0);
   if (!signe(y)) { *rem = 0; return gen_0; }
 
   ly = lgefint(y);
@@ -1007,10 +1007,11 @@ red_montgomery(GEN T, GEN N, ulong inv)
   }
   if (carry)
   { /* Td > N overflows (k+1 words), set Td := Td - N */
+    GEN NE = N + k+2;
     Td = Te;
     Nd = Ne;
     t = subll(*++Td, *++Nd); *Td = t;
-    while (Td < (GEN)av) { t = subllx(*++Td, *++Nd); *Td = t; }
+    while (Nd < NE) { t = subllx(*++Td, *++Nd); *Td = t; }
   }
 
   /* copy result */
@@ -1037,8 +1038,7 @@ red_montgomery(GEN T, GEN N, ulong inv)
 
 /* EXACT INTEGER DIVISION */
 
-#if 1 && !defined(_WIN64) /* use undocumented GMP interface */
-                          /* mpz_divexact_ui is not LLP64 friendly */
+/* use undocumented GMP interface */
 static void
 GEN2mpz(mpz_t X, GEN x)
 {
@@ -1054,69 +1054,80 @@ mpz2GEN(GEN z, mpz_t Z)
   z[1] = evalsigne(l > 0? 1: -1) | evallgefint(labs(l)+2);
 }
 
+#ifdef mpn_divexact_1
+static GEN
+diviuexact_i(GEN x, ulong y)
+{
+  long l = lgefint(x);
+  GEN z = cgeti(l);
+  mpn_divexact_1(LIMBS(z), LIMBS(x), NLIMBS(x), y);
+  if (z[l-1] == 0) l--;
+  z[1] = evallgefint(l) | evalsigne(signe(x));
+  return z;
+}
+#elif 1 && !defined(_WIN64) /* mpz_divexact_ui is not LLP64 friendly   */
+                            /* assume y != 0 and the division is exact */
+static GEN
+diviuexact_i(GEN x, ulong y)
+{
+  long l = lgefint(x);
+  GEN z = cgeti(l);
+  mpz_t X, Z;
+  GEN2mpz(X, x);
+  Z->_mp_alloc = l-2;
+  Z->_mp_size  = l-2;
+  Z->_mp_d = LIMBS(z);
+  mpz_divexact_ui(Z, X, y);
+  mpz2GEN(z, Z); return z;
+}
+#else
 /* assume y != 0 and the division is exact */
+static GEN
+diviuexact_i(GEN x, ulong y)
+{
+  /*TODO: implement true exact division.*/
+  return divii(x,utoi(y));
+}
+#endif
+
 GEN
 diviuexact(GEN x, ulong y)
 {
+  GEN z;
   if (!signe(x)) return gen_0;
-  {
-    long l = lgefint(x);
-    mpz_t X, Z;
-    GEN z = cgeti(l);
-    GEN2mpz(X, x);
-    Z->_mp_alloc = l-2;
-    Z->_mp_size  = l-2;
-    Z->_mp_d = LIMBS(z);
-    mpz_divexact_ui(Z, X, y);
-    mpz2GEN(z, Z); return z;
-  }
+  z = diviuexact_i(x, y);
+  if (lgefint(z) == 2) pari_err_OP("exact division", x, utoi(y));
+  return z;
 }
 
 /* Find z such that x=y*z, knowing that y | x (unchecked) */
 GEN
 diviiexact(GEN x, GEN y)
 {
+  GEN z;
   if (!signe(y)) pari_err_INV("diviiexact",y);
+  if (!signe(x)) return gen_0;
   if (lgefint(y) == 3)
   {
-    GEN z = diviuexact(x, y[2]);
+    z = diviuexact_i(x, y[2]);
     if (signe(y) < 0) togglesign(z);
-    return z;
   }
-  if (!signe(x)) return gen_0;
+  else
   {
     long l = lgefint(x);
     mpz_t X, Y, Z;
-    GEN z = cgeti(l);
+    z = cgeti(l);
     GEN2mpz(X, x);
     GEN2mpz(Y, y);
     Z->_mp_alloc = l-2;
     Z->_mp_size  = l-2;
     Z->_mp_d = LIMBS(z);
     mpz_divexact(Z, X, Y);
-    mpz2GEN(z, Z); return z;
+    mpz2GEN(z, Z);
   }
+  if (lgefint(z) == 2) pari_err_OP("exact division", x, y);
+  return z;
 }
-#else
-/* assume y != 0 and the division is exact */
-GEN
-diviuexact(GEN x, ulong y)
-{
-  /*TODO: implement true exact division.*/
-  return divii(x,utoi(y));
-}
-
-/* Find z such that x=y*z, knowing that y | x (unchecked)
- * Method: y0 z0 = x0 mod B = 2^BITS_IN_LONG ==> z0 = 1/y0 mod B.
- *    Set x := (x - z0 y) / B, updating only relevant words, and repeat */
-GEN
-diviiexact(GEN x, GEN y)
-{
-  /*TODO: use mpn_bdivmod instead*/
-  if (!signe(y)) pari_err_INV("diviiexact",y);
-  return divii(x,y);
-}
-#endif
 
 /* assume yz != and yz | x */
 GEN
