@@ -3,6 +3,10 @@
 #define __DIVRR_GMP_LIMIT                4
 #define __EXPNEWTON_LIMIT                66
 #define __F2x_MUL_KARATSUBA_LIMIT        15
+#define __F2xqX_BARRETT_LIMIT            48
+#define __F2xqX_DIVREM_BARRETT_LIMIT     292
+#define __F2xqX_INVBARRETT_LIMIT         46
+#define __F2xqX_REM_BARRETT_LIMIT        292
 #define __Flx_BARRETT_HALFMULII_LIMIT    21
 #define __Flx_BARRETT_KARATSUBA_LIMIT    1172
 #define __Flx_BARRETT_MULII2_LIMIT       16
@@ -61,16 +65,21 @@
 #define __LOGAGM_LIMIT                   6
 #define __MULII_FFT_LIMIT                -1
 #define __MULII_KARATSUBA_LIMIT          -1
-#define __MULRR_MULII_LIMIT              74
+#define __MULRR_MULII_LIMIT              55
 #define __RgX_MUL_LIMIT                  9
 #define __RgX_SQR_LIMIT                  38
 #define __SQRI_FFT_LIMIT                 -1
 #define __SQRI_KARATSUBA_LIMIT           -1
+#define __SQRR_SQRI_LIMIT                12
 #else
 #define __AGM_ATAN_LIMIT                 89
 #define __DIVRR_GMP_LIMIT                4
 #define __EXPNEWTON_LIMIT                197
 #define __F2x_MUL_KARATSUBA_LIMIT        23
+#define __F2xqX_BARRETT_LIMIT            17
+#define __F2xqX_DIVREM_BARRETT_LIMIT     46
+#define __F2xqX_INVBARRETT_LIMIT         22
+#define __F2xqX_REM_BARRETT_LIMIT        48
 #define __Flx_BARRETT_HALFMULII_LIMIT    23
 #define __Flx_BARRETT_KARATSUBA_LIMIT    905
 #define __Flx_BARRETT_MULII2_LIMIT       647
@@ -129,9 +138,10 @@
 #define __LOGAGM_LIMIT                   45
 #define __MULII_FFT_LIMIT                -1
 #define __MULII_KARATSUBA_LIMIT          -1
-#define __MULRR_MULII_LIMIT              8
+#define __MULRR_MULII_LIMIT              19
 #define __RgX_MUL_LIMIT                  7
 #define __RgX_SQR_LIMIT                  34
 #define __SQRI_FFT_LIMIT                 -1
 #define __SQRI_KARATSUBA_LIMIT           -1
+#define __SQRR_SQRI_LIMIT                9
 #endif
