@@ -1,7 +1,5 @@
 #line 2 "../src/kernel/ix86/asm0.h"
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -21,7 +19,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 ASM addll mulll bfffo divll
 */
 #ifdef ASMINLINE
-/* Written by Bruno Haible, 1996-1998. */
+/* Written by Bruno Haible, 1996-1998.
+   addllx8/subllx8 by Bill Allombert, 2011.
+*/
 
 /* This file can assume the GNU C extensions.
    (It is included only if __GNUC__ is defined.) */
@@ -48,6 +48,22 @@ __extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b), __temp; \
   __value; \
 })
 
+#define addllx8(a,b,c,overflow) \
+do { long *__arg1 = a, *__arg2 = b, *__out = c; \
+     ulong __temp; \
+   __asm__ ("subl %5, %0 \n\t" \
+            "movl    (%2), %0 ; adcl    (%3),%0; movl %0,    (%4) \n\t" \
+            "movl  -4(%2), %0 ; adcl  -4(%3),%0; movl %0,  -4(%4) \n\t" \
+            "movl  -8(%2), %0 ; adcl  -8(%3),%0; movl %0,  -8(%4) \n\t" \
+            "movl -12(%2), %0 ; adcl -12(%3),%0; movl %0, -12(%4) \n\t" \
+            "movl -16(%2), %0 ; adcl -16(%3),%0; movl %0, -16(%4) \n\t" \
+            "movl -20(%2), %0 ; adcl -20(%3),%0; movl %0, -20(%4) \n\t" \
+            "movl -24(%2), %0 ; adcl -24(%3),%0; movl %0, -24(%4) \n\t" \
+            "movl -28(%2), %0 ; adcl -28(%3),%0; movl %0, -28(%4) \n\t" \
+            "adcl  %1, %1" \
+        : "=&r" (__temp), "=&r" (overflow) \
+        : "r" (__arg1), "r" (__arg2), "r" (__out), "g" (overflow), "0" ((ulong)0), "1" ((ulong)0)        : "cc"); \
+} while(0)
 
 #define subll(a,b) \
 __extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
@@ -67,6 +83,24 @@ __extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b), __temp; \
   __value; \
 })
 
+#define subllx8(a,b,c,overflow) \
+do { long *__arg1 = a, *__arg2 = b, *__out = c; \
+     ulong __temp; \
+   __asm__ ("subl %5, %0 \n\t" \
+            "movl    (%2), %0 ; sbbl    (%3),%0; movl %0,    (%4) \n\t" \
+            "movl  -4(%2), %0 ; sbbl  -4(%3),%0; movl %0,  -4(%4) \n\t" \
+            "movl  -8(%2), %0 ; sbbl  -8(%3),%0; movl %0,  -8(%4) \n\t" \
+            "movl -12(%2), %0 ; sbbl -12(%3),%0; movl %0, -12(%4) \n\t" \
+            "movl -16(%2), %0 ; sbbl -16(%3),%0; movl %0, -16(%4) \n\t" \
+            "movl -20(%2), %0 ; sbbl -20(%3),%0; movl %0, -20(%4) \n\t" \
+            "movl -24(%2), %0 ; sbbl -24(%3),%0; movl %0, -24(%4) \n\t" \
+            "movl -28(%2), %0 ; sbbl -28(%3),%0; movl %0, -28(%4) \n\t" \
+            "adcl  %1, %1" \
+        : "=&r" (__temp), "=&r" (overflow) \
+        : "r" (__arg1), "r" (__arg2), "r" (__out), "g" (overflow), "0" ((ulong)0), "1" ((ulong)0)        : "cc"); \
+} while(0)
+
+
 #define mulll(a,b) \
 __extension__ ({ ulong __valuelo, __arg1 = (a), __arg2 = (b); \
    __asm__ ("mull %3" \
@@ -86,7 +120,7 @@ __extension__ ({ ulong __valuelo, __arg1 = (a), __arg2 = (b), __temp; \
 #define divll(a,b) \
 __extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
    __asm__ ("divl %4" \
-        : "=a" /* %eax */ (__value), "=&d" /* %edx */ (hiremainder) \
+        : "=a" /* %eax */ (__value), "=d" /* %edx */ (hiremainder) \
         : "0" /* %eax */ (__arg1), "1" /* %edx */ (hiremainder), "mr" (__arg2)); \
    __value; \
 })
