@@ -37,7 +37,7 @@ addsi_sign(long x, GEN y, long sy)
   ly=lgefint(y);
   if (ly==3)
   {
-    const long d = y[2] - x;
+    const long d = (long)(uel(y,2) - (ulong)x);
     if (!d) return gen_0;
     z=cgeti(3);
     if (y[2] < 0 || d > 0) {
@@ -176,41 +176,6 @@ addsr(long x, GEN y) { return addsr_sign(x, y, signe(y)); }
 GEN
 subsr(long x, GEN y) { return addsr_sign(x, y, -signe(y)); }
 
-/* return x + 1, assuming x > 0 is a normalized t_REAL of exponent 0 */
-GEN
-addrex01(GEN x)
-{
-  long l = lg(x);
-  GEN y = cgetr(l);
-  y[1] = evalsigne(1) | _evalexpo(1);
-  y[2] = HIGHBIT | (((ulong)x[2] & ~HIGHBIT) >> 1);
-  shift_right(y, x, 3,l, x[2], 1);
-  return y;
-}
-/* return subrs(x,1) to the min of (prec(x), prec(x-1) + 1),
- * assuming x > 1 is a normalized t_REAL of exponent 0
- * [ goal: avoid the loss of significant bits form subrs ]*/
-GEN
-subrex01(GEN x)
-{
-  long i, sh, k, ly, lx = lg(x);
-  ulong u;
-  GEN y;
-  k = 2;
-  u = (ulong)x[2] & (~HIGHBIT);
-  while (!u) u = x[++k]; /* terminates: x not a power of 2 */
-  ly = (k == 2)? lx: lx - k+3; /* NB: +3, not +2: 1 extra word */
-  y = cgetr(ly);
-  sh = bfffo(u);
-  if (sh)
-    shift_left(y+2, x+k, 0, lx-k-1, 0, sh);
-  else
-  { for (i = 2; i < lx-k+2; i++) y[i] = x[k-2 + i]; }
-  for (i = lx-k+2; i < ly; i++) y[i] = 0;
-  y[1] = evalsigne(1) | evalexpo(- (bit_accuracy(k) + sh));
-  return y;
-}
-
 GEN
 addrr_sign(GEN x, long sx, GEN y, long sy)
 {
@@ -263,7 +228,7 @@ addrr_sign(GEN x, long sx, GEN y, long sy)
       const ulong sh = BITS_IN_LONG-m;
       GEN p1 = x; x = new_chunk(lx + lz + 1);
       shift_right(x,p1,2,lx, 0,m);
-      if (extend) x[lx] = p1[lx-1] << sh;
+      if (extend) uel(x,lx) = uel(p1,lx-1) << sh;
       avma = av; /* HACK: cgetr(lz) will not overwrite x */
     }
   }
@@ -298,7 +263,7 @@ addrr_sign(GEN x, long sx, GEN y, long sy)
     if (overflow)
     {
       z[1] = 1; /* stops since z[1] != 0 */
-      for (;;) { z[i] = (ulong) y[i]+1; if (z[i--]) break; }
+      for (;;) { z[i] = uel(y,i)+1; if (z[i--]) break; }
       if (i <= 0)
       {
         shift_right(z,z, 2,lz, 1,1);
@@ -314,8 +279,8 @@ addrr_sign(GEN x, long sx, GEN y, long sy)
   else
   {
     i = 2; while (i < lx && x[i] == y[i]) i++;
-    if (i==lx) return real_0_bit(ey+1 - bit_accuracy(lx));
-    f2 = ((ulong)y[i] > (ulong)x[i]);
+    if (i==lx) return real_0_bit(ey+1 - prec2nbits(lx));
+    f2 = (uel(y,i) > uel(x,i));
   }
   /* result is non-zero. f2 = (y > x) */
   i = lz-1; z = cgetr(lz);
@@ -326,7 +291,7 @@ addrr_sign(GEN x, long sx, GEN y, long sy)
     else        z[i] = subll(y[i], x[j--]);
     for (i--; j>=2; i--) z[i] = subllx(y[i], x[j--]);
     if (overflow) /* stops since y[1] != 0 */
-      for (;;) { z[i] = (ulong) y[i]-1; if (y[i--]) break; }
+      for (;;) { z[i] = uel(y,i)-1; if (y[i--]) break; }
     for (; i>=2; i--) z[i] = y[i];
     sx = sy;
   }
