@@ -253,7 +253,7 @@ sqrz_i(GEN z, GEN x, long lz)
   LOCAL_HIREMAINDER;
   LOCAL_OVERFLOW;
 
-  if (lz > MULRR_MULII_LIMIT)
+  if (lz > SQRR_SQRI_LIMIT)
   {
     pari_sp av = avma;
     GEN hi = sqrispec_mirror(x+2, lz-2);
@@ -841,6 +841,20 @@ int2u(ulong n) {
   for (i = 2; i < l; i++) z[i] = 0;
   *int_MSW(z) = 1UL << m; return z;
 }
+/* 2^n - 1 */
+GEN
+int2um1(ulong n) {
+  ulong i, m, l;
+  GEN z;
+  if (n == 0) return gen_0;
+
+  l = dvmduBIL(n, &m);
+  l += m? 3: 2;
+  z = cgetipos(l);
+  for (i = 2; i < l; i++) z[i] = ~0UL;
+  if (m) *int_MSW(z) = (1UL << m) - 1;
+  return z;
+}
 
 GEN
 shifti(GEN x, long n)
