@@ -16,9 +16,6 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "anal.h"
 #include "opcode.h"
 
-INLINE void
-clone_unlock_deep(GEN C) { if (isclone(C)) gunclone_deep(C); }
-
 /********************************************************************/
 /*                                                                  */
 /*                   break/next/return handling                     */
@@ -1291,7 +1288,7 @@ closure_eval(GEN C)
     case 20: sp-=20; f(st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7],st[sp+8],st[sp+9],st[sp+10],st[sp+11],st[sp+12],st[sp+13],st[sp+14],st[sp+15],st[sp+16],st[sp+17],st[sp+18],st[sp+19]); break; \
     default: \
       pari_err_IMPL("functions with more than 20 parameters");\
-      goto endeval; /*not reached*/ \
+      goto endeval; /*LCOV_EXCL_LINE*/ \
   }
 
     case OCcallgen:
@@ -1624,24 +1621,14 @@ GEN
 pareval(GEN C)
 {
   pari_sp av = avma;
-  long l = lg(C), i, pending = 0, workid;
-  struct pari_mt pt;
-  GEN worker, V, done;
+  long l = lg(C), i;
+  GEN worker;
   if (!is_vec_t(typ(C))) pari_err_TYPE("pareval",C);
   for (i=1; i<l; i++)
     if (typ(gel(C,i))!=t_CLOSURE)
       pari_err_TYPE("pareval",gel(C,i));
   worker = snm_closure(is_entry("_pareval_worker"), NULL);
-  V = cgetg(l, t_VEC);
-  mt_queue_start(&pt, worker);
-  for (i=1; i<l || pending; i++)
-  {
-    mt_queue_submit(&pt, i, i<l? mkvec(gel(C,i)): NULL);
-    done = mt_queue_get(&pt, &workid, &pending);
-    if (done) gel(V,workid) = done;
-  }
-  mt_queue_end(&pt);
-  return gerepilecopy(av, V);
+  return gerepileupto(av, gen_parapply(worker, C));
 }
 
 GEN
@@ -1663,7 +1650,7 @@ parvector(long n, GEN code)
   GEN worker = snm_closure(is_entry("_parvector_worker"), mkvec(code));
   GEN a, V, done;
   struct pari_mt pt;
-  mt_queue_start(&pt, worker);
+  mt_queue_start_lim(&pt, worker, n);
   a = mkvec(cgetipos(3)); /* left on the stack */
   V = cgetg(n+1, t_VEC);
   for (i=1; i<=n || pending; i++)
