@@ -1780,7 +1780,7 @@ sumnummonien(void *E, GEN (*eval)(void*,GEN), GEN n0, GEN tab, long prec)
     S = gadd(S, gmul(gel(vwt,i), eval(E, gel(vabs,i))));
     S = gprec_wensure(S, prec);
   }
-  return gerepileupto(av, gprec_w(S, prec));
+  return gerepilecopy(av, gprec_w(S, prec));
 }
 
 static GEN
@@ -1950,7 +1950,7 @@ intnumgauexp(void *E, GEN (*eval)(void*,GEN), GEN gN, GEN tab, long prec)
     S = gadd(S, gmul(gdiv(w,x), cxtoreal(t)));
     S = gprec_wensure(S, prec);
   }
-  return gerepileupto(av, gprec_w(S, prec));
+  return gerepilecopy(av, gprec_w(S, prec));
 }
 
 GEN
@@ -2068,7 +2068,7 @@ intnumainfrat(GEN F, long N, double r, long prec)
   for (k = lim; k >= v; k--) /* S <- (S + coeff(ser,k)/(k-1)) / N */
     S = gdivgs(gadd(S, gdivgs(sercoeff(ser,k), k-1)), N);
   if (v-2) S = gdiv(S, powuu(N, v-2));
-  return gerepileupto(av, gprec_w(S, prec));
+  return gerepilecopy(av, gprec_w(S, prec));
 }
 
 static GEN
@@ -2131,7 +2131,7 @@ static GEN
 sumnumrat_i(GEN F, GEN F0, GEN vF, long prec)
 {
   long B = prec2nbits(prec), vx, j, k, N;
-  GEN S, S1, S2, intf;
+  GEN S, S1, S2, intf, _1;
   double r;
   if (poldegree(F, -1) > -2) pari_err(e_MISC, "sum diverges in sumnumrat");
   vx = varn(gel(F,2));
@@ -2139,10 +2139,11 @@ sumnumrat_i(GEN F, GEN F0, GEN vF, long prec)
   get_kN((long)ceil(r), B, &k,&N);
   intf = intnumainfrat(F, N, r, prec);
   /* N > ratpolemax(F) is not a pole */
-  S1 = gmul2n(gtofp(poleval(F, utoipos(N)), prec), -1);
+  _1 = real_1(prec);
+  S1 = gmul2n(gmul(_1, gsubst(F, vx, utoipos(N))), -1);
   S1 = add_sumrfrac(S1, F, vF, N-1);
   if (F0) S1 = gadd(S1, F0);
-  S = gmul(real_1(prec), gsubst(F, vx, gaddgs(pol_x(vx), N)));
+  S = gmul(_1, gsubst(F, vx, gaddgs(pol_x(vx), N)));
   S = rfrac_to_ser(S, k + 2);
   S2 = gen_0;
   for (j = 2; j <= k; j += 2)
@@ -2302,7 +2303,7 @@ sumeulerrat(GEN F, GEN s, long a, long prec)
   u_forprime_init(&T, a, N);
   while ( (p = u_forprime_next(&T)) )
     res = gadd(res, gsubst(F, vx, gpow(utoipos(p), s, prec)));
-  return gerepileupto(av, gprec_w(res, prec));
+  return gerepilecopy(av, gprec_w(res, prec));
 }
 
 /* prod_{p prime, p >= a} F(p^s), F rational function */
@@ -2341,7 +2342,7 @@ prodeulerrat(GEN F, GEN s, long a, long prec)
   u_forprime_init(&T, a, N);
   while ( (p = u_forprime_next(&T)) )
     res = gmul(res, gsubst(F, vx, gpow(utoipos(p), s, prec)));
-  return gerepileupto(ltop, gprec_w(res, prec));
+  return gerepilecopy(ltop, gprec_w(res, prec));
 }
 
 /* Compute $\sum_{n\ge a}c(n)$ using Lagrange extrapolation.
@@ -2545,7 +2546,7 @@ sumnumlagrange(void *E, GEN (*eval)(void*,GEN,long), GEN a, GEN tab, long prec)
     s = gprec_wensure(s, prec);
   }
   if (!gequal1(S)) s = gdiv(s,S);
-  return gerepileupto(av, gprec_w(s, prec));
+  return gerepilecopy(av, gprec_w(s, prec));
 }
 
 GEN
