@@ -44,8 +44,8 @@ typedef struct {
   GEN p;
   GEN top;  /* absolute polynomial of a = z + pi (mod pr) */
   GEN topr; /* top mod p */
-  GEN eis;  /* relative polynomial of pi in terms of z (mod pr)  */
   GEN z;    /* root of T in terms of a (mod pr) */
+  GEN eis;  /* relative polynomial of pi in terms of z (mod pr)  */
   GEN pi;   /* prime element in terms of a */
   GEN ipi;  /* p/pi in terms of a (mod pr) (used to divide by pi) */
   GEN pik;  /* [1, pi, ..., pi^(e-1), pi^e/p] in terms of a (mod pr).
@@ -479,7 +479,7 @@ static GEN
 TamelyRamifiedCase(KRASNER_t *data)
 {
   long av = avma;
-  long g = ugcd(data->e, umodiu(data->qm1, data->e)); /* (e, q-1) */
+  long g = ugcdui(data->e, data->qm1); /* (e, q-1) */
   GEN rep, eis, Xe = gpowgs(pol_x(0), data->e), m = stoi(data->e / g);
 
   rep = zerovec(g);
@@ -697,7 +697,7 @@ setUnramData(KRASNER_t *d)
   {
     GEN L, z, T, p = d->p;
     d->T = T = init_Fq(p, d->f, d->v);
-    L = gel(factoru(ugcd(d->e, umodiu(d->qm1, d->e))), 1);
+    L = gel(factoru(ugcdui(d->e, d->qm1)), 1);
     z = gener_FpXQ_local(T, p, zv_to_ZV(L));
     d->u = ZpXQ_sqrtnlift(pol_1(d->v), d->qm1, z, T, p, d->r);
     d->frob = ZpX_ZpXQ_liftroot(T, FpXQ_pow(pol_x(d->v), p, T, p), T, p, d->r);
@@ -738,7 +738,7 @@ GetRamifiedPol(GEN p, GEN efj, long flag)
   data.b   = j%e;
   data.c   = (e+2*j)/e+1;
   data.q   = powiu(p, f);
-  data.qm1 = subis(data.q, 1);
+  data.qm1 = subiu(data.q, 1);
   data.v   = v;
   data.r   = 1 + (long)ceildivuu(2*j + 3, e); /* enough precision */
   data.pr  = powiu(p, data.r);
@@ -878,7 +878,7 @@ checkpols(GEN p, GEN EFJ, GEN pols)
     if (poldegree(p1, -1) == e*f) pari_err_BUG("fields are isomorphic");
     /*
       p1 = nfinitall(mkvec2(pol, mkvec(p)), 0, DEFAULTPREC);
-      p2 = gel(idealprimedec(p1, p), 1);
+      p2 = idealprimedec_galois(p1, p);
       if (!cmpis(mulii(gel(p2, 3), gel(p2, 4)), e*f)) pari_err_BUG("fields are isomorphic");
     */
   }
