@@ -540,7 +540,7 @@ choose_prime(primedata *S, GEN pol, GEN dpol)
     if (r == N || r >= BIL) continue;
 
     n = cgetg(r+1, t_VECSMALL); lcm = n[1] = degpol(gel(ff,1));
-    for (j=2; j<=r; j++) { n[j] = degpol(gel(ff,j)); lcm = clcm(lcm, n[j]); }
+    for (j=2; j<=r; j++) { n[j] = degpol(gel(ff,j)); lcm = ulcm(lcm, n[j]); }
     if (lcm <= oldlcm) continue; /* false when oldlcm = 0 */
 
     if (DEBUGLEVEL) err_printf("p = %lu,\tlcm = %ld,\torbits: %Ps\n",p,lcm,n);
@@ -619,7 +619,7 @@ compute_data(blockdata *B)
   if (DATA)
   {
     GEN Xm1 = gsub(pol_x(varn(pol)), gen_1);
-    GEN TR = addis(gel(DATA,5), 1);
+    GEN TR = addiu(gel(DATA,5), 1);
     GEN mTR = negi(TR), interp, bezoutC;
 
     if (DEBUGLEVEL>1) err_printf("... update (translate) an existing DATA\n\n");
@@ -811,15 +811,15 @@ subfields_poldata(GEN T, poldata *PD)
   PD->pol = T; setvarn(T, 0);
   if (nf)
   {
-    PD->den = Q_denom(nf_get_zk(nf));
+    PD->den = nf_get_zkden(nf);
     PD->roo = nf_get_roots(nf);
-    PD->dis = mulii(absi(nf_get_disc(nf)), sqri(nf_get_index(nf)));
+    PD->dis = mulii(absi_shallow(nf_get_disc(nf)), sqri(nf_get_index(nf)));
   }
   else
   {
     PD->den = initgaloisborne(T,NULL,nbits2prec(bit_accuracy(ZX_max_lg(T))), &L,NULL,&dis);
     PD->roo = L;
-    PD->dis = absi(dis);
+    PD->dis = absi_shallow(dis);
   }
 }
 
