@@ -22,6 +22,7 @@ struct mt_queue
 {
   long no;
   pari_sp avma;
+  struct pari_mainstack *mainstack;
   GEN input, output;
   GEN worker;
   long workid;
@@ -132,6 +133,7 @@ mt_queue_run(void *arg)
   pthread_cleanup_push(mt_queue_cleanup,NULL);
   LOCK(mq->pmut)
   {
+    mq->mainstack = pari_mainstack;
     mq->avma = av;
     pthread_cond_signal(mq->pcond);
   } UNLOCK(mq->pmut);
@@ -143,6 +145,7 @@ mt_queue_run(void *arg)
       while(!mq->input)
         pthread_cond_wait(&mq->cond, &mq->mut);
     } UNLOCK(&mq->mut);
+    pari_mainstack = mq->mainstack;
     avma = mq->avma;
     work = mq->input;
     pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);
@@ -150,6 +153,7 @@ mt_queue_run(void *arg)
     pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,NULL);
     LOCK(mq->pmut)
     {
+      mq->mainstack = pari_mainstack;
       mq->avma = av;
       mq->input = NULL;
       mq->output = done;
@@ -158,7 +162,7 @@ mt_queue_run(void *arg)
   }
   pthread_cleanup_pop(1);
 #ifdef __GNUC__
-  return NULL; /* NOT REACHED */
+  return NULL; /* LCOV_EXCL_LINE */
 #endif
 }
 
@@ -208,11 +212,16 @@ mtpthread_queue_get(struct mt_state *junk, long *workid, long *pending)
   } UNLOCK(&mt->pmut);
   BLOCK_SIGINT_END
   mq = mt->mq+last;
-  if (mq->output==err_e_STACK) pari_err(e_STACKTHREAD);
   done = gcopy(mq->output);
   mq->output = NULL;
   if (workid) *workid = mq->workid;
-  if (typ(done)==t_ERROR) pari_err(0,done);
+  if (typ(done) == t_ERROR)
+  {
+    if (err_get_num(done)==e_STACK)
+      pari_err(e_STACKTHREAD);
+    else
+      pari_err(0,done);
+  }
   mt->last = last;
   mt->pending--;
   *pending = mt->pending;
@@ -240,7 +249,19 @@ mtpthread_queue_submit(struct mt_state *junk, long workid, GEN work)
   {
     mq->output = NULL;
     mq->workid = workid;
-    mq->input = gcopy_avma(work, &mq->avma);
+    BLOCK_SIGINT_START
+    {
+      pari_sp av = avma;
+      struct pari_mainstack *st = pari_mainstack;
+      pari_mainstack = mq->mainstack;
+      avma = mq->avma;
+      mq->input = gcopy(work);
+      mq->avma = avma;
+      mq->mainstack = pari_mainstack;
+      pari_mainstack = st;
+      avma = av;
+    }
+    BLOCK_SIGINT_END
     pthread_cond_signal(&mq->cond);
   } UNLOCK(&mq->mut);
   mt->pending++;
@@ -259,7 +280,7 @@ mt_queue_reset(void)
     pthread_join(mt->th[i],NULL);
   pari_mt = NULL;
   BLOCK_SIGINT_END
-  if (DEBUGLEVEL) pari_warn(warner,"stop threads");
+  if (DEBUGLEVEL) pari_warn(warner,"stopping %ld threads", mt->n);
   for (i=0;i<mt->n;i++)
   {
     struct mt_queue *mq = mt->mq+i;
@@ -274,33 +295,35 @@ mt_queue_reset(void)
 }
 
 void
-mt_queue_start(struct pari_mt *pt, GEN worker)
+mt_queue_start_lim(struct pari_mt *pt, GEN worker, long lim)
 {
-  if (pari_mt || pari_mt_nbthreads == 1)
+  if (lim==0) lim = pari_mt_nbthreads;
+  else        lim = minss(pari_mt_nbthreads, lim);
+  if (pari_mt || lim <= 1)
     mtsingle_queue_start(pt, worker);
   else
   {
-    long NBT = pari_mt_nbthreads;
     struct mt_pstate *mt =
            (struct mt_pstate*) pari_malloc(sizeof(struct mt_pstate));
     long mtparisize = GP_DATA->threadsize? GP_DATA->threadsize: pari_mainstack->rsize;
     long mtparisizemax = GP_DATA->threadsizemax;
     long i;
-    mt->mq  = (struct mt_queue *) pari_malloc(sizeof(*mt->mq)*NBT);
-    mt->th  = (pthread_t *) pari_malloc(sizeof(*mt->th)*NBT);
-    mt->pth = (struct pari_thread *) pari_malloc(sizeof(*mt->pth)*NBT);
+    mt->mq  = (struct mt_queue *) pari_malloc(sizeof(*mt->mq)*lim);
+    mt->th  = (pthread_t *) pari_malloc(sizeof(*mt->th)*lim);
+    mt->pth = (struct pari_thread *) pari_malloc(sizeof(*mt->pth)*lim);
     mt->pending = 0;
-    mt->n = NBT;
+    mt->n = lim;
     mt->nbint = 0;
     mt->last = 0;
     pthread_cond_init(&mt->pcond,NULL);
     pthread_mutex_init(&mt->pmut,NULL);
     pari_thread_sync();
-    for (i=0;i<NBT;i++)
+    for (i=0;i<lim;i++)
     {
       struct mt_queue *mq = mt->mq+i;
       mq->no     = i;
       mq->avma   = 0;
+      mq->mainstack = NULL;
       mq->worker = worker;
       mq->input  = NULL;
       mq->output = NULL;
@@ -313,9 +336,9 @@ mt_queue_start(struct pari_mt *pt, GEN worker)
       else
         pari_thread_alloc(&mt->pth[i],mtparisize,(GEN)mq);
     }
-    if (DEBUGLEVEL) pari_warn(warner,"start threads");
+    if (DEBUGLEVEL) pari_warn(warner,"starting %ld threads", lim);
     BLOCK_SIGINT_START
-    for (i=0;i<NBT;i++)
+    for (i=0;i<lim;i++)
       pthread_create(&mt->th[i],NULL, &mt_queue_run, (void*)&mt->pth[i]);
     pari_mt = mt;
     BLOCK_SIGINT_END
