@@ -114,6 +114,14 @@ rand_Flx(long n, ulong l)
   return x;
 }
 
+static GEN
+rand_F2xqX(long n, GEN T)
+{
+  GEN x;
+  do x = random_F2xqX(n+1, 0, T); while (degpol(x) < n);
+  return x;
+}
+
 static GEN
 rand_FlxqX(long n, GEN T, ulong l)
 {
@@ -147,6 +155,14 @@ rand_NFlx(long n, ulong l)
   return gerepileuptoleaf(av, x);
 }
 
+static GEN
+rand_NF2xqX(long n, GEN T)
+{
+  pari_sp av = avma;
+  GEN x = F2xX_add(monomial(pol1_F2x(0),n,0), random_F2xqX(n, 0, T));
+  return gerepileupto(av, x);
+}
+
 static GEN
 rand_NFlxqX(long n, GEN T, long l)
 {
@@ -176,10 +192,12 @@ rand_NFpXQX(long n, GEN T)
 #define t_NFl2x   214
 #define t_FpX     300
 #define t_NFpX    310
-#define t_FlxqX   400
-#define t_NFlxqX  410
-#define t_FpXQX   500
-#define t_NFpXQX  510
+#define t_F2xqX   400
+#define t_NF2xqX  410
+#define t_FlxqX   500
+#define t_NFlxqX  510
+#define t_FpXQX   600
+#define t_NFpXQX  610
 
 static GEN
 rand_g(speed_param *s)
@@ -201,6 +219,8 @@ rand_g(speed_param *s)
     case t_NFl2x: return rand_NFlx(n,DFLT_mod2);
     case t_FpX:    return rand_FpX(n);
     case t_NFpX:   return rand_NFpX(n);
+    case t_F2xqX:  return rand_F2xqX(n, s->T);
+    case t_NF2xqX: return rand_NF2xqX(n, s->T);
     case t_FlxqX:  return rand_FlxqX(n, s->T, s->l);
     case t_NFlxqX: return rand_NFlxqX(n, s->T, s->l);
     case t_FpXQX:  return rand_FpXQX(n, s->T);
@@ -244,6 +264,8 @@ dftmod(speed_param *s)
     case t_NFl2x: s->l=DFLT_mod2;  return;
     case t_FpX:  s->p=LARGE_mod; return;
     case t_NFpX: s->p=LARGE_mod; return;
+    case t_F2xqX:  s->l=2;  dft_Flxq(s); return;
+    case t_NF2xqX: s->l=2;  dft_Flxq(s); return;
     case t_FlxqX:  s->l=DFLT_mod;  dft_Flxq(s); return;
     case t_NFlxqX: s->l=DFLT_mod;  dft_Flxq(s); return;
     case t_FpXQX:  s->p=LARGE_mod; dft_FpXQ(s); return;
@@ -284,9 +306,15 @@ static void disable(speed_param *s)
 static double speed_mulrr(speed_param *s)
 { TIME_FUN(mulrr(s->x, s->y)); }
 
+static double speed_sqrr(speed_param *s)
+{ TIME_FUN(sqrr(s->x)); }
+
 static double speed_mulii(speed_param *s)
 { TIME_FUN(mulii(s->x, s->y)); }
 
+static double speed_sqri (speed_param *s)
+{ TIME_FUN(sqri(s->x)); }
+
 static double speed_exp(speed_param *s)
 { TIME_FUN(mpexp(s->x)); }
 
@@ -306,9 +334,6 @@ static double speed_atan(speed_param *s)
   gatan(s->x, 0);
   TIME_FUN(gatan(s->x, 0)); }
 
-static double speed_sqri (speed_param *s)
-{ TIME_FUN(sqri(s->x)); }
-
 static double speed_Fp_pow(speed_param *s)
 { TIME_FUN( Fp_pow(s->x, subis(s->y,1), s->y)); }
 
@@ -384,6 +409,27 @@ static double speed_FpX_gcd(speed_param *s)
 static double speed_FpX_extgcd(speed_param *s)
 { GEN u,v; TIME_FUN(FpX_extgcd(s->x, s->y, s->p, &u, &v)); }
 
+static double speed_F2xqX_inv(speed_param *s)
+{ TIME_FUN(F2xqX_invBarrett(s->x, s->T)); }
+
+static double speed_F2xqX_divrem(speed_param *s)
+{
+  GEN r, x = rand_NF2xqX((degpol(s->x)-1)*2, s->T);
+  TIME_FUN(F2xqX_divrem(x, s->x, s->T, &r));
+}
+
+static double speed_F2xqX_rem(speed_param *s)
+{
+  GEN x = rand_NF2xqX((degpol(s->x)-1)*2, s->T);
+  TIME_FUN(F2xqX_rem(x, s->x, s->T));
+}
+
+static double speed_F2xqXQ_red(speed_param *s) {
+  GEN x = rand_NF2xqX((degpol(s->x)-1)*2, s->T);
+  GEN q = F2xqX_get_red(s->x, s->T);
+  TIME_FUN(F2xqX_rem(x, q, s->T));
+}
+
 static double speed_FlxqX_inv(speed_param *s)
 { TIME_FUN(FlxqX_invBarrett(s->x, s->T, s->l)); }
 
@@ -447,9 +493,9 @@ static double speed_FpXQX_gcd(speed_param *s)
 
 /* small coeffs: earlier thresholds for more complicated rings */
 static double speed_RgX_sqr(speed_param *s)
-{ TIME_FUN(RgX_sqr(s->x)); }
+{ TIME_FUN(RgX_sqr_i(s->x)); }
 static double speed_RgX_mul(speed_param *s)
-{ TIME_FUN(RgX_mul(s->x, s->y)); }
+{ TIME_FUN(RgX_mul_i(s->x, s->y)); }
 
 enum { PARI = 1, GMP = 2 };
 #ifdef PARI_KERNEL_GMP
@@ -467,6 +513,7 @@ static tune_param param[] = {
 {PARI,var(MULII_FFT_LIMIT),        t_INT, 500,0, speed_mulii,0.02},
 {PARI,var(SQRI_FFT_LIMIT),         t_INT, 500,0, speed_sqri,0.02},
 {0,   var(MULRR_MULII_LIMIT),      t_REAL,4,0, speed_mulrr},
+{0,   var(SQRR_SQRI_LIMIT),        t_REAL,4,0, speed_sqrr},
 {0,   var(Fp_POW_REDC_LIMIT),      t_INT, 3,100, speed_Fp_pow,0,0,&Fp_POW_BARRETT_LIMIT},
 {0,   var(Fp_POW_BARRETT_LIMIT),   t_INT, 3,0, speed_Fp_pow},
 {0,   var(INVNEWTON_LIMIT),        t_REAL,66,0, speed_inv,0.03},
@@ -521,6 +568,10 @@ static tune_param param[] = {
             speed_Flx_halfgcd,0,0,NULL,&Flx_MUL_MULII2_LIMIT},
 {0,  var(Flx_GCD_LIMIT),           t_Flx,10,0, speed_Flx_gcd,0.1},
 {0,  var(Flx_EXTGCD_LIMIT),        t_Flx,10,0, speed_Flx_extgcd},
+{0,  var(F2xqX_INVBARRETT_LIMIT),t_NF2xqX,10,0, speed_F2xqX_inv,0.05},
+{0,  var(F2xqX_BARRETT_LIMIT),   t_NF2xqX,10,0, speed_F2xqXQ_red,0.05},
+{0,  var(F2xqX_DIVREM_BARRETT_LIMIT), t_NF2xqX,10,0, speed_F2xqX_divrem,0.05},
+{0,  var(F2xqX_REM_BARRETT_LIMIT), t_NF2xqX,10,0, speed_F2xqX_rem,0.05},
 {0,  var(FlxqX_INVBARRETT_LIMIT),t_NFlxqX,10,0, speed_FlxqX_inv,0.05},
 {0,  var(FlxqX_BARRETT_LIMIT),   t_NFlxqX,10,0, speed_FlxqXQ_red,0.05},
 {0,  var(FlxqX_DIVREM_BARRETT_LIMIT), t_NFlxqX,10,0, speed_FlxqX_divrem,0.05},
@@ -599,7 +650,15 @@ time_fun(speed_function_t fun, speed_param *s, long enabled)
     }
   }
   pari_err(e_MISC,"couldn't measure time");
-  return -1.0; /* not reached */
+  return -1.0; /* LCOV_EXCL_LINE */
+}
+
+int
+cmpdat(const void *a, const void *b)
+{
+  struct dat_t *da =(struct dat_t *)a;
+  struct dat_t *db =(struct dat_t *)b;
+  return da->size-db->size;
 }
 
 void
@@ -612,6 +671,7 @@ add_dat(long size, double d)
   }
   dat[ndat].size = size;
   dat[ndat].d    = d; ndat++;
+  qsort(dat, ndat, sizeof(*dat), cmpdat);
 }
 
 void
@@ -663,13 +723,14 @@ analyze_dat(int final)
 }
 
 void
-Test(tune_param *param)
+Test(tune_param *param, long linear)
 {
   int since_positive, since_change, thresh, new_thresh;
   speed_param s;
   long save_var_disable = -1;
   pari_timer T;
   pari_sp av=avma;
+  long good = -1, bad = param->min_size;
 
   if (param->kernel == AVOID) { print_define(param->name, -1); return; }
 
@@ -719,34 +780,59 @@ Test(tune_param *param)
 
 #define SINCE_POSITIVE 20
 #define SINCE_CHANGE 50
-    /* Stop if method B has been consistently faster for a while */
-    if (d >= 0)
-      since_positive = 0;
-    else
-      if (++since_positive > SINCE_POSITIVE)
-      {
-        if (option_trace >= 1)
-          diag("Stop: since_positive (%d)\n", SINCE_POSITIVE);
-        break;
-      }
-    /* Stop if method A has become slower by a certain factor */
-    if (t1 >= t2 * param->stop_factor)
+    if (linear)
     {
-      if (option_trace >= 1)
-        diag("Stop: t1 >= t2 * factor (%.1f)\n", param->stop_factor);
-      break;
-    }
-    /* Stop if threshold implied hasn't changed for a while */
-    if (thresh != new_thresh)
-      since_change = 0, thresh = new_thresh;
-    else
-      if (++since_change > SINCE_CHANGE)
+      /* Stop if method B has been consistently faster for a while */
+      if (d >= 0)
+        since_positive = 0;
+      else
+        if (++since_positive > SINCE_POSITIVE)
+        {
+          if (option_trace >= 1)
+            diag("Stop: since_positive (%d)\n", SINCE_POSITIVE);
+          break;
+        }
+      /* Stop if method A has become slower by a certain factor */
+      if (t1 >= t2 * param->stop_factor)
       {
         if (option_trace >= 1)
-          diag("Stop: since_change (%d)\n", SINCE_CHANGE);
+          diag("Stop: t1 >= t2 * factor (%.1f)\n", param->stop_factor);
         break;
       }
-    s.size += maxss((long)floor(s.size * param->step_factor), 1);
+      /* Stop if threshold implied hasn't changed for a while */
+      if (thresh != new_thresh)
+        since_change = 0, thresh = new_thresh;
+      else
+        if (++since_change > SINCE_CHANGE)
+        {
+          if (option_trace >= 1)
+            diag("Stop: since_change (%d)\n", SINCE_CHANGE);
+          break;
+        }
+      s.size += maxss((long)floor(s.size * param->step_factor), 1);
+    }
+    else
+    {
+      if (t2 <= t1)
+        new_thresh = good = s.size;
+       else
+        bad = s.size;
+      if (bad == -1)
+        linear = 1;
+      else if (good == -1)
+      {
+        long new_size = minss(2*s.size,param->max_size-1);
+        if (new_size==s.size) linear = 1;
+        s.size = new_size;
+      }
+      else if (good-bad < 20*param->step_factor*bad)
+      {
+        linear = 1;
+        new_thresh = s.size = bad + 1;
+      }
+      else s.size = (good+bad)/2;
+      err_printf("bad= %ld good = %ld thresh = %ld linear = %ld\n",bad, good, thresh, linear);
+    }
     if (s.size >= param->max_size)
     {
       if (option_trace >= 1)
@@ -784,6 +870,7 @@ int
 main(int argc, char **argv)
 {
   int i, r, n = 0;
+  int linear = 1;
   GEN v;
   pari_init(8000000, 2);
   DFLT_mod = 27449;
@@ -806,7 +893,8 @@ main(int argc, char **argv)
         case 't': option_trace++;
           while (*++s == 't') option_trace++;
           break;
-
+        case 'd': linear = 1-linear;
+          break;
         case 'p':
           if (!*++s)
           {
@@ -836,8 +924,8 @@ main(int argc, char **argv)
       v[n++] = r;
     }
   }
-  if (n) { for (i = 0; i < n; i++) Test(&param[ v[i] ]); return 0; }
+  if (n) { for (i = 0; i < n; i++) Test(&param[ v[i] ], linear); return 0; }
   n = numberof(param);
-  for (i = 0; i < n; i++) Test(&param[i]);
+  for (i = 0; i < n; i++) Test(&param[i], linear);
   return 0;
 }
